/**
 * @file phal_hw_api.h
 * @brief PHAL Hardware Library API
 */

/* particular definitions */
#include <assert.h>
#include <stdio.h>

#define RUN_AS_PROC	/**< Define to load daemons as stand-alone processes */

#define xprintf printf	
#define time_t struct timeval

#define PLATFORM_LNX26	1
#define PLATFORM_C6000	2

#define  NEW(p) ((p) = calloc(1,(long)sizeof *(p)))
#include <assert.h>


/* common definitions begin here ... */


 

/** @defgroup base Base API Functions
 * 
 * This functions include some common and basic 
 * functions required to use the HW API Library
 *
 * @{
 */



/** Initialize Hardware API.
 *
 * Initializes hwapi library. This function must be called before
 * calling any other hwapi function.
 *
 * 
 * @return >0 if successfully registered
 * @return <0 if failed
 *
 */

int hwapi_init(void);


/** Exit function. 
 *
 * Clears hwapi resources, no more hwapi functions can be called after 
 * this functions is called. 
 * 
 * The function does NOT kill process thread, just clears shared memory resources
 * 
 */

void hwapi_exit();


/** Check Daemon Presence Function:
 *
 * @depreciated
 *
 * Checks if daemon with name 'daemon_name' have been registered in the same
 * HWAPI Library
 * 
 * @param daemon_name Name of the daemon to check
 *
 * @return 1 dameon is present and registered
 * @return 0 daemon is present but not registered
 * @return -1 daemon is not present
 */

/*int hwapi_is_daemon(char *daemon_name);*/

/** Get my object id 
 */
int hwapi_myobjid (void);
 

/** Process Relinquish.
 * 	
 * Function used by processes to relinquish their cpu ownership.
 * This function must be called by SERVICES API before the end of the
 * time slot, otherwise RTCFault will be flagged.
 * 
 * @param slots number of slots to relinquish
 * @return >=0 number of processes
 */
void hwapi_relinquish(int slots);

/** Daemon Relinquish.
 *
 * Same function as proc_relinquish but used by SW daemons. 
 * Here time slotting is not required (in Linux), so processes are
 * sleeped a fraction of the time slot
 * 
 */
void hwapi_relinquish_daemon();


/** @} */ // end group











/** @defgroup itf Packet Interface Management Functions
 * 
 * HW API provides the user to a set of functions to create FIFO-like packet
 * oriented data interface. 
 * 
 * This interfaces have a read and a write side. Once it is created, the user
 * must attach a file descriptor to the read or write side, thus, communication is possible
 * and packets written in the write side will be available in the read side. 
 *
 * A name and an owner Id (object Id) can be assigned to any of the sides so separate 
 * processes can share an interface easier.
 * 
 * @{
 */


#define FLOW_READ_ONLY	1
#define FLOW_WRITE_ONLY	2
#define FLOW_READ_WRITE	3
#define FLOW_WRITE_READ	4




/** Create Interface.
 *
 * Creates a packet-oriented internal data interface.  
 * If id parameter is different of 0 it will be the 
 * assigned id for the interface. Otherwise it will 
 * be generated.
 * 
 * @param id 0 for automatic, else for fixed
 * @param size maximum packet size allowed in words
 * 
 * @return !=0 Id of the interface
 * @return =0 error creating interface
 */
int hwapi_itf_create(int id, int size);



/** Delete Interface.
 *
 * Deletes internal data interface with id='id' (and clears resources)
 * 
 * @param id Id of the interface to delete
 *
 * @return >0 Deleted successfully
 * @return <0 Error deleting interface
 */
int hwapi_itf_delete(int id);


/** Delete all own interfaces.
 *
 * Given an object name, deletes all interfaces which it has been
 * matched in write mode 
 * 
 * @param name Name of the object
 *
 * @return 1 Deleted successfully
 * @return 0 Error deleting interface
 */ 
  int
hwapi_itf_delete_own (char *obj_name);

/** Attach to an Interface.
 *
 * Creates a file descriptor to use the interface. This function
 * should be called 'after' creating the interface and 'before' using it.
 * 
 * It must be attached in FLOW_READ_ONLY (for reading) or FLOW_WRITE_ONLY 
 * mode (for writting). A file descriptor will be returned which can be used
 * to read/write packets
 * 
 * @param id Id of the interface created with hwapi_create_itf()
 * @param mode FLOW_READ_ONLY or FLOW_WRITE_ONLY constants
 *
 * @return >0 File descriptor
 * @return =0 Interface not available
 * @return <0 Error attaching interface
 */
int hwapi_itf_attach(int id, int mode);



/** Unattach from an Interface.
 *
 * Un-attaches a file descriptor from the (previously attached)
 * interface in mode 'mode'
 * 
 * @param fd file descriptor of the interface
 * @param mode FLOW_READ_ONLY or FLOW_WRITE_ONLY constants
 * 
 * @return >0 Successfully unattached
 * @return <0 Error unattaching
 */
int hwapi_itf_unattach(int fd, int mode);



/** Send a packet.
 *
 * Sends a packet of data through a data interface. The user needs
 * to provide a file descriptor, get after calling the attaching function.
 * 
 * @param fd file descriptor
 * @param buffer pointer to data to send
 * @param size size of the packet in words (32-bit)
 * 
 * @return >0 Successfully sended. Returns amount of words sended. 
 * @return =0 Packet could not be send (full queue)
 * @return <0 Error sending packet
 */
int hwapi_itf_snd(int fd, int *buffer,int size);


/** Receive a packet.
 *
 * Recevies a packet through a file descriptor. User must
 * indicate provided buffer's length (in words).
 *
 * @param fd file descriptor
 * @param buffer pointer to buffer where receive data
 * @param size size of the buffer in words (32-bit)
 * 
 * @return >0 Successfully received. Returns amount of words. 
 * @return =0 No packets available (empty queue)
 * @return <0 Error receiving packet
 */
int hwapi_itf_rcv(int fd, int *buffer, int size);

/** Link Physical Interface to a logical Interface
 *
 * Links a physical interface (external) to an internal packet
 * oriented interface. After the link is successfully realized,
 * all packets written to the internal interface will be bridged
 * (by RUNPH_NET) towards the physical interface. Conversely, all
 * packets received to the phy. itf. will be bridged to the internal 
 * interface.
 * 
 * Interface must be created before calling this function. Then, in 
 * order to use it, it must be attached.
 * 
 * @param itf_id Local interface Id
 * @param phy_itf_id External Physical Interface Id
 * @param mode FLOW_READ_ONLY or FLOW_WRITE_ONLY constant
 *
 * @return >0 Successfully linked
 * @return <0 Error linking
 */
int hwapi_itf_link_phy(int itf_id, int phy_itf_id, int mode);



/** Match Interface.
 * 
 * Assign an object id and interface name for the read and write side 
 *
 * @return 1 if ok 0 if error
 */
int hwapi_itf_match (int itf_id, char *w_obj, char *w_itf, char *r_obj, char *r_itf);


/** Get itf id given a name/obj_id.
 *
 * Inverse function to the hwapi_itf_match(). Now, given an itf name
 * and object id, returns the itf_id.
 */
int hwapi_itf_find (char *obj_name, char *itf_name, int mode);

/** @} */ // end group


/** @defgroup hw Hardware Monitoring Functions
 * 
 * A set of functions are available to gather information about
 * underlying hardware characteristics and current status. 
 * 
 * @{
 */

/** Structure For CPU Information 
 * @todo Define other RT or monitoring info
 */
struct hwapi_cpu_i
{
  int pe_id;
  float C;		/**< Computing Capacity (in MIPS) */
  float intBW;		/**< Internal memory BW (in Mbps) */
  int status;		/**< status flag */
  int tslen_usec;	/**< Duration of the time slot (in usec) */
  int plat_family;	/**< Platform family identification */
  int debug_level;	/**< Platform debug level */
  char disable_stats;
  char disable_logs;
  int cpu_usage;
  int total_fifo;
  int used_fifo;
};




/** Structure for External interfaces Information 
 */
struct hwapi_xitf_i
{
  int id; 	/**< Id of the interface */
  int mode;	/**< r/w mode */
  int delay;	/**< delay (in usec) */
  float BW;	/**< BW in kilobits (flost) */
};


/** Physical Interface Information.
 *	
 *	
 *
 * Obtains information of the Physical External interfaces. It is read from
 * the physical interfaces configuration file (RUNPH Launcher).
 * 
 * User must provide the maximum amount of external interfaces he wants to read.
 * It depends on the length of the xitf buffer (number of elements)
 * 
 * @param xitf Pointer to information structure (check phal_hw_api.h)
 * @param max_itf Maximum number of interfaces (buffer length, in elements)
 *
 * @return >=0 Number of interfaces read
 * @return <0 error
 */
int hwapi_hwinfo_xitf(struct hwapi_xitf_i *xitf, int max_itf);

/** Get Processor Information.
 *
 *
 * @todo Definition of structure, dynamic change? battery? etc.  
 * 
 * 
 * Function used to gather processing resource (CPU) information. 
 * Data is writted to the struct pointed by info:
 * 
 * @see hwapi_cpu_i 
 * @param info CPU information struct
 *
 */
void hwapi_hwinfo_cpu(struct hwapi_cpu_i *info);


/** Set processor id
 * @param pe_id New processor id
 */
int hwapi_hwinfo_setid (int pe_id);

/** @} */ //end group




/** @defgroup proc Process Management Functions 
 * 
 * This group of functions provide tools for creating, monitoring and 
 * changing the status of processes (objects) running in the local processor.
 *  
 * @{
 */

#define PROC_MODE_LINKED	0x1
#define PROC_MODE_COMPILED	0x2
#define PROC_MODE_SOURCE	0x4

/** Object status structure
 */
struct hwapi_proc_status
{
  int cur_status;
  int next_status;
  int next_tstamp;
};

#define MEAN_CPU_LENGTH		8

/** Process Information/Monitoring Structure 
 */
struct hwapi_proc_i 
{
  /* char proc_name[24];*/
  char obj_name[24];
  char app_name[24];
  char path[128];
  int app_id;
  int obj_id;
  int pid;
  int rel_tstamp;
  int change_progress;
  int relinquished;
  time_t cpu_time[3];
  int _mean_cpu_usec[MEAN_CPU_LENGTH];
  int mean_cpu_usec;
  int tti;
  struct hwapi_proc_status status;
};

/** Process loading propierties
 */
struct hwapi_proc_launch 
{
  int app_id;
  int obj_id;
  char *obj_name;
  char *app_name;
  char *exe_name;
  int blen;
  int wlen;
  /* only required by absolute-addressing computers */
  unsigned int mem_st;
  unsigned int mem_sz;
  unsigned int stack_st;
  unsigned int stack_sz;
};
  

/** Request New Proces Creation.
 * 
 *
 * Use this function before creating a new process to request a 
 * memory region where load the process.
 *
 * @todo New process creation for absolute addressing computers
 *  
 */
/*int hwapi_proc_req(void);*/


/** Create a New Process.
 *
 * This function creates a process and launches it. As the process resides in user
 * memory (pdata buffer), it must be writed to the disk before launching it. This file
 * will be deleted when the process does.
 *
 * @todo (in Linux) path to downloaded software should be either relative or guessed but not a constant
 * 
 * @param pdata Pointer to program data
 * @param pinfo Pointer to process loading information structure
 * 
 * @return >0 Pid of the process
 * @return <0 error
 */
int hwapi_proc_create(char *pdata, struct hwapi_proc_launch *pinfo);



/** Remove Process Function:
 *	
 * Removes a process from the system and de-allocates all resources
 * it was using.
 * 
 * @param obj_id Id of the object to remove
 *
 * @return >0 success
 * @return <0 error
 */
int hwapi_proc_remove(int obj_id);

/** Get object status structure pointer 
 */
int hwapi_proc_mystatus (void);


/** Set Object status.
 *
 * Sets a new object status. Status can only be changed 
 * if the object successfully changed its previous status 
 * and the hwapi_proc_ackstatus() has been called.
 */
int hwapi_proc_status_new(int obj_id, int next_status, int next_tstamp);


/** Confirm a successful change in an object status.
 * After this call, another change in the status could
 * be performed.
 */
int hwapi_proc_status_ack(int obj_id);



/** Proces Monitoring.
 * 
 * Function used to gather information about current process status,
 * future status, cpu timing and other related information
 * 
 * @param proc Pointer to proc_db structure
 * @param obj_id id of the object assigned to the process
 * 
 * @return >0 ok
 * @return <0 error (not found)
 */
int hwapi_proc_info(int obj_id, struct hwapi_proc_i *pinfo);

/** My process information.
 * 
 * Fills hwapi_proc_i structure with the caller's info.
 * 
 * @param proc Pointer to hwapi_proc_i structure
 * 
 */ 
  int
hwapi_proc_myinfo (struct hwapi_proc_i *pinfo);



/** Get Processes List.
 *
 * Returns the number of processes (not including daemons, so, only objects,
 * running in the local processor and fills the pinfo pointer with 
 * their hwapi_proc_i structures (until a maximum of nelems).
 * 
 * @param pinfo array of hwapi_proc_i where info must be saved
 * @param nelems size of the array in elements
 *
 * @return >=0 number of processes
 */
int hwapi_proc_list(struct hwapi_proc_i *pinfo, int nelems);

/** @} */








/** @defgroup stats Global Variable Functions
 *
 * The Hardware API enables processes to share global variables in an easy, fast
 * and efficient way.
 * 
 * 
 * @{
 */
 
/** Variable structure information
*/
struct hwapi_var_i {
	char name[24];
	int obj_id;
	int stat_id;
	int size;
	int cur_size;
	int opts;
	int value_offset;
};

/** Create a new shared variable.
*/
int hwapi_var_create(char *name, int size, int opts);

/** Delete a variable
 */
int hwapi_var_delete (int stat_id);


/** List all shared variables.
 * 
 * Obtains a list of all variables created in the system 
 * matching the given option (set to 0 to list all)
 */
int hwapi_var_list (int opts, struct hwapi_var_i *var_list, int nof_elems);

/** Modify a variable option
 */
int hwapi_var_setopt (int stat_id, int new_opt);

/** Get a variable option
 */
int hwapi_var_getopt (int stat_id);

/** Read a variable value
 */
int hwapi_var_read (int stat_id, void *value, int buff_sz);

/** Write a variable value
 */
int hwapi_var_write (int stat_id, void *value, int value_sz);

/* @} */

/** @defgroup others
 * @{
 */

void get_time_interval(time_t * tdata);

/** @} */

