 /** @file hwapi_backd.c
 * @brief Hardware API Background Daemon for Linux
 *
 * This file contains the implementation of the HWAPI
 * Background Daemon for Linux.
 *
 * This program creats all resources that will be used by
 * PHAL Software Daemons as well as acts as a bridge
 * towards external network interfaces.
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <signal.h>
#include <errno.h>
#include <string.h>
#include <sched.h>
#include <time.h>
#include <ctype.h>
#include <sys/ipc.h>
#include <sys/shm.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/socket.h>
#include <sys/wait.h>
#include <sys/stat.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <fcntl.h>

#include <sys/msg.h>

#include "mem.h"

#include "phal_hw_api.h"
#include "hwapi_backd.h"
#include "hwapi_netd.h"

#define REMOVE_FILE

/*Info for at exit called function*/
int exit_level = 0;

/*HWAPI shared memory id and pointer*/
key_t shm_key;
int shm_id;
struct hw_api_db *hw_api_db= NULL;
struct net_db *net_db;

/*HWAPI msg mailboxes */
int api_msg_key, data_msg_key;
int data_msg_id, api_msg_id;


/* priority for objects */
int objects_nice_prio;
int do_nice=0;

/*Helpful internal functions*/
void print_frame(char *title);

void sig_chld(int sig, siginfo_t * sinfl, void *v);
void sig_usr(int sig, siginfo_t * sinfl, void *v);
void sig_int(int sig, siginfo_t * sinfl, void *v);

void clear_files();
void clear_zombies();

void clear_shm();
void remove_file(int i);

int Answer_command(int cmd, int *packet, int len);

void launch_daemons();
void kill_chlds();

void clear_zombies();

/* memory for messages */
struct msgbuff msg;

/* string */
char filelockname[64];

/* runph pid */
pid_t runph_pid;

#define LAUNCH_DAEMONS	10

struct launch_daemons daemons[LAUNCH_DAEMONS];

/* number of running daemons */
int nof_daemons;

#define DEB_XITF

/** @defgroup launcher PHAL Hardware API Background Daemon for Linux
 *
 * This program creats all resources that will be used by
 * PHAL Software Daemons as well as acts as a bridge
 * towards external network interfaces.
 *
 * Communication with Daemons is done through two interfaces:
 * 	- Shared Memory: Common definitions, running processes, shared
 *		interfaces, etc.
 *	- Commands (messages): requesting for some services (see Answer_command()) like creating
 * a new process, starting an external interface, etc.
 *
 * @{
 */

/** Exit Function
 *
 * Function called at exit to clear resources
 */
void exit_runph()
{
	struct sigaction action;
	int sleep_time;

	if (getpid() != runph_pid) {
		return;
	}

	//    printf("----- EXIT LEVEL %d -----\n", exit_level);

	if (exit_level == 10)
		sleep_time = 100;
	else
		sleep_time = 50;

	if (exit_level > 3) {
		printf("Killing chlds...\n");

		kill_chlds(sleep_time);
		clear_files();

	}
	if (exit_level > 2) {
		printf("Killing net chlds...\n");

	}
	if (exit_level > 2) {
		clear_zombies();
	}

	if (exit_level > 1) {

		printf("cleaning shared memory resources...\n");

		clear_shm();

	}

	if (exit_level > 0) {
		/* deleting file */
		if (strlen(filelockname) > 0) {
			if (remove(filelockname))
				perror("remove");
		}
	}

	printf("Exit.\n");
	fflush(stdout);

}

/*************
 *   USAGE   *
 *************/
void usage(char *prog)
{
	printf("%s [work_path] -c [platform_cfg_file] -i [xitf_cfg_file] -p [prio] -d [debug_mode]\n", prog);
	printf("\t [work_path]: Path for file resources (for manager daemons)\n");
	printf("\t -c [platform_config_file]: Platform configuration file\n");
	printf("\t -i [xitf_cfg_file]: External interface configuration file\n");
	printf("\t -p: Number indicating kernel priority for objects (must be root)\n");
	printf("\t -d: Number indicating debug mode level (daemons will read it)\n");
	exit(0);
}

void clear_shm()
{
	/* if shared memory is created */
	if (shmget(shm_key, sizeof(struct hw_api_db), 0) != -1) {
		/* unattach pointer if exists*/
		if (hw_api_db) {
			shmdt(hw_api_db);
		}
		/* and clear shared memory */
		shmctl(shmget(shm_key, sizeof(struct hw_api_db), 0),
		IPC_RMID, NULL);
	}

	/* same for msg for api */
	if (msgget(api_msg_key, 0) != -1) {
		msgctl(msgget(api_msg_key, 0), IPC_RMID, NULL);
	}
	if (msgget(data_msg_key, 0) != -1) {
		msgctl(msgget(data_msg_key, 0), IPC_RMID, NULL);
	}

}

void create_shm()
{
	struct msqid_ds info;

	if ((shm_id = shmget(shm_key, sizeof(struct hw_api_db),
	IPC_CREAT | IPC_EXCL | 0666)) == -1) {
		fprintf(stderr, "HWAPI_BACKD: Shared memory could not be created\n");
		perror("shmget");
		exit(-1);
	}
	if ((hw_api_db = (struct hw_api_db *) shmat(shm_id, NULL, 0))
			== (struct hw_api_db *) -1) {
		fprintf(stderr, "Shared memory could not be attached\n");
		perror("shmat");
		exit(-1);
	}
	printf("HWAPI_BACKD: Local shared memory created with id %d size %d\n",
			shm_id, sizeof(struct hw_api_db));

	/* Create daemons' api communications mbox */
	if ((api_msg_id = msgget(api_msg_key, IPC_CREAT | IPC_EXCL | 0666))
			== -1) {
		fprintf(stderr,
		"HWAPI_BACKD: Message queue for deamons control could not be created\n");
		perror("msgget");
		exit(-1);
	}
	printf("HWAPI_BACKD: Msg queue for api comms created with id %d\n",
			api_msg_id);

	/* Create daemons data communications mbox */
	if ((data_msg_id = msgget(data_msg_key, IPC_CREAT | IPC_EXCL | 0666))
			== -1) {
		fprintf(stderr,"HWAPI_BACKD: Message queue for data could not be created\n");
		perror("msgget");
		exit(-1);
	}

	/* Set maximum packet size for data mbox */
	msgctl(data_msg_id, IPC_STAT, &info);
	info.msg_qbytes = 1024*1024;
	if (msgctl(data_msg_id, IPC_SET, &info) == -1) {
		printf ("HWAPI_BACKD: Error setting size of message queue (1Mb).\n"
				"Please set /proc/sys/kernel/msgmnb to 1048576.\n");
		perror("msgctl");
		exit(-1);
	}
	msgctl(data_msg_id, IPC_STAT, &info);
	printf("HWAPI_BACKD: Msg queue for data comms created with id %d, size %d\n",
			data_msg_id, info.msg_qbytes);

}

/** Init Shared Memory
 *
 * Initializes shared memory regions
 */
void init_shm_db()
{
	memset(hw_api_db, 0, sizeof(struct hw_api_db));

	/* create memory pool object for statistics variables */
	/* default slot size to 32 bytes */
	if (!mem_new(hw_api_db->var_table, VAR_TABLE_SZ, 32)) {
		printf("HWAPI_BACKD: Error creating shared memory!\n");
	}

	//    printf ("table @ %p (%d)\n", hw_api_db->var_table, hw_api_db->var_table[0]);
	/* init time */
	get_time(&hw_api_db->init_time);
}



#define REPORT_WINDOW	100
int fifo_usage[REPORT_WINDOW];
int report_idx=0;

char line[64];
void report_hwinfo(int fd) {
	struct msqid_ds msgstatus;
	
	msgctl(data_msg_id, IPC_STAT, &msgstatus);
	
	fifo_usage[report_idx] = msgstatus.__msg_cbytes;
	report_idx++;
	if (report_idx == REPORT_WINDOW) {
		int i;
		
		for (i=0;i<REPORT_WINDOW;i++) {
			sprintf(line, "%d\n", fifo_usage[i]);
			write(fd, line, strlen(line));
		}
		report_idx=0;
	}
}



/** Main Daemon Function
 *
 * This is the main function for the Daemon. It creates shared memory resources
 * reads some configuration files (for interfaces) and launches network childs.
 */
int main(int argc, char **argv)
{

	struct hw_api_db tmp_db;
	struct sigaction action;
	int i, n, debug_level=0;
	int fd;
	FILE *fit=NULL, *fcpu=NULL;
	char loc_str[64];
	char *daemons_file=NULL, *xitf_file=NULL, *res_path=DEFAULT_RES_PATH;
	char *report_file=NULL;
	int report_file_fd=0;

	exit_level = 0;
	
	/** Register exit function */
	atexit(exit_runph);
	runph_pid = getpid();

	/* Init external interfaces database */
	memset(&net_db, 0, sizeof(struct net_db));
	memset(filelockname, 0, 64);

	/* if invalid arguments print usage */
	if (argc < 2) {
		usage(argv[0]);
		exit(-1);
	}

	daemons_file = DEFAULT_DAEMONS_FILE;
	xitf_file = DEFAULT_XITF_FILE;

	/*Read parameters */

	/* first one is working path */
	res_path = argv[1];
	if (open(res_path, O_RDONLY) == -1) {
		printf("HWAPID: Error opening work path %s\n", res_path);
		perror("open");
		exit(-1);
	}

	for (i = 2; i < argc; i++) {
		if (argv[i][0] == '-') {
			switch (argv[i][1]) {
			case 'c':
				if ((argc - 1) > i) {
					/* save daemons config file */
					daemons_file = argv[i + 1];
					i++;
				} else {
					printf("Switch '-c' needs a config file name.\n");
					usage(argv[0]);
				}
				break;
			case 'i':
				if ((argc - 1) > i) {
					/* save daemons config file */
					xitf_file = argv[i + 1];
					i++;
				} else {
					printf("Switch '-I' needs a config file name.\n");
					usage(argv[0]);
				}
				break;
			case 'p':
				if ((argc - 1) > i) {
					if (getuid() != 0) {
						printf("Program must be run as root to set priorities.\n");
						exit(0);
					}
					do_nice = 1;
					objects_nice_prio = atoi(argv[i+1]);
					i++;
				} else {
					printf("Switch '-p' needs a priority level.\n");
					usage(argv[0]);
				}
				break;
			case 'd':
				if ((argc - 1) > i) {
					debug_level=atoi(argv[i+1]);
					printf ("\nUsing DEBUG Level %d\n\n", debug_level);
					i++;
				} else {
					printf("Switch '-d' needs a debug level.\n");
					usage(argv[0]);
				}
				break;
			case 'r':
				if ((argc - 1) > i) {
					report_file = argv[i+1];
					printf ("\nReporting hwinfo to %s\n\n", report_file);
					i++;
				} else {
					printf("Switch '-r' needs a report file.\n");
					usage(argv[0]);
				}
				break;
			default:
				usage(argv[0]);
				break;
			}
		} else
			usage(argv[0]);
	}


	/* reset daemons to launch structure */
	memset(daemons, 0, sizeof(struct launch_daemons)*LAUNCH_DAEMONS);

	/* Parse CPU config file */
	if (report_file) {
		report_file_fd = open(report_file, O_CREAT | O_WRONLY | O_APPEND, S_IRUSR | S_IWUSR);
		if (report_file_fd < 0) {
			printf("Error while opening report file %s\n", report_file);
			perror("open");
			exit(-1);
		}		
	}

	/* Parse CPU config file */
	fcpu = fopen(daemons_file, "r");
	if (fcpu == NULL) {
		printf("Error while opening config file %s\n", daemons_file);
		perror("open1");
		exit(-1);
	}

	printf("Parsing cpu config file...");
	fflush(stdout);
	memset(&tmp_db, 0, sizeof (struct hw_api_db));
	n = parse_cpu_cfg(fcpu, daemons, LAUNCH_DAEMONS, &tmp_db);
	fclose(fcpu);
	if (n < 0) {
		printf("\nThere was an error parsing file %s (%d)\n",
                               daemons_file, n);
		exit(-1);
	} else {
		nof_daemons = n;
		printf("ok %d daemons\n", n);
	}


	if (tmp_db.virtual_plt) {
		sprintf (filelockname, "%s/phal.%d.lock", LOCK_PATH, getpid ());
	} else {
		sprintf(filelockname, "%s/%s", LOCK_PATH, LOCK_FILE);
	}

	exit_level++;

	fd = open(filelockname, O_WRONLY | O_CREAT | O_EXCL);
	if (fd < 0) {
		if (errno == EEXIST) {
			/* remove it if exists */
			remove(filelockname);

			/* and create new */
			fd = open(filelockname, O_WRONLY | O_CREAT | O_EXCL);
			if (fd < 0) {
				perror("open");
				exit(-1);
			}
		} else {
			printf("Error opening %s\n", filelockname);
			perror("open_lock");
			exit(-1);
		}
	}

	printf("HWAPI_BACKD: Created lock file %s (%d)\n", filelockname, fd);
	sprintf(loc_str, "HWAPI_BACKD shared resources files\n");
	write(fd, loc_str, strlen(loc_str));
	close(fd);

	/*Create shared memory control area */
	shm_key = ftok(filelockname, HW_API_SHM_KEY);
	api_msg_key = ftok(filelockname, HW_API_MSG_KEY);
	data_msg_key = ftok(filelockname, DATA_MSG_KEY);
	exit_level++;

	clear_shm();

	create_shm();

	init_shm_db();

	hw_api_db->virtual_plt = tmp_db.virtual_plt;
	memcpy(&hw_api_db->cpu_info, &tmp_db.cpu_info, sizeof (struct hwapi_cpu_i));
	memcpy(&hw_api_db->daemons, daemons, nof_daemons*sizeof(struct launch_daemons));
	strncpy(hw_api_db->res_path, res_path, 128);
	hw_api_db->cpu_info.debug_level = debug_level;
		
	exit_level++;

	net_db = &hw_api_db->net_db;


	/* Parse interface config file */
	if (xitf_file) {
		fit = fopen (xitf_file, "r");
		if (!fit) {
			printf ("Error opening itf file %s\n", xitf_file);
			exit(-1);
		}
		printf("Parsing interface config file...");
		fflush(stdout);
		n = parse_itf_cfg(fit, net_db->ext_itf, MAX_EXT_ITF);
		fclose(fit);
		if (n < 0) {
			printf("\nThere was an error parsing file %s (%d)\n",
					xitf_file, n);
			exit(-1);
		} else {
			printf("ok %d itf\n", n);
			net_db->nof_ext_itf = n;
		}
	}

	/* Check CPU INFO */
	if (!hw_api_db->cpu_info.C) {
		xprintf
		("HWAPI_BACKD: CAUTION no CPU info was entered in daemons cfg. Setting default %.f\n",
				DEFAULT_CPU_MIPS);
		hw_api_db->cpu_info.C = DEFAULT_CPU_MIPS;
	} else {
		xprintf ("HWAPI_BACKD: Local Computing Capacity: %.2f MIPS\n",
				hw_api_db->cpu_info.C);
	}

	/* Check TSLOT */
	if (!hw_api_db->cpu_info.tslen_usec) {
		xprintf
		("HWAPI_BACKD: Error can't run with tslot 0\n");
		exit(-1);
	} else {
		xprintf ("HWAPI_BACKD: Time slot duration set to %d usec\n",
				hw_api_db->cpu_info.tslen_usec);
	}

	hw_api_db->cpu_info.plat_family = PLATFORM_LNX26;

	printf("\n\n");
	print_frame("P-HAL for LINUX");
	printf("\n");

	/* Init network thread */
	net_db->net_chld = 0;
	if (net_db->nof_ext_itf) {
		Net_thread(net_db, data_msg_id);
	}
	exit_level++;

	/* wait to be created */
	sleep_ms(500);

	/*Capture SIGCHLD signals */
	action.sa_sigaction = sig_chld;
	action.sa_flags = SA_SIGINFO;
	sigemptyset(&action.sa_mask);
	if (sigaction(SIGCHLD, &action, NULL) < 0) {
		perror("sigaction");
		exit(-1);
	}

	/* capture SIGUSR1 signal */
	action.sa_sigaction = sig_usr;
	action.sa_flags = SA_SIGINFO;
	sigemptyset(&action.sa_mask);
	if (sigaction(SIGUSR1, &action, NULL) < 0) {
		perror("sigaction");
		exit(-1);
	}

	/* capture SIGINT signal */
	action.sa_sigaction = sig_int;
	action.sa_flags = SA_SIGINFO;
	sigemptyset(&action.sa_mask);
	if (sigaction(SIGINT, &action, NULL) < 0) {
		perror("sigaction");
		exit(-1);
	}
	/* start daemons at boot */
	launch_daemons();
	exit_level++;

	/** fall to IDLE
	 */
	while (1) {
		sleep_ms(100);
		fflush(stdout);
		if (report_file_fd>0) {
			report_hwinfo(report_file_fd);
		}
		//    clear_zombies ();
	}

	/* type ctrl+c to exit */
	return (1);
}


/** Died child signal capture.
 *
 * Captures signals for SIGCHLD (a child dies).
 */
void sig_chld(int sig, siginfo_t * sinfl, void *v)
{
	int i;
	time_t t;
	int status;

	//printf("HWAPI_BACKD: Received a SIGCHLD from pid %d.\n", sinfl->si_pid);

	/* if it is the main network child, this is a fatal error, exit program */
	if (net_db->net_chld > 0) {
		if (waitpid(net_db->net_chld, NULL, WNOHANG) > 0) {
			printf("HWAPI_BACKD: Fatal error. Main network child died. Should exit...\n");
			net_db->net_chld = 0;
		}
	}

	/* if it is daemon, it is also a fatal error, exit program*/
	for (i = 0; i < nof_daemons; i++) {
		if (daemons[i].pid) {
			if (waitpid(daemons[i].pid, NULL, WNOHANG) > 0) {
				printf(
						"HWAPI_BACKD: Fatal error. Daemon %s died. Should exit...\n",
						daemons[i].path);
				daemons[i].pid = 0;
			}
		}
	}

	/* if it is an object, set pid to 0. Here as we don't want to exit,
	 * we will check all childs
	 */
	for (i = 0; i < MAX_PROCS; i++) {
		if (hw_api_db->proc_db[i].pid) {
			if (waitpid(hw_api_db->proc_db[i].pid, NULL, WNOHANG)
					> 0) {
				get_time(&t);
				printf("HWAPI_BACKD: Object %s (Id: 0x%x) died at %d:%d\n",
						hw_api_db->proc_db[i].obj_name,
						hw_api_db->proc_db[i].obj_id,
						t.tv_sec,t.tv_usec);

				remove_file(i);

				/** @todo remove also interfaces and variables when object dies
				 */
				//hw_api_db->proc_db[i].obj_id = 0;
				hw_api_db->proc_db[i].pid = 0;
			}
		}
	}

	/* finnally, check all network childs */
	for (i = 0; i < net_db->nof_ext_itf; i++) {
		if (net_db->ext_itf[i].pid > 0) {
			if (waitpid(net_db->ext_itf[i].pid, NULL, WNOHANG) > 0) {
				printf(
						"HWAPI_BACKD: Network child pid %d has died.\n",
						net_db->ext_itf[i].pid);

				net_db->ext_itf[i].pid = -1;
				if (net_db->ext_itf[i].mode & FLOW_WRITE_ONLY) {
					net_db->ext_itf[i].out_mbox = 0;
				}
			}
		}
		if (net_db->ext_itf[i].bpid > 0) {
			if (waitpid(net_db->ext_itf[i].bpid, NULL, WNOHANG) > 0) {
				printf(
						"HWAPI_BACKD: Network bidirectional child pid %d has died.\n",
						sinfl->si_pid);

				net_db->ext_itf[i].bpid = 0;
				net_db->ext_itf[i].out_mbox = 0;
			}
		}
	}
}

void sig_int(int sig, siginfo_t * sinfl, void *v)
{
	exit_level = 10;
	exit_runph();
	exit_level = -1;
	exit(0);
}

/** Message signal capture (command from a daemon).
 *
 */
void sig_usr(int sig, siginfo_t * sinfl, void *v)
{
	int n;
	int ret;

	/* receive message */
	n = msgrcv(api_msg_id, &msg, MSG_SZ, MSG_MAGIC, IPC_NOWAIT);
	if (n > 0) {
		/* call the answer command function */
		ret = Answer_command(msg.body[0], (int *) &msg.body[1], ((n
				- HEAD_SZ) >> 2) - 1);
	} else if (n == -1) {
		if (errno != ENOMSG) {
			perror("msgrcv");
			exit(-1);
		} else {
			return;
		}
	}

	/* answer with return value */
	msg.head.dst = msg.head.src;
	msg.body[0] = ret;
	msgsnd(api_msg_id, &msg, HEAD_SZ + 1, 0);
}

/** Command Processing.
 *
 * @todo Relative path when launching received processes
 *
 * Accepts the following commands with the following contents. First word is the
 * command type, following is data needed by this command:
 *	- HWD_SETUP_XITF: Starts an external interface (configured at interface configuration file).
 *		- word 1: Id of the external interface.
 *		- word 2: Mailbox of the internal interface to be bridged.
 *		- word 3: Mode (READ_ONLY/WRITE_ONLY).
 *	- CREATE_PROCESS: Creates a process.
 *		- word 0: Position in the process database of the process to be
 *			created.
 */
int Answer_command(int cmd, int *packet, int len)
{
	int i;
	int pid;


	switch (cmd) {
	case HWD_SETUP_XITF:

		i = 0;
		while (i < net_db->nof_ext_itf && net_db->ext_itf[i].id
				!= packet[0]) {
			i++;
		}

		if (i == net_db->nof_ext_itf) {
			printf("HWAPI_BACKD: Itf id 0x%x does not exist.\n",
					packet[0]);
			return -1;
		}
		if (packet[2] == FLOW_READ_ONLY) {
			net_db->ext_itf[i].in_mbox = packet[1];
		} else if (packet[2] == FLOW_WRITE_ONLY) {
			net_db->ext_itf[i].out_mbox = packet[1];
		} else {
			printf("HWAPI_BACKD: Invalid mode 0x%x while setting itf\n",
					packet[2]);
			return -1;
		}

#ifdef DEB_XITF
		printf("HWAPI_BACKD: Itf id 0x%x configured with mbox 0x%x. Pos %d mode %d\n",
				packet[0], packet[1], i, packet[2]);
#endif
		return 1;
		break;

	case HWD_CREATE_PROCESS:

		/* process index is received in the packet */
		i = packet[0];

		//      printf ("creating proc id %d\n", hw_api_db->proc_db[i].obj_id);

		/* create child */
		switch (pid = fork()) {
		case -1:
			perror("fork");
			return -1;
		case 0:

			/* set pid, this indicates a successfull creation of the child */
			hw_api_db->proc_db[i].pid = getpid();

			//        xprintf ("launching %s\n", hw_api_db->proc_db[i].path);

			execl(hw_api_db->proc_db[i].path,
					hw_api_db->proc_db[i].path, "-m", "1", NULL);
			printf(
					"HWAPI_BACKD: Error while launching process %s\n",
					hw_api_db->proc_db[i].path);
			exit(-1);
		default:
			if (do_nice) {
				if (setpriority(PRIO_PROCESS, pid, objects_nice_prio)) {
					printf ("HWAPI_BACKD: Error setting process priority\n");
					perror("setpriority");
				}
				
			}

			/* parent returns child pid */
			return pid;
		}

		break;
	default:
		printf("HWAPI_BACKD: Error unknown command 0x%x\n", cmd);
		break;
	}

	printf("HWAPI_BACKD: Error answering command... Not suposed to arrive here.\n");
	return -1;
}

/** Start Daemons
 *
 * Starts all daemons (parsed from the daemon configuration file and saved in daemons structure)
 * by forking and exec()
 */
void launch_daemons()
{
	int i;
	int pid;

	for (i = 0; i < nof_daemons; i++) {
		/* create child */
		switch (pid = fork()) {
		case -1:
			perror("fork");
			exit(-1);
		case 0:
			/* launch process */
			execlp(daemons[i].path, daemons[i].path, NULL);
			printf("\nHWAPI_BACKD: Error launching %s\n\n",
					daemons[i].path);
			perror("execlp");
			daemons[i].pid = 0;
			exit(-1);
		default:
			/* save pid */
			daemons[i].pid = pid;
		}
	}
}

/** Close Daemons
 *
 * Terminates all daemons and proccesses by sending a SIGTERM
 */
void kill_chlds(int wait_time)
{
	int i, c;
	int sleep_time;
	int error = 0;

	if (!wait_time) {
		wait_time = 500;
	}

	sleep_time = wait_time / 10;

	/* kill all phal processes */
	for (i = 0; i < MAX_PROCS; i++) {
		if (hw_api_db->proc_db[i].pid > 0) {
			printf(
					"HWAPI_BACKD: Terminating object %s (Id 0x%x) with PID=%d...\n",
					hw_api_db->proc_db[i].obj_name,
					hw_api_db->proc_db[i].obj_id,
					hw_api_db->proc_db[i].pid);

			kill(hw_api_db->proc_db[i].pid, SIGINT);

			c = 0;
			while (hw_api_db->proc_db[i].pid > 0) {
				sleep_ms(sleep_time);
				c++;
				if (c == 10) {
					printf(
							"HWAPI_BACKD: Object %s still alive, sending kill...\n",
							hw_api_db->proc_db[i].obj_name);
					kill(hw_api_db->proc_db[i].pid, SIGKILL);
					error++;
					break;
				}
			}
		}
	}
	/* and daemons */
	for (i = 0; i < nof_daemons; i++) {
		if (daemons[i].pid > 0) {
			printf(
					"HWAPI_BACKD: Terminating daemon %s with pid %d...\n",
					daemons[i].path, daemons[i].pid);

			kill(daemons[i].pid, SIGKILL);

			c = 0;
			while (daemons[i].pid > 0) {
				sleep_ms(sleep_time);
				c++;
				if (c == 10) {
					printf(
							"HWAPI_BACKD: Daemon %s (%d) still alive, sending kill...\n",
							daemons[i].path,
							daemons[i].pid);
					kill(daemons[i].pid, SIGKILL);
					error++;
					break;
				}
			}
		}
	}

	if (error) {
		printf(
				"HWAPI_BACKD: Some daemons could not be closed properly (%d).\n",
				error);
	}

}

void remove_file(int i)
{
	/* ensure any other process is using the same executable */

#ifdef REMOVE_FILE
	int j;

	for (j = 0; j < MAX_PROCS; j++) {
		if (j != i) {
			if (!strncmp(hw_api_db->proc_db[i].path,
					hw_api_db->proc_db[j].path, 128)) {
				break;
			}
		}
	}
	if (j == MAX_PROCS) {
		//printf("HWAPI_BACKD: Removing file %s\n ",
		//		hw_api_db->proc_db[i].path);

		if (remove(hw_api_db->proc_db[i].path)) {
			perror("remove");
		}

	}
#endif
	hw_api_db->proc_db[i].path[0] = '\0';
}

void clear_files()
{
	int i;
	for (i=0; i<MAX_PROCS; i++) {
		if (hw_api_db->proc_db[i].path[0] != '\0') {
			remove_file(i);
		}
	}
}

/** Clear all zombies (my childs)
 */
void clear_zombies()
{
	while (waitpid(-1, NULL, WNOHANG)>0)
		;
	/* ensure there are any other child */
	while (wait(NULL)>0)
		;
}

/* print frame function */
void print_frame(char *title)
{
	printf("======================================\n");
	printf("     -= %s =-               \n", title);
	printf("======================================\n");

}

/** @depreciated erase ?? */
/*int
 get_itf_r_pid (int id)
 {
 int i;

 i = 0;
 while (i < MAX_INT_ITF && hw_api_db->int_itf[i].id != id)
 i++;
 if (i < MAX_INT_ITF)
 return hw_api_db->int_itf[i].r_pid;
 else
 return 0;
 }
 */
/** @} */
