/**
 * @file phal_hw_api.c
 * @brief PHAL Hardware Library LINUX Implementation
 *
 * This file contains the implementation of the HW Library for
 * linux platforms.
 */

/** default includes
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/msg.h>
#include <sys/ipc.h>
#include <sys/shm.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <sys/wait.h>
#include <sys/time.h>
#include <time.h>
#include <errno.h>
#include <unistd.h>
#include <fcntl.h>

#include "mem.h"
#include "phal_hw_api.h"
#include "hwapi_backd.h"

//#define PRINT_OBJ_TIME
//#define DEBUG_ITF

/* fraction of the time slot sw daemons must sleep */
#define DAEMON_REL_FRACTION	3

/* general string buffers */
char aux_string[128];

/** hw_api general database */
struct hw_api_db *hw_api_db;

/** pointer to proc db */
struct hwapi_proc_i *proc_db;

/** save my proc pointer */
struct hwapi_proc_i *my_proc;

/** shared vars db */
struct hwapi_var_i *var_db;

/** time slot duration pointer*/
int *ts_len;

/** pointer to internal interfaces db */
struct int_itf *int_itf;

/** init interface id counter */
int last_id = 0;

/** shared memory ids */
int api_shm_id, api_msg_id, data_msg_id;

/** structs for messages */
struct msgbuff msg;

/** internal itf database definition */
struct file_des
{
	int int_itf_idx;
	int id;
	int mode;
	int pending_data;
	int pending_rpm;
	int *pending_pkt;
};

/** number of internal interfaces (for this process) */
#define MAX_FILE_DES	20
/** database for local itf */
struct file_des file_des[MAX_FILE_DES];

/************************************************
 INTERNAL FUNCTION DECLARATIONS
 *************************************************/
int send_parent_cmd(int cmd, int *data, int size);
void get_time(time_t * t);


/** @defgroup hwapifunc Hardware API Functions Implementation
 *
 * Here HW API functions are implemented for Linux Platforms
 *
 * @{
 */

/** @defgroup base Base API Functions
 *
 * This functions include some common and basic
 * functions required to use the HW API Library
 *
 * @{
 */

/** Initialize Hardware API.
 *
 * Initializes hwapi library. This function must be called before
 * calling any other hwapi function.
 *
 *
 * @return 1 if successfully registered
 * @return 0 if failed
 *
 */
int hwapi_init(void)
{
	int *ptr;
	key_t key;
	int flags;
	int i, n;
	flags = 0;



	/* try to attach to parent process shared memory */
	sprintf (aux_string, "%s/phal.%d.lock", LOCK_PATH, getppid ());
	key = ftok(aux_string, HW_API_SHM_KEY);
	api_shm_id = shmget(key, sizeof(struct hw_api_db), flags);
	if (api_shm_id < 0) {
		/* if do not exist, try to attach to default lock file (no virtual platform) */
		sprintf(aux_string, "%s/%s", LOCK_PATH, LOCK_FILE);
		key = ftok(aux_string, HW_API_SHM_KEY);
		api_shm_id = shmget(key, sizeof(struct hw_api_db), flags);
		if (api_shm_id < 0) {
			printf("Error shared memory for api doesn't exist. Run console first\n");
			perror("shmget");
			return 0;
		}
	}
	ptr = (int *) shmat(api_shm_id, NULL, 0);
	if (ptr == (int *) -1) {
		perror("shmat");
		return 0;
	} else {
		hw_api_db = (struct hw_api_db *) ptr;
		proc_db = hw_api_db->proc_db;
		int_itf = hw_api_db->int_itf;
		var_db = hw_api_db->var_db;
		ts_len = &hw_api_db->cpu_info.tslen_usec;
	}

	/* get lock key for control messages */
	key = ftok(aux_string, HW_API_MSG_KEY);

	/* get id for control messages */
	api_msg_id = msgget(key, flags);
	if (api_msg_id < 0) {
		printf("Error initiating message queue. Console not initiated\n");
		perror("msgget");
		return 0;
	}

	/* get lock key for data messages */
	key = ftok(aux_string, DATA_MSG_KEY);

	/* get id for data messages */
	data_msg_id = msgget(key, flags);
	if (data_msg_id < 0) {
		printf("Error initiating message queue. Console not initiated\n");
		perror("msgget");
		return 0;
	}

	/* initialize internal interface database */
	memset(file_des, 0, MAX_FILE_DES * sizeof (struct file_des));

	/* check if I am an object process or a daemon */
	i = 0;
	n = getpid();
	while (i < MAX_PROCS && proc_db[i].pid != n) {
		i++;
	}
	if (i < MAX_PROCS) {
		my_proc = &proc_db[i];
		/* set to relinquished */
		my_proc->relinquished = 1;
	}

	else {
		my_proc = 0;
	}


	/* redirect output if requested */
	i=0;
	while(i==0 && hw_api_db->daemons[i].pid != n) {
		i++;
	}
	if (i==MAX_DAEMONS) {
		xprintf("CAUTION! did not find daemon in daemon's db\n");
		return 1;
	}
	
	if (strlen(hw_api_db->daemons[i].output) > 0) {
		int fd = open(hw_api_db->daemons[i].output, O_CREAT | O_WRONLY | O_APPEND);
		if (fd < 0) {
			xprintf ("Could not open output file %s\n", hw_api_db->daemons[i].output);
			perror("open");
		} else {
			printf("Output is being redirected to %s for daemon %s...\n", hw_api_db->daemons[i].output, hw_api_db->daemons[i].path);
			close(1);
			dup(fd);
		}
	}
//	printf("hola\n");
//	fflush(stdout);

	//  xprintf ("pid %d hwapi %p table %p (%d)\n", getpid (), hw_api_db, hw_api_db->var_table, hw_api_db->var_table[0]);
	return 1;
}

/** Exit function.
 *
 * Clears hwapi resources, no more hwapi functions can be called after
 * this functions is called.
 *
 * The function does NOT kill process thread, just clears shared memory resources
 *
 */
void hwapi_exit()
{

	/* clear shared memory resources */
	shmdt(hw_api_db);

	/* to do... clear messages resources */
	/* more stuff?? */
}

/** Check Daemon Presence Function:
 *
 * Checks if daemon with name 'daemon_name' have been registered in the same
 * HWAPI Library
 *
 * @param daemon_name Name of the daemon to check
 *
 * @return 1 dameon is present and registered
 * @return 0 daemon is present but not registered
 * @return -1 daemon is not present
 */

/* to erase ?
 int hwapi_is_daemon(char *daemon_name) {
 int i;

 i=0;
 while(i<MAX_DAEMONS && strcmp(daemon_db[i].sname,daemon_name))
 i++;
 if (i==MAX_DAEMONS)
 return -1;
 else
 return daemon_db[i].status;
 }
 */

/** Get my object id
 */
int hwapi_myobjid(void)
{
	if (my_proc) {
		return my_proc->obj_id;
	}

	else {
		return 0;
	}
}

int compute_mean(int *x, int len, int input) {
	int i;
	
	int output=0;	
	for (i=0;i<len;i++) {
		output += x[i]/len;
	}
	x[0] = input;
	for (i=len-1;i>0;i--) {
		x[i] = x[i-1];
	}

	return output;	
}

/** Process Relinquish.
 *
 * Function used by processes to relinquish their cpu ownership.
 * This function must be called by SERVICES API before the end of the
 * time slot, otherwise RTCFault will be flagged.
 *
 * @param slots number of slots to relinquish
 * @return >=0 number of processes
 */
void hwapi_relinquish(int slots)
{
	unsigned int actual_time, tstamp;
	int sleep_time;
	time_t tdata;
	int wake_time;
	int tslot_time;
	time_t wake, tti;

	/* check params */
	if (!slots || slots < 0)
	{
		slots = 1;
	}

	/* check my proc is correct */
	if (!my_proc) {
		printf("HWAPI: Fatal Error!!\n");
		exit(-1);
	}

	/* get actual time stamp */
	tstamp = get_tstamp();

	/* get actual time */
	get_time(&tdata);

	/* express time in microseconds */
	actual_time = (unsigned int) tdata.tv_sec * 1000000 + tdata.tv_usec;

	/* calc sleep_time */
	sleep_time = (tstamp + slots) * (*ts_len) - actual_time;
	
	/* get time we finished execution */
	memcpy(&my_proc->cpu_time[2], &tdata, sizeof(time_t));

	/* and cpu utilization (time we started (2) - time we entered (1) */
	get_time_interval(my_proc->cpu_time);
	
	/* save mean cpu */
	my_proc->mean_cpu_usec = compute_mean(my_proc->_mean_cpu_usec, MEAN_CPU_LENGTH, my_proc->cpu_time[0].tv_usec);
	
	my_proc->rel_tstamp = tstamp;
	my_proc->relinquished = 1;

	/* go sleep */
	if (sleep_time > ((*ts_len)/TS_ERROR_FRAC)) {
		sleep_us(sleep_time);
	}

	my_proc->relinquished = 0;

	/* here we wake */
	get_time(&wake);
	timersub (&wake, &my_proc->cpu_time[1], &tti);
	my_proc->tti = tti.tv_usec;

	memcpy(&my_proc->cpu_time[1], &wake, sizeof(time_t));

#ifdef PRINT_OBJ_TIME
	printf
	("OBJECT %s:\tCPU %d usec. CPU/TTI %.2f%%. Sleep %d usec. TTI %d usec (error %d usec).\n",
			my_proc->obj_name, my_proc->cpu_time[0].tv_usec,
			(float) my_proc->cpu_time[0].tv_usec/sleep_time,
			sleep_time, tti.tv_usec, abs (tti.tv_usec - (*ts_len)));
#endif

	fflush(stdout);
}

/** Daemon Relinquish.
 *
 * Same function as proc_relinquish but used by SW daemons.
 * Here time slotting is not required (in Linux), so processes are
 * sleeped a fraction of the time slot
 *
 */
void hwapi_relinquish_daemon()
{

	sleep_us(*ts_len / DAEMON_REL_FRACTION);
	fflush(stdout);
}

/** @} */

/** @defgroup itf Packet Interface Management Functions
 *
 * HW API provides the user to a set of functions to create FIFO-like packet
 * oriented data interface.
 *
 * This interfaces have a read and a write side. Once it is created, the user
 * must attach a file descriptor to the read or write side, thus, communication is possible
 * and packets written in the write side will be available in the read side.
 *
 * A name and an owner Id (object Id) can be assigned to any of the sides so separate
 * processes can share an interface easier.
 *
 * @{
 */

/** Create Interface.
 *
 * Creates a packet-oriented internal data interface.
 * If id parameter is different of 0 it will be the
 * assigned id for the interface. Otherwise it will
 * be generated.
 *
 * @param id 0 for automatic, else for fixed
 * @param size maximum packet size allowed in words
 *
 * @return !=0 Id of the interface
 * @return =0 error creating interface
 */
int hwapi_itf_create(int id, int size)
{
	int i;

	/* check if id already used */
	if (id) {
		i = 0;
		while (i < MAX_INT_ITF && int_itf[i].id != id) {
			i++;
		}
		if (i < MAX_INT_ITF) {
			xprintf ("HWAPI: Error Id=0x%x already used\n", id);
			return 0;
		}
	} else {
		/* search an used one */
		do {
			id = (((rand()) & 0xf) << 12) | (last_id & 0xfff);
			last_id = (last_id + 1) & 0xfff;
			i = 0;
			while (i < MAX_INT_ITF && int_itf[i].id != id) {
				i++;
			}
		} while (i < MAX_INT_ITF);
	}

	/* search for a free position in db */
	i = 0;
	while (i < MAX_INT_ITF + 1 && int_itf[i].id) {
		i++;
	}
	if (i == MAX_INT_ITF) {
		xprintf ("HWAPI: Error no more itf allowed (%d)\n", i);
		return 0;
	}

	/* create interface */
	int_itf[i].id = id;

	//  xprintf ("HWAPI: Created interface 0x%x\n", id);

	/* return id */
	return id;
}

/** Delete Interface.
 *
 * Deletes internal data interface with id='id' (and clears resources)
 *
 * @param id Id of the interface to delete
 *
 * @return 1  Deleted successfully
 * @return 0 Error deleting interface
 */
int hwapi_itf_delete(int id)
{
	int i;

	i = 0;
	while (i < MAX_INT_ITF && int_itf[i].id != id) {
		i++;
	}
	if (i == MAX_INT_ITF) {
		xprintf ("HWAPI: Error can't find itf id=0x%x.\n", id);
		return 0;
	}

	//xprintf ("HWAPI: Deleted itf 0x%x\n", int_itf[i].id);

	/* clear interface resources */
	memset(&int_itf[i], 0, sizeof(struct int_itf));

	return 1;
}

/** Delete all own interfaces.
 *
 * Given an object name, deletes all interfaces which it has been
 * matched in write mode
 *
 * @param name Name of the object
 *
 * @return 1 Deleted successfully
 * @return 0 Error deleting interface
 */
int hwapi_itf_delete_own(char *obj_name)
{
	int i;

	for (i = 0; i < MAX_INT_ITF; i++) {
		if (!strncmp(int_itf[i].w_obj, obj_name, 32)) {
			hwapi_itf_delete(int_itf[i].id);
		}
	}

	return 1;
}

/** Attach to an Interface.
 *
 * Creates a file descriptor to use the interface. This function
 * should be called 'after' creating the interface and 'before' using it.
 *
 * It must be attached in FLOW_READ_ONLY (for reading) or FLOW_WRITE_ONLY
 * mode (for writting). A file descriptor will be returned which can be used
 * to read/write packets
 *
 * @param id Id of the interface created with hwapi_create_itf()
 * @param mode FLOW_READ_ONLY or FLOW_WRITE_ONLY constants
 *
 * @return >0 File descriptor
 * @return =-2 Interface not available
 * @return =-1 Error attaching interface
 */
int hwapi_itf_attach(int id, int mode)
{
	int i, j;

	/* first of all, search itf in interfaces database */
	i = 0;
	while (i < MAX_INT_ITF && int_itf[i].id != id) {
		i++;
	}
	if (i == MAX_INT_ITF) {
		return -2;
	}

	/* then search a free position in file descriptors database */
	j = 3; /* first 2 are reserved */
	while (j < MAX_FILE_DES && file_des[j].mode) {
		j++;
	}
	if (j == MAX_FILE_DES) {
		xprintf ("HWAPI: Error no more local itf allowed (%d).\n", j);
		return -1;
	}

	/* now fill local db */
	file_des[j].id = int_itf[i].id;
	file_des[j].mode = mode;
	file_des[j].int_itf_idx = i;

	return j;
}

/** Unattach from an Interface.
 *
 * Un-attaches a file descriptor from the (previously attached)
 * interface in mode 'mode'
 *
 * @param fd file descriptor of the interface
 * @param mode FLOW_READ_ONLY or FLOW_WRITE_ONLY constants
 *
 * @return >0 Successfully unattached
 * @return <0 Error unattaching
 */
int hwapi_itf_unattach(int fd, int mode)
{

	/* check correct fd */
	if (fd < 0 || fd > MAX_FILE_DES) {
		xprintf ("HWAPI: Error invalid local fd %d.\n", fd);
		return -1;
	}

	/* clear resources */
	memset(&file_des[fd], 0, sizeof(struct file_des));

	return 1;
}

/** Send a packet.
 *
 * Sends a packet of data through a data interface. The user needs
 * to provide a file descriptor, get after calling the attaching function.
 *
 * @param fd file descriptor
 * @param buffer pointer to data to send
 * @param size size of the packet in words (32-bit)
 *
 * @return >0 Successfully sended. Returns amount of words sended.
 * @return =0 Packet could not be send (full queue)
 * @return <0 Error sending packet
 */
int hwapi_itf_snd(int fd, int *buffer, int size)
{
	int n;
	int tstamp;
	int id;
	int delay = 0;
	int urgent = 0;
	int do_signal = 0;
	int flags;
	int external = 0;
	int int_pos;
	struct msqid_ds msgstatus;
	assert (buffer && size >= 0);

	/* check for valid fd */
	if (fd < 0 || fd > MAX_FILE_DES) {
		printf("HWAPI: Error invalid fd=%d. 0<=fd<=%d.\n", fd,
		MAX_FILE_DES);
		return -1;
	}

	/* check if fd is write enabled */
	if (!(file_des[fd].mode & FLOW_WRITE_ONLY)) {
		xprintf
		("HWAPI: Error flow is not write enable (fd=%d id=0x%x mode=%d pid=%d)\n",
				fd, file_des[fd].id, file_des[fd].mode,getpid());
		return -1;
	}

	/* get interface id */
	id = file_des[fd].id;

	/* get interface position */
	int_pos = file_des[fd].int_itf_idx;

	/* get interface delay */
	/** @todo itf delay */
	delay = 0;

	/* check urgent flag */
	/** @todo urgent interfaces */
	/* check external flag */
	/** @todo external interfaces */

	/* check if size exceeds internal limit */
	if (size > MSG_BUFF_SZ) {
		printf("HWAPI: Error packet too long (%d/%d)\n", size,
		MSG_BUFF_SZ);
		return -1;
	}
	
	if (my_proc)
		delay=1;

	/* calculate packet time stamp and destination */
	if (delay) {
		/* write future tstamp. Packet is only available on next tstamp */
		tstamp = get_tstamp();
		msg.head.dst = (id << TSTAMP_PKT_BITS) | ((tstamp + delay)
				& TSTAMP_PKT_MASK);
		flags = 0;
	} else {
		msg.head.dst = (id << TSTAMP_PKT_BITS);
		flags = 0;
	}

	/* fill message header */
	msg.head.src = 0;
	msg.head.opts = int_itf[int_pos].opts;
	msg.head.tstamp = msg.head.dst & TSTAMP_PKT_MASK;

	/* copy packet body */
	memcpy(msg.body, buffer, size << 2);

/*	if (my_proc) {
		msgctl(data_msg_id, IPC_STAT, &msgstatus);
		if (msgstatus.__msg_cbytes > msgstatus.msg_qbytes/10) {
			if ((get_tstamp() % 100)==0)
				printf ("CAUTION WITH FIFOS: %d/%d\n", msgstatus.__msg_cbytes, msgstatus.msg_qbytes);
		}		
	}
*/	
	
	flags = IPC_NOWAIT;
	n = msgsnd(data_msg_id, &msg, (size << 2) + HEAD_SZ, flags);
	if (n == -1 && errno == EAGAIN) {
		/* full queue, print only for objects interfaces */
		if (my_proc) {
			msgctl(data_msg_id, IPC_STAT, &msgstatus);
			time_t t;
			get_time(&t);
			printf("\nCAUTION: Full msg queue at %d:%d (ts=%d), caller object %s. Current status: %d/%d bytes, %d msg\n\n",
				t.tv_sec, t.tv_usec, get_tstamp(), 
				my_proc->obj_name,msgstatus.msg_cbytes, msgstatus.msg_qbytes,
					msgstatus.msg_qnum);
		}
		return 0;
	} else if (n == -1) {

		/* error */
		perror("msgsnd");
		return -1;
	}
	/* success */

#ifdef DEBUG_ITF
	if (my_proc) {
		xprintf ("%d send packet to 0x%x\n", getpid(), id);
	}
#endif
	return size;
}

int hwapi_itf_status(int fd)
{
	struct msqid_ds info;
	int id;

	/* check valid file descriptor */
	if (fd < 0 || fd > MAX_FILE_DES) {
		printf("HWAPI: Error invalid fd=%d. 0<=fd<=%d.\n", fd,
				MAX_FILE_DES);
		return -1;
	}
	
	if (!msgctl(data_msg_id, IPC_STAT, &info)) {
		return info.__msg_cbytes;
	} else {
		return 0;
	}
}

/** Receive a packet.
 *
 * Recevies a packet through a file descriptor. User must
 * indicate provided buffer's length (in words).
 *
 * @param fd file descriptor
 * @param buffer pointer to buffer where receive data
 * @param size size of the buffer in words (32-bit)
 *
 * @return >0 Successfully received. Returns amount of words.
 * @return =0 No packets available (empty queue)
 * @return <0 Error receiving packet
 */
int hwapi_itf_rcv(int fd, int *buffer, int size)
{
	int n;
	int type;
	int tstamp;
	int c;
	int id;
	int urgent;
	int delay, external;
	int int_pos;

	assert (buffer && size >= 0);

	/* check valid file descriptor */
	if (fd < 0 || fd > MAX_FILE_DES) {
		printf("HWAPI: Error invalid fd=%d. 0<=fd<=%d.\n", fd,
		MAX_FILE_DES);
		return -1;
	}

	/* check if reading enabled */
	if (!(file_des[fd].mode & FLOW_READ_ONLY)) {
		xprintf
				("HWAPI: Error flow is not read enable (fd=%d id=0x%x mode=%d pid=%d)\n",
						fd, file_des[fd].id, file_des[fd].mode,getpid());

		return -1;
	}

	/* get interface id */
	id = file_des[fd].id;

	/* get interface db position */
	int_pos = file_des[fd].int_itf_idx;

	/* obtain current timestamp */
	tstamp = get_tstamp();
	c = 0;

	if (my_proc) 
		delay = 1;

	/* receive current tstamp or tstamp - TSTAMP_MAX_RED */
	do {
		if (!delay) {
			type = id << TSTAMP_PKT_BITS;
		} else {
			type = (id << TSTAMP_PKT_BITS) | ((tstamp - c)
					& TSTAMP_PKT_MASK);
		}

		/* receive */
		n = msgrcv(data_msg_id, &msg, MSG_SZ, type, IPC_NOWAIT);
		c++;
	} while (n == -1 && errno == ENOMSG && c < TSTAMP_MAX_RED - 1 && delay);

	/* any message available */
	if (n == -1 && errno == ENOMSG) {
		return 0;
	}

	/* check error */
	if (n < 0) {
		if (errno == E2BIG) {
			printf(
					"msgsz %d bytes size is %d words, mtext size %d bytes\n",
					(size << 2) + HEAD_SZ, size,
					(MSG_BUFF_SZ << 2));
		}

		/* other error */
		perror("msgrcv");
		return n;
	}

	/* check packet lenght does not exceed user buffer length */
	n = ((n - HEAD_SZ - 1) >> 2) + 1;
	if (n > size)

	{
		n = size;
	}

#ifdef DEBUG_ITF
	if (my_proc)
		xprintf ("%d recv packet from 0x%x\n", getpid(), id);
#endif
	/* copy packet to user buffer */
	memcpy(buffer, msg.body, n << 2);

	return n;
}

/** Link Physical Interface to a logical Interface
 *
 *
 *
 * Links a physical interface (external) to an internal packet
 * oriented interface. After the link is successfully realized,
 * all packets written to the internal interface will be bridged
 * (by RUNPH_NET) towards the physical interface. Conversely, all
 * packets received to the phy. itf. will be bridged to the internal
 * interface.
 *
 * Interface must be created before calling this function. Then, in
 * order to use it, it must be attached.
 *
 * @param itf_id Local interface Id
 * @param phy_itf_id External Physical Interface Id
 * @param mode FLOW_READ_ONLY or FLOW_WRITE_ONLY constant
 *
 * @return >0 Successfully linked
 * @return <0 Error linking
 */
int hwapi_itf_link_phy(int itf_id, int phy_itf_id, int mode)
{
	int n, i;
	int cmd[3];

	/* check correct params */
	if (!itf_id) {
		printf("HWAPI: Error invalid parameter itf_id must be != 0\n");
		return -1;
	}
	if (!phy_itf_id) {
		printf("HWAPI: Error invalid parameter xitf_id must be !=0\n");
		return -1;
	}

	/* search internal interface */
	i = 0;
	while (i < MAX_INT_ITF && int_itf[i].id != itf_id)

	{
		i++;
	}
	if (i == MAX_INT_ITF) {
		printf("HWAPI: Can't find itf id 0x%x.\n", itf_id);
		return -1;
	}

	/* send message to RUNPH process */
	cmd[0] = phy_itf_id;
	cmd[1] = itf_id;
	cmd[2] = mode;
	n = send_parent_cmd(HWD_SETUP_XITF, cmd, 3);
	if (n > 0) {
		/* if success, set interface options to external */
		/** @todo Set option to external */
		/*	int_itf[i].opts |= (ITF_OPTS_EXT & ITF_OPT_TYPE_MASK);
		 */
	}

	else {
		/* error */
		printf("HWAPI: Error while setting xitf.\n");
	}
	return n;
}

/** Match Interface.
 *
 * Assign an object id and interface name for the read and write side
 *
 * @return 1 if ok 0 if error
 */
int hwapi_itf_match(int itf_id, char *w_obj, char *w_itf, char *r_obj,
	char *r_itf)
{
	assert (w_itf && w_obj && r_itf && r_obj);
	int i;

	i = 0;
	while (i < MAX_INT_ITF && int_itf[i].id != itf_id) {
		i++;
	}
	if (i == MAX_INT_ITF) {
		xprintf ("HWAPI: Interface 0x%x not found\n", itf_id);
		return 0;
	}
	strncpy(int_itf[i].w_obj, w_obj, 32);
	strncpy(int_itf[i].w_itf, w_itf, 32);
	strncpy(int_itf[i].r_obj, r_obj, 32);
	strncpy(int_itf[i].r_itf, r_itf, 32);
	return 1;
}

/** Get itf id given an itf name and object name.
 * If
 *
 */
int hwapi_itf_find(char *obj_name, char *itf_name, int mode)
{
	int i;

	/* first of all, search itf in interfaces database */
	i = MAX_INT_ITF;
	if (mode & FLOW_WRITE_ONLY) {
		i = 0;
		while (i < MAX_INT_ITF && (strncmp(itf_name, int_itf[i].w_itf,
				32) || strncmp(obj_name, int_itf[i].w_obj, 32)))

		{
			i++;
		}
	}
	if ((mode & FLOW_READ_ONLY) && (i == MAX_INT_ITF)) {
		i = 0;
		while (i < MAX_INT_ITF && (strncmp(itf_name, int_itf[i].r_itf,
				32) || strncmp(obj_name, int_itf[i].r_obj, 32)))

		{
			i++;
		}
	}
	if (i == MAX_INT_ITF) {
		 xprintf ("HWAPI: notfound itf %d %s:%s r/w\n", mode, itf_name, obj_name);
		return 0;
	}
	 /* xprintf ("found itf 0x%x %d %s:%s r %s:%s search %s:%s\n", int_itf[i].id, mode,
	 int_itf[i].w_itf, int_itf[i].w_obj,
	 int_itf[i].r_itf, int_itf[i].r_obj,
	 itf_name, obj_name);
	 */
	return int_itf[i].id;
}

/** @} */

/** @defgroup hw Hardware Monitoring Functions
 *
 * A set of functions are available to gather information about
 * underlying hardware characteristics and current status.
 *
 * @{
 */

/** Physical Interface Information.
 *
 * Obtains information of the Physical External interfaces. It is read from
 * the physical interfaces configuration file (RUNPH Launcher).
 *
 * User must provide the maximum amount of external interfaces he wants to read.
 * It depends on the length of the xitf buffer (number of elements)
 *
 * @param xitf Pointer to information structure (check phal_hw_api.h)
 * @param max_itf Maximum number of interfaces (buffer length, in elements)
 *
 * @return >=0 Number of interfaces read
 * @return <0 error
 */
int hwapi_hwinfo_xitf(struct hwapi_xitf_i *xitf, int max_itf)
{
	int i;
	assert (xitf && max_itf >= 0);

	/* avoid overwrite buffer */
	if (max_itf > hw_api_db->net_db.nof_ext_itf) {
		max_itf = hw_api_db->net_db.nof_ext_itf;
	}

	/* copy all interfaces */
	for (i = 0; i < max_itf; i++) {
		memcpy(&xitf[i], &hw_api_db->net_db.ext_itf[i].info,
				sizeof(struct hwapi_xitf_i));
	}
	return max_itf;
}

/** Get Processor Information.
 *
 *
 * @todo Definition of structure, dynamic change? battery? etc.
 *
 *
 * Function used to gather processing resource (CPU) information.
 * Data is writted to the struct pointed by info:
 *
 *
 * @see hwapi_cpu_i
 * @param info CPU information struct
 *
 */
void hwapi_hwinfo_cpu(struct hwapi_cpu_i * info)
{
	assert (info);
	struct msqid_ds msgstatus;
	//  printf ("HWAPI: Local CPU is %.2f MIPS\n", hw_api_db->cpu_info.C);

	/* fill info, copy it from hw_api_db struct */
	memcpy(info, &hw_api_db->cpu_info, sizeof(struct hwapi_cpu_i));
	
	hw_api_db->cpu_info.cpu_usage = 0;
	int i;
	for (i=0;i<MAX_PROCS;i++) {
		if (proc_db[i].obj_id) {
			hw_api_db->cpu_info.cpu_usage += proc_db[i].mean_cpu_usec;
		}
	}
	
	msgctl(data_msg_id, IPC_STAT, &msgstatus);
	hw_api_db->cpu_info.used_fifo =  msgstatus.__msg_cbytes;
	hw_api_db->cpu_info.total_fifo =  msgstatus.msg_qbytes;
}

int hwapi_hwinfo_setid (int pe_id)
{
	hw_api_db->cpu_info.pe_id = pe_id;
}


/** Execute a Hardware-API Command
 * 
 * Possible commands:
 * 	ps: 	list processes
 * 
 * @param cmd Command string to execute
 * @returns 1 ok 0 error
 */
int hwapi_hwcmd (char *cmd) {
	if (!strcmp(cmd, "stats_on")) {
		hw_api_db->cpu_info.disable_stats = 0;
	} else if (!strcmp(cmd, "stats_off")) {
		hw_api_db->cpu_info.disable_stats = 1;
	} else if (!strcmp(cmd, "logs_on")) {
		hw_api_db->cpu_info.disable_logs = 0;
	} else if (!strcmp(cmd, "logs_off")) {
		hw_api_db->cpu_info.disable_logs = 1;
	} else {
		printf("HWAPI: Unknown command\n");
	}
}

/** @} */

/** @defgroup proc Process Management Functions
 *
 * This group of functions provide tools for creating, monitoring and
 * changing the status of processes (objects) running in the local processor.
 *
 * @{
 */

/** Request New Proces Creation.
 *
 *
 * Use this function before creating a new process to request a
 * memory region where load the process.
 *
 * @todo New process creation for absolute addressing computers
 *
 */
int hwapi_proc_req(void)
{
	xprintf ("HWAPI: Error no relocating available\n");
	return 0;
}

/** Create a New Process.
 *
 * This function creates a process and launches it. As the process resides in user
 * memory (pdata buffer), it must be writed to the disk before launching it. This file
 * will be deleted when the process does.
 *
 * @todo (in Linux) path to downloaded software should be either relative or guessed but not a constant
 *
 * @param pdata Pointer to program data
 * @param pinfo Pointer to process information structure
 *
 * @return >0 Pid of the process
 * @return <0 error
 */
int hwapi_proc_create(char *pdata, struct hwapi_proc_launch *pinfo)
{
	int n;
	int i;
	int f;
	int pid;
	int blen, wlen;
	int c;

	assert (pdata && pinfo);

	if (!pinfo->app_name || !pinfo->exe_name || !pinfo->obj_name) {
		printf("HWAPI: Error launching executable. Some parameters are empty\n");
		return -1;
	}

	blen = pinfo->blen;
	wlen = pinfo->wlen;

	i = 0;
	while (i < MAX_PROCS && (proc_db[i].pid || proc_db[i].obj_id)) {
		i++;
	}
	if (i == MAX_PROCS) {
		printf("HWAPI: No more space in process db.\n");
		return -1;
	}

	/* check if file is already created */
	sprintf(aux_string, "%s/%s", RCV_PROC_PATH, pinfo->exe_name);
	f = open(aux_string, O_CREAT | O_WRONLY | O_EXCL, S_IRWXU);
	/* check error */
	if (f < 0 && errno != EEXIST) {
		perror("fopen");
		return -1;
	}

	/* write buffer to temporal file unless file exists */
	if (f > 0) {
		n = write(f, pdata, blen);
		if (n!= blen) {
			perror("write");
			printf("removing file...\n");
			if (remove(aux_string)) {
				perror("remove");
			}
		}
	}

	/* close file */
	close(f);

	memset(&proc_db[i], 0, sizeof(struct hwapi_proc_i));

	proc_db[i].app_id = pinfo->app_id;
	proc_db[i].obj_id = pinfo->obj_id;
	proc_db[i].change_progress = 1;
	proc_db[i].status.cur_status = 0;
	proc_db[i].status.next_status = 1;
	proc_db[i].status.next_tstamp = get_tstamp();
	strncpy(proc_db[i].path, aux_string, 128);
	strncpy(proc_db[i].obj_name, pinfo->obj_name, 24);
	strncpy(proc_db[i].app_name, pinfo->app_name, 24);

	//printf ("saved pid %d id %d pos %d\n", proc_db[i].pid, proc_db[i].obj_id, i);

	//  printf ("sending parent command...\n");

	/* tell background daemon to launch the process */
	pid = send_parent_cmd(HWD_CREATE_PROCESS, &i, 1);

	// printf ("2saved pid %d id %d pos %d\n", proc_db[i].pid, proc_db[i].obj_id, i);

	/* wait creation of the process */
	c = 0;
	while (pid != proc_db[i].pid && c<4 && pid>0) {
		sleep_ms(500);
		c++;
	}

	// printf ("3saved pid %d id %d pos %d\n", proc_db[i].pid, proc_db[i].obj_id, i);


	/* if error, remove file and reset database structure */
	if (pid != proc_db[i].pid) {
		printf("HWAPI: Child did not started, removing it...\n");
		memset(&proc_db[i], 0, sizeof(struct hwapi_proc_i));

		return -1;
	}

	return pid;
}

/** Remove Process Function:
 *
 * Removes a process from the system and de-allocates all resources
 * it was using.
 *
 * @param obj_id Id of the object to remove
 *
 * @return 1 success
 * @return 0 error
 */
int hwapi_proc_remove(int obj_id)
{
	int c;
	int i;

	/* search it in the database */
	i = 0;
	while (i < MAX_PROCS && proc_db[i].obj_id != obj_id) {
		i++;
	}
	if (i == MAX_PROCS) {
		printf("HWAPI: Error removing object with id 0x%x, not found.\n",
				obj_id);
		return 0;
	}

	/* clear db */
	proc_db[i].obj_id = 0;

	if (!proc_db[i].pid) {
		return 0;
	}
	//  xprintf ("removing id 0x%c, pid %d\n", obj_id, proc_db[i].pid);

	/* send sigterm */
	kill(proc_db[i].pid, SIGTERM);
	c = 0;
	while (proc_db[i].pid) {
		sleep_ms(100);
		c++;
		if (c == 10) {
			printf("HWAPI: Error terminating process %d, sending kill signal...\n",
					proc_db[i].pid);
			kill(proc_db[i].pid, SIGKILL);
			return 0;
		}
	}

	return 1;
}

/** Get object new status
 *
 * Obtains the status the object must run in the current timestamp.
 *
 * This function is called by Services API during the Status()
 * call, in every object execution's loop.
 *
 * During an status change procedure, the new status won't be
 * returned after the timestamp is same or greater than the one
 * passed as the third parameter in the hwapi_proc_status_new() function
 *
 */
int hwapi_proc_status_get(void)
{
	assert (my_proc);

	if (my_proc->status.next_status != my_proc->status.cur_status) {
		if (my_proc->change_progress < 2) {
			if (get_tstamp() >= my_proc->status.next_tstamp) {
				my_proc->change_progress = 2;
				return my_proc->status.next_status;
			}
		} else {
			my_proc->status.cur_status = my_proc->status.next_status;
		}
	}
	return my_proc->status.cur_status;
}

/** Set a new status for an object
 * @returns 1 if changed ok, 0 if can't change, -1 if error
 */
int hwapi_proc_status_new(int obj_id, int next_status, int next_tstamp)
{
	int i;
	i = 0;

	if (!next_tstamp) {
		next_tstamp = get_tstamp();
	}

	while (i < MAX_PROCS && proc_db[i].obj_id != obj_id) {
		i++;
	}
	if (i == MAX_PROCS) {
		printf("HWAPI: Error object 0x%x not found\n", obj_id);
		return -1;
	}

	if ((!proc_db[i].change_progress && proc_db[i].status.next_status
			== proc_db[i].status.cur_status)
			|| next_tstamp == -1) {

		proc_db[i].status.next_status = next_status;
		proc_db[i].status.next_tstamp = next_tstamp;
		proc_db[i].change_progress = 1;

		return next_status;
	} else {
		return 0;
	}
}

/** Confirm a successfull status change.
 * @returns 1 if changed ok, 0 if still not changed.
 */
int hwapi_proc_status_ack(int obj_id)
{
	int i;
	i = 0;
	while (i < MAX_PROCS && proc_db[i].obj_id != obj_id) {
		i++;
	}
	if (i == MAX_PROCS) {
		return -1;
	}
	if (proc_db[i].status.cur_status == proc_db[i].status.next_status
			&& proc_db[i].change_progress == 2) {
		proc_db[i].change_progress = 0;
		return 1;
	} else {
		return 0;
	}
}

/** Proces Monitoring.
 *
 * Function used to gather information about current process status,
 * future status, cpu timing and other related information
 *
 * @param proc Pointer to hwapi_proc_i structure
 * @param obj_id id of the object assigned to the process
 *
 * @return 1 ok
 * @return 0 error (not found)
 */
int hwapi_proc_info(int obj_id, struct hwapi_proc_i *pinfo)
{
	int i;
	i = 0;

	assert (pinfo);

	while (i < MAX_PROCS && proc_db[i].obj_id != obj_id) {
		i++;
	}
	if (i == MAX_PROCS) {
		printf("HWAPI: Error object 0x%x not found\n", obj_id);
		return 0;
	}
	memcpy(pinfo, &proc_db[i], sizeof(struct hwapi_proc_i));
	return 1;
}

/** My process information.
 *
 * Fills hwapi_proc_i structure with the caller's info.
 *
 * @param proc Pointer to hwapi_proc_i structure
 *
 */
int hwapi_proc_myinfo(struct hwapi_proc_i *pinfo)
{
	assert (pinfo && my_proc);
	memcpy(pinfo, my_proc, sizeof(struct hwapi_proc_i));
	return 1;
}


/** Get Processes List.
 *
 */
int hwapi_proc_list(struct hwapi_proc_i *pinfo, int nelems)
{
	int i, k;

	assert (pinfo);

	k = 0;
	for (i = 0; i < MAX_PROCS; i++) {
		if (proc_db[i].obj_id) {
			if (k < nelems) {
				memcpy(&pinfo[k], &proc_db[i],
						sizeof(struct hwapi_proc_i));
			}
			k++;
		}
	}
	if (k > nelems) {
		printf("HWAPI: Caution more processes runnning but buffer too short (%d<%d)\n",
				nelems, k);
		k = nelems;
	}
	return k;
}

/** @} */

/** @defgroup stats Global Variable Functions
 *
 * The Hardware API enables processes to share global variables in an easy, fast
 * and efficient way.
 *
 *
 * @{
 */

/** Create a new shared variable.
 *
 * @param size Size of the variable, in bytes
 * @returns Id for the variable, 0 if error
 */
int hwapi_var_create(char *name, int size, int opts)
{
	int i;
	char *p;

	assert (var_db && name && size>=0);

	if (!my_proc) {
		printf("HWAPI: Error only objects can create variables\n");
		return -1;
	}

	/* alloc data*/
	p = mem_alloc(hw_api_db->var_table, size);
	if (!p) {
		xprintf ("HWAPI: Error allocating space for variable\n");
		return -1;
	}

	i = 1;
	while (i < MAX_VARS && var_db[i].obj_id) {
		i++;
	}
	if (i == MAX_VARS) {
		printf("HWAPI: Can't create more variables\n");
		return -1;
	}

	/* fill database */
	strncpy(var_db[i].name, name, 24);
	var_db[i].obj_id = my_proc->obj_id;
	var_db[i].stat_id = i;
	var_db[i].size = size;
	var_db[i].value_offset = (int) (p - hw_api_db->var_table);
	var_db[i].opts = opts;

	/* stat id is position + 1 */
	return i;
}

/** Delete a variable
 */
int hwapi_var_delete(int stat_id)
{
	assert (var_db);

	if (stat_id < 0 || stat_id > MAX_VARS) {
		return 0;
	}

	/* dealloc buffer */
	if (var_db[stat_id].value_offset) {
		mem_free(hw_api_db->var_table, hw_api_db->var_table
				+ var_db[stat_id].value_offset);
	}

	memset(&var_db[stat_id], 0, sizeof(struct hwapi_var_i));

	return 1;
}

/** Get a variable option
 */
int hwapi_var_getopt(int stat_id)
{
	assert (var_db);

	if (stat_id < 0 || stat_id > MAX_VARS) {
		return 0;
	}
	return var_db[stat_id].opts;
}

/** List all shared variables.
 *
 * Obtains a list of all variables created in the system
 * matching the given option (set to 0 to list all)
 */
int hwapi_var_list(int opts, struct hwapi_var_i *var_list, int nof_elems)
{
	int i, k;

	assert (var_db && var_list);

	k=0;

	for (i = 0; i < MAX_VARS && k < nof_elems; i++) {
		if (opts == var_db[i].opts || !opts) {
			memcpy(&var_list[k], &var_db[i],
					sizeof(struct hwapi_var_i));
			k++;
		}
	}

	return k;
}

/** Modify a variable option
 */
int hwapi_var_setopt(int stat_id, int new_opt)
{
	assert (var_db);

	if (stat_id < 0 || stat_id > MAX_VARS) {
		return 0;
	}
	var_db[stat_id].opts = new_opt;

	//xprintf ("pid %s set var %d opt %d\n", my_proc->obj_name, stat_id, new_opt);
	return 1;
}

/** Read a variable value
 * @param value_sz Size of the buffer, In bytes
 */
int hwapi_var_read(int stat_id, void *value, int buff_sz)
{
	int len;
	
	assert (var_db && value && buff_sz>=0);

	if (stat_id < 0 || stat_id > MAX_VARS) {
		return -1;
	}
	len = var_db[stat_id].cur_size;
	if (len > buff_sz) {
		len = buff_sz;
	}
	
	
	memcpy(value, hw_api_db->var_table + var_db[stat_id].value_offset, len);

//	printf ("stat read len %d (%s)\n", len, (char*) value);

	return len;
}

/** Write a variable value
 * @param value_sz Size of the variable, In bytes
 */
int hwapi_var_write(int stat_id, void *value, int value_sz)
{
	int len;
	
	assert (var_db && value && value_sz>=0);

	if (stat_id < 0 || stat_id > MAX_VARS) {
		return -1;
	}

	len = value_sz;;
	if (len >= var_db[stat_id].size) {
		len = var_db[stat_id].size;
	} else {
		var_db[stat_id].cur_size = len;
	}

	//  printf ("stat write 0x%x at 0x%x\n", *((char*) value), var_db[stat_id].value_offset );
	memcpy(hw_api_db->var_table + var_db[stat_id].value_offset, value, len);

	return len;
}

/* @} */

/**********************************************************

 int send_parent_cmd(int cmd, int *data, int size);

 Internal function, send a command to parent process (runph)

 **********************************************************/
int send_parent_cmd(int cmd, int *data, int size)
{
	int n;

	/* clear msb bit */
	msg.head.dst = MSG_MAGIC;
	msg.head.src = getpid();

	/* check size */
	if (size < 0 || size > MSG_SZ) {
		printf("HWAPI: Error with size %d\n", size);
		return -1;
	}

	msg.body[0] = cmd;

	/* copy body */
	memcpy(&msg.body[1], data, size << 2);

	/* send message */
	n = msgsnd(api_msg_id, &msg, size*4 + HEAD_SZ + 1, 0);
	if (n == -1) {
		perror("msgsnd");
		return -1;
	}

//	  printf ("sending signal to %d\n", getppid ());

	kill(getppid(), SIGUSR1);

	/* return message type is my pid */
	n = msgrcv(api_msg_id, &msg, MSG_SZ, getpid(), 0);
	if (n == -1) {
		perror("msgrcv");
		return -1;
	}

	/* return value is first word */
	return msg.body[0];
}

/*****************************************************************************
 * TIMING FUNCTIONS TO REVIEW







 *****************************************************************************/

/***********************
 *   SLEEP FUNCTIONS   *
 ***********************/
int sleep_ms(int msec)
{
	struct timespec treq, trem;
	if (msec <= 0)
		return (0);
	treq.tv_sec = msec / 1000;
	treq.tv_nsec = (msec % 1000) * 1000000;
	nanosleep(&treq, &trem);
	return (1);
}

int sleep_us(int usec)
{
	struct timespec treq, trem;
	if (usec <= 0)
		return (0);
	treq.tv_sec = usec / 1000000;
	treq.tv_nsec = (usec % 1000000) * 1000;
	nanosleep(&treq, &trem);
	return (1);
}
void get_time(time_t * t)
{
	time_t tact;
	gettimeofday(&tact, NULL);
	timersub (&tact, &hw_api_db->init_time, t);
}
void

set_time(time_t * t)
{
	time_t tact;
	gettimeofday(&tact, NULL);
	timersub (&tact, t, &hw_api_db->init_time);
}
void init_time(void)
{
	gettimeofday(&hw_api_db->init_time, NULL);
}
void

add_time(time_t * t1, time_t * t2, time_t * r)
{
	timeradd (t1, t2, r);
}
void

sub_time(time_t * t1, time_t * t2, time_t * r)
{
	timersub (t1, t2, r);
}
void

get_time_interval(time_t * tdata)
{

	/*First element stores the result,
	 *second element is the start time and
	 *third element is the end time.
	 */
	tdata[0].tv_sec = tdata[2].tv_sec - tdata[1].tv_sec;
	tdata[0].tv_usec = tdata[2].tv_usec - tdata[1].tv_usec;
	if (tdata[0].tv_usec < 0) {
		tdata[0].tv_sec--;
		tdata[0].tv_usec += 1000000;
	}
}

int time_to_tstamp(time_t *tdata)
{
	int tstamp;

	if (!ts_len) {
		ts_len = &hw_api_db->cpu_info.tslen_usec;
	}
	tstamp=(int) (tdata->tv_sec*(1000000/(*ts_len))+tdata->tv_usec
			/(*ts_len));
	return tstamp;
}
void tstamp_to_time(int tstamp, time_t *tdata)
{
	tdata[0].tv_sec=(tstamp*(*ts_len))/1000000;
	tdata[0].tv_usec=(tstamp*(*ts_len))-tdata[0].tv_sec*1000000;
}

int get_tstamp(void)
{
	time_t tdata;
	get_time(&tdata);
	return time_to_tstamp(&tdata);
}

/** @} */

