/**
* @file phal_hw_api.c
* @brief PHAL Hardware Library LINUX Implementation
*
* This file contains the implementation of the HW Library for 
* linux platforms. 	
*/  
  
/** default includes 
*/ 
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/msg.h>
#include <sys/ipc.h>
#include <sys/shm.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <sys/wait.h>
#include <sys/time.h>
#include <time.h>
#include <errno.h>
#include <unistd.h>
#include <fcntl.h>

#include "mem.h"
#include "phal_hw_api.h"  
#include "phal_hw_api_man.h"
#include "hwapi_backd.h"


/** hw_api general database */ 
extern struct hw_api_db *hw_api_db;
  
/** @defgroup res File Resource Management Functions
 * 
 * Accessing on-disk files for reading application, statistics or parameter
 * configurations may be realized through this group of functions.
 *  
 * @{
 */ 

/** Open a file, parse it to a buffer and close.
 *
 * Useful for reading configuration files. Reads a text
 * file, parses it (erase spaces, remove comments, etc.)
 * and saves it into an allocated buffer of memory. The
 * allocation is done inside this function.
 * The file is also closed by this function when job is 
 * finished and data is saved into memory. 
 *
 * @warnign This function requires the utils library to be 
 * linked to the program.
 *
 * @param res_name Name of the resource (file)
 * @param buffer Address of the pointer where allocated data
 * will be saved. 
 * @return Amount of bytes saved into memory, -1 if error
 */
 int
hwapi_res_parseall (char *res_name, char **buffer)
{
  FILE *f;
  int offset;
  char *path;

  path = calloc (1, strlen (res_name) + strlen (hw_api_db->res_path) + 2);
  
  if (!path)
    {
      return -1;
    }
    
  /* get file path */ 
  sprintf (path, "%s/%s", hw_api_db->res_path, res_name);
  
//  printf ("obro %s\n", path);
  
  f = fopen (path, "r");
  
  free (path);
  
  if (!f)
    {
      perror ("fopen");
      return -1;
    }
  
  offset = read_file (f, buffer, 0);
  
  fclose (f);
  
  if (offset < 0)
    {
      printf ("HWAPI: Error reading file.\n");
      return -1;
    }
  else
    {
      return offset;
    }
}
  
  



  
/** Open Resource (File).
 *
 * @todo path of resource files
 *
 * This function opens a resource so it can be read/written 
 * 
 * The function returns a file descriptor which will be later used
 * by read/write functions.
 *  
 * @param res_name String indicating the name of the resource
 * @param mode FLOW_READ_ONLY, FLOW_READ_WRITE, FLOW_WRITE_ONLY global constants
 * @param file_len fills with length of the file 
 *
 * @return >0 file descriptor
 * @return <0 error
 */ 
  int
hwapi_res_open (char *res_name, int mode, int *file_len)
{
  int n;
  //char args[4];
  int args;
  
    //int seek_end=0;
  int filelen;
  
  assert (res_name);
  
  char *path;

  path = calloc (1, strlen (res_name) + strlen (hw_api_db->res_path) + 2);
  
  if (!path)
    {
      return -1;
    }
    
  /* get file path */ 
  sprintf (path, "%s/%s", hw_api_db->res_path, res_name);
  
  /* get args param */ 
  switch (mode)
    {
    case FLOW_WRITE_ONLY:
      args = O_WRONLY | O_CREAT;
      
//              sprintf(args,"wb");
	break;
    case FLOW_READ_ONLY:
      args = O_RDONLY;
//              sprintf(args,"rb");
      break;
    case FLOW_READ_WRITE:
      args = O_RDWR | O_CREAT;
//              sprintf(args,"r+");
      break;
    }
  
    
  /* get file path */ 
  sprintf (path, "%s/%s", hw_api_db->res_path, res_name);
  
  /* open file */ 
  n = open (path, args, S_IRUSR | S_IWUSR);
  
  if (n < 0)
    {
      printf ("Error opening %s\n", path);
      perror ("open");
    }

  free (path);
  
  if (file_len)
    
    {
      filelen = lseek (n, 0, SEEK_END);
      lseek (n, 0, SEEK_SET);
      *file_len = filelen;
    }
  return n;
}




/** Close Resource.
 *	
 * Closes and deallocates resources associated with file descriptor fd.
 * 
 * No further read/write is possible after this call.
 * 
 * @param fd File descriptor
 *
 * @return >0 success
 * @return <0 error
 */
  int
hwapi_res_close (int fd)
{
  int n;
  n = close (fd);
  if (n < 0)
    {
      printf ("HWAPI: Error while closing fd=%d\n", fd);
      perror ("close");
      return -1;
    }
  return 1;
}


/** Read Resource.
 *
 * Reads blen bytes from the resource fd and saves to user buffer
 * 
 * If buffer is NULL pointer, it is equivalent to POSIX's lseek 
 * 
 * @param fd file descriptor 
 * @param buffer pointer where data must be saved
 * @param blen number of bytes to read
 *
 * @return >0 number of bytes successfully read
 * @return <0 error
 */
  int
hwapi_res_read (int fd, char *buffer, int blen)
{
  int n;
  assert (blen >= 0);

  if (buffer) {
	  n = read (fd, buffer, blen);	  
  } else {
	  n = lseek(fd, blen, SEEK_CUR);
  }
  if (n < 0)
    {
      perror ("read");
      return n;
    }
}



/** Read Line of Resource (File) Function:
 * 
 * Reads a text line (terminated by '\n') from the resource fd. 
 * User must provide buffer length (blen)
 * 
 */ 
#ifdef DEPRECIATED
  int
hwapi_read_line_res (int fd, char *buffer, int blen)
{
  
    /*         OLD
       
       
       int n,i;

       i=0;
       while(i<MAX_RESOURCES && hw_api_db->resource_db[i].id!=resource_id)
       i++;
       if (i==MAX_RESOURCES) {      
       printf("HWAPI: Error can't find resource with id 0x%x\n",resource_id);
       return -1;
       }

       if (hw_api_db->resource_db[i].f==0) {
       printf("HWAPI: Error while reading from resource 0x%x. Not opened.\n",resource_id);
       return -1;
       }
       n=ph_read_line(hw_api_db->resource_db[i].f,(char*) buffer,len);
       if (n<0) {
       printf("HWAPI: Error while reading from resource=0x%x f=0x%x\n",resource_id,hw_api_db->resource_db[i].f);
       return -1;
       }
     */ 
  int i = 0;
  int n;
  char c = 0;
  
  /* old
       n=fread(&c,sizeof(char),1,f);
    */ 
  n = read (fd, &c, 1);
  if (n < 0)
    {
      perror ("read");
      return -1;;
    }
  while (c != '\n')
    {
      buffer[i] = c;
      n = read (fd, &c, 1);
      if (!n)
	break;
      if (n < 0)
	{
	  perror ("read");
	  return -1;
	}
      if (i >= (blen - 1))
	{
	  printf ("HWAPI: Caution line bigger than buffer (%d<%d)\n", i,
		  blen);
	  buffer[i] = '\0';
	  return (i);
	}
      i++;
    }
  buffer[i] = '\0';
  return i;
}

#endif /*  */

/** Write Resource.
 *	
 * Writes blen bytes from the user buffer to the resource 
 * associated with file descpriptor
 * 
 * @param fd file descriptor 
 * @param buffer pointer where data must be read
 * @param blen number of bytes to write
 *
 * @return >0 number of bytes successfully read
 * @return <0 error
 */ 
  int
hwapi_res_write (int fd, char *buffer, int blen)
{
  int n;
  assert (buffer && blen >= 0);

  n = write (fd, buffer, blen);
  if (n < 0)
    
    {
      perror ("write");
    }
  return n;
}


/** @} */

