/** @file hwapi_netd.c
 * @brief PHAL HWAPI Linux Network Daemon
 *
 * This file contains the implementation of the Network Bridge
 * of the HWAPI for Linux.
 *
 * This program creates, manages and bridges tcp/ip interfaces
 * to local phal interfaces
 */



#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <signal.h>
#include <errno.h>
#include <string.h>
#include <sched.h>
#include <time.h>
#include <ctype.h>
#include <sys/ipc.h>
#include <sys/shm.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/socket.h>
#include <sys/wait.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <fcntl.h>

#include "phal_hw_api.h"
#include "hwapi_backd.h"
#include "hwapi_netd.h"


/**********************************************************
			DEBUGGING CONSTANTS 
**********************************************************/
/*
#define PKT_DEBHEAD
#define PKT_DEBCONT
#define CONN_DEBUG
#define PKT_DEBUG
#define PKT_DEBUG2
#define DEB_SIG
*/

/**********************************************************
			GLOBAL CONSTANTS
**********************************************************/

struct msgbuff msg_rcv, msg_snd;

int data_msg_id;


/** @defgroup launcher_net PHAL Launcher for Linux Network Bridge Functions
 *
 * This group of functions are called by runph.c main function at boot and act like
 * bridge between logical and external physical interfaces.
 *
 * @{
 */



/** Network Send Packet
 *
 * Checks if a packet is available in logical interface and sends it throught an 
 * output interface. The external PHAL header is attached at the beggining so other 
 * platforms understand it.
 * 
 * @param fd File descriptor to the external socket
 * @param mbox Mailbox of the internal interface
 *
 * @return >0 number of bytes sent
 * @return 0 no packets available
 * @return <0 error 
 */
int
net_packet_snd (int fd, long mbox)
{
  int n, l;
  int tstamp;
  int type;
  packet_header_t *pkth;
  int *buffer;

#ifdef PKT_DEBUG
  time_t tdata;
#endif

  tstamp = get_tstamp ();
  type = mbox << TSTAMP_PKT_BITS;

#ifdef PKT_DEBUG2
  printf ("HWAPI_NETD: --SND-- waiting fd=%d mbox=0x%x...\n", fd, mbox);
#endif

  n = msgrcv (data_msg_id, &msg_snd, MSG_SZ, type, 0);
  if (n == -1 && errno == ENOMSG)
    {
      return 0;
    }
  if (n == -1)
    {
      perror ("msgrcv");
      return n;
    }

#ifdef PKT_DEBCONT
  xprintf ("RUNPH: SNDCONT: 0x%x-0x%x-0x%x\n", msg_snd.body[0],
	   msg_snd.body[1], msg_snd.body[2]);
#endif


  buffer = malloc (n - HEAD_SZ + (PACKET_HEADER_SIZE << 2));
  pkth = (packet_header_t *) buffer;

  pkth->id = PACKET_MAGIC;
  pkth->opts = msg_snd.head.opts;

  pkth->tstamp = msg_snd.head.tstamp;
  pkth->size = (n - HEAD_SZ) >> 2;


  memcpy (&buffer[PACKET_HEADER_SIZE], msg_snd.body, n - HEAD_SZ);

#ifdef PKT_DEBUG
  get_time (&tdata);
  printf
    ("HWAPI_NETD: -- SND -- TS=%d (t=%d:%d) type=0x%x size=%d opts=0x%x\n",
     get_tstamp (), tdata.tv_sec, tdata.tv_usec, type, pkth->size,
     pkth->opts);
#endif
  l = n - HEAD_SZ + (PACKET_HEADER_SIZE << 2);
  do
    {
      n = write (fd, buffer, l);
    }
  while (n == -1 && errno == EINTR);

#ifdef PKT_DEBHEAD
  printf ("RUNPH: SNDHEAD: id=0x%x,opts=0x%x,ts=0x%x,sz=0x%x\n", pkth->id,
	  pkth->opts, pkth->tstamp, pkth->size);
#endif
  if (n != l)
    {
      printf ("SND ALARM: (%d!=%d)\n", n, l);
    }

  free (buffer);

  if (n == -1)
    {
      perror ("write");
      return -1;
    }

  return n;
}




/** Network Receive Packet
 *
 * Checks if a packet is available in external physical interface and bridges it to the
 * internal interfcae.  
 * 
 * @param fd File descriptor of the external socket
 * @param mbox Mailbox of the internal interface
 *
 * @return >0 number of bytes received
 * @return 0 no packets available
 * @return <0 error 
 */
int
net_packet_rcv (int fd, long mbox)
{
  packet_header_t pkth;
  int n, rpm;
  int size;
  int signal_pid = 0;
  int do_signal = 0;
  char *bptr;

#ifdef PKT_DEBUG
  time_t tdata;
#endif

#ifdef PKT_DEBUG2
  printf ("HWAPI_NETD: --RCV-- Waiting fd=%d,mbox=0x%x\n", fd, mbox);
#endif

  bptr = (char *) &pkth;
  rpm = 0;
  do
    {
      n = read (fd, &bptr[rpm], sizeof (packet_header_t));
      if (n >= 0)
	rpm += n;
    }
  while (n > 0 && rpm < (sizeof (packet_header_t)));

  if (n == 0)
    return -1;

  if (n < 0)
    {
      fflush (0);
      perror ("read");

      return -1;
    }

  if (n != sizeof (packet_header_t))
    {
      xprintf ("HEADER ERROR (%d!=%d)!!\n", n, sizeof (packet_header_t));
    }

#ifdef PKT_DEBHEAD
  printf ("RUNPH: RCVHEAD: id=0x%x,opts=0x%x,ts=0x%x,sz=0x%x\n", pkth.id,
	  pkth.opts, pkth.tstamp, pkth.size);
#endif

  if (pkth.id != PACKET_MAGIC)
    {
      printf
	("RUNPHD: Received an invalid packet 0x%x. Closing connection.\n",
	 pkth.id);
      return -1;
    }
  size = pkth.size;
  if (size > MSG_BUFF_SZ)
    {
      printf
	("RUNPHD: Caution: packet size exceed msg buffer size (%d>%d)... trunkating\n",
	 size, MSG_BUFF_SZ);
      size = MSG_BUFF_SZ;
    }
  bptr = (char *) msg_rcv.body;
  rpm = 0;
  do
    {
      n = read (fd, &bptr[rpm], (size << 2) - rpm);
      if (n >= 0)
	rpm += n;
//              printf("n=%d,rpm=%d,sz=%d\n",n,rpm,size<<2);
    }
  while ((n == -1 && errno == EINTR) || (n > 0 && rpm < (size << 2)));

  if (n == 0)
    {
      return 0;
    }

  if (n < 0)
    {
      perror ("read");
      return n;
    }

  if (rpm != size << 2)
    {
      xprintf ("CONTENTS ERROR (%d!=%d)!!\n", rpm, size << 2);
    }

#ifdef PKT_DEBCONT
  xprintf ("RUNPH: RCVCONT: 0x%x-0x%x-0x%x\n", msg_rcv.body[0],
	   msg_rcv.body[1], msg_rcv.body[2]);
#endif


  /* read all packet although buffer exceed */
  if (pkth.size > size)
    {
      n = lseek (fd, (pkth.size - size) << 2, SEEK_CUR);
      if (n < 0)
	return n;
    }

  /* bridge packet to msg mbox */
/*  if ((pkth.opts & ITF_OPT_TIME_MASK) == ITF_OPTS_URG)
    {
      signal_pid = get_itf_r_pid (pkth.opts & ITF_ARGS_MASK);
      do_signal = 1;
    }
*/  msg_rcv.head.dst =
    mbox << TSTAMP_PKT_BITS | (pkth.tstamp & TSTAMP_PKT_MASK);
  msg_rcv.head.opts = pkth.opts;


  do
    {
      n = msgsnd (data_msg_id, &msg_rcv, HEAD_SZ + (size << 2), 0);
    }
  while (n == -1 && errno == EINTR);

  if (n < 0)
    {
      perror ("msgsnd");
      return n;
    }

#ifdef PKT_DEBUG
  get_time (&tdata);
  printf
    ("HWAPI_NETD: -- RCV -- TS=%d (t=%d:%d) type=0x%x size=%d opts=0x%x\n",
     get_tstamp (), tdata.tv_sec, tdata.tv_usec, msg_rcv.head.dst, size,
     pkth.opts);
#endif

  if (do_signal)
    {
      //signal_pid=pkth.opts&ITF_ARGS_MASK;
#ifdef DEB_SIG2
      printf ("RUNPHD: Sending SIGUSR2 to process %d itf 0x%x\n", signal_pid,
	      pkth.opts & ITF_ARGS_MASK);
#endif
/*      if (signal_pid > 0)
	sigqueue (signal_pid, SIGUSR2,
		  (union sigval) (pkth.opts & ITF_ARGS_MASK));
*/    }
  return 1;
}


/** Close Networds child
 *
 * Closes a network process created for bridging
 *
 * @sa Net_thread()
 */
int
Net_close_chld (struct ext_itf *ext_itf, int wait_time)
{
  int c;
  int sleep_time;
  
  if (!wait_time)
    {
      wait_time = 500;
    }
  
  sleep_time = wait_time / 10;

  if (ext_itf->bpid > 0)
    {
      kill (ext_itf->bpid, SIGTERM);
      c = 0;
      while (ext_itf->bpid > 0)
	{
	  sleep_ms (sleep_time);
	  c++;
	  if (c == 10)
	    {
	      printf ("RUNPHD: Trouble closing network b-child. Kill %d.\n",
		      ext_itf->bpid);
	      kill (ext_itf->bpid, SIGKILL);
	      return -1;
	    }
	}
    }
  if (ext_itf->pid > 0)
    {
      kill (ext_itf->pid, SIGTERM);
      c = 0;
      while (ext_itf->pid > 0)
	{
	  sleep_ms (sleep_time);
	  c++;
	  if (c == 10)
	    {
	      printf ("RUNPHD: Trouble closing network child. Kill %d.\n",
		      ext_itf->pid);
	      kill (ext_itf->pid, SIGKILL);
	      return -1;
	    }
	}
    }
  return 1;
}


/** Net Close
 *
 * Closes main network thread and calls Net_close_chld for each
 * process created (one for each interface)
 */
void
Net_close (struct net_db *net_db, int wait_time)
{
  int i, c;
  int sleep_time;
  int error = 0;

  if (!wait_time)
    {
      wait_time = 500;
    }

  sleep_time = wait_time / 10;


  /* kill all network childs */
  for (i = 0; i < net_db->nof_ext_itf; i++)
    {
      if (Net_close_chld (&net_db->ext_itf[i], wait_time) < 0)
	error++;
    }

  /* now kill main net thread */
  if (net_db->net_chld > 0)
    {
      kill (net_db->net_chld, SIGTERM);
      c = 0;
      while (net_db->net_chld > 0)
	{
	  sleep_ms (sleep_time);
	  c++;
	  if (c == 10)
	    {
	      printf ("RUNPHD: Trouble closing main network child. Kill %d\n",
		      net_db->net_chld);
	      kill (net_db->net_chld, SIGKILL);
	      error++;
	      break;
	    }
	}
    }


  if (error)
    {
      printf
	("RUNPHD: Some network childs could not be closed properly (%d).\n",
	 error);
    }
}


/** Network Interface thread
 *
 * This is the main thread for every network inteface. Depending on its type, it listens to 
 * either external or internal interface for new packets and bridge to the opposite site. 
 * In case of bidirectional interfaces, another thread is launched from this function.
 *
 * @todo Check this function
 */
void
do_net_chld (struct ext_itf *xitf)
{

  int n, pid, fd;
  int i_listen, do_snd, do_rcv;

  switch (xitf->mode)
    {
    case FLOW_READ_ONLY:
      i_listen = 1;
      do_snd = 0;
      do_rcv = 1;
      break;
    case FLOW_WRITE_ONLY:
      i_listen = 0;
      do_snd = 1;
      do_rcv = 0;
      break;
    case FLOW_READ_WRITE:
      i_listen = 1;
      do_snd = 1;
      do_rcv = 1;
      break;
    case FLOW_WRITE_READ:
      i_listen = 0;
      do_snd = 1;
      do_rcv = 1;
      break;
    }
#ifdef CONN_DEBUG
  printf ("RUNPHD: Pid %d starting on itf 0x%x mode 0x%x...\n", getpid (),
	  xitf->id, xitf->mode);
#endif
  while (1)
    {
      /* start the connection or setup server */
      if (i_listen)
	{
#ifdef CONN_DEBUG
	  printf ("RUNPH_NET: Pid %d waiting for connection...\n", getpid ());
#endif
	  fd = setup_server (htonl (INADDR_ANY), xitf->port, &xitf->fsock);
	  if (fd < 0)
	    {
	      printf ("RUNPH_NET: Error setting server\n");
	      exit (-1);
	    }
	  xitf->fd = fd;
#ifdef CONN_DEBUG
	  printf ("RUNPH_NET: Input connection. id=0x%x fd=%d\n", xitf->id,
		  fd);
#endif
	}
      else
	{
#ifdef CONN_DEBUG
	  printf ("RUNPH_NET: Pid %d waiting to configure output itf...\n",
		  getpid ());
#endif
	  while (!xitf->out_mbox)
	    sleep_ms (100);
	  fd = setup_client (inet_addr (xitf->address), xitf->port);
	  if (fd < 0)
	    {
	      printf ("RUNPH_NET: Error connecting to server\n");
	      exit (-1);
	    }
	  xitf->fd = fd;
#ifdef CONN_DEBUG
	  printf ("RUNPH_NET: Output connection. id=0x%x fd=%d\n", xitf->id,
		  fd);
#endif
	}
      if (do_snd && do_rcv)
	{
	  pid = fork ();
	  if (pid < 0)
	    {
	      perror ("fork");
	      exit (-1);
	    }
	  if (pid > 0)
	    {
#ifdef CONN_DEBUG
	      printf ("RUNPH_NET: Child %d created\n", pid);
#endif
	      xitf->bpid = pid;
	    }
	}
      else
	pid = 0;


      if (do_snd || !pid)
	{
	  if (do_snd && !pid)
	    {
	      if (i_listen)
		close (xitf->fsock);

#ifdef CONN_DEBUG
	      printf ("RUNPH_NET: Waiting to configure output itf...\n");
#endif
	      while (!xitf->out_mbox)
		sleep_ms (100);

#ifdef CONN_DEBUG
#endif
	      do
		{
		  n = net_packet_snd (fd, xitf->out_mbox);
		}
	      while (n > 0);

#ifdef CONN_DEBUG
	      printf ("RUNPH_NET: Out Connection lost. id 0x%x pid %d.\n",
		      xitf->id, getpid ());
#endif
	      close (fd);
	      xitf->out_mbox = 0;
	      xitf->fd = 0;
	      if (do_rcv)
		exit (0);
	    }
	  else
	    {
#ifdef CONN_DEBUG
	      printf ("RUNPH_NET: Waiting in mbox to configure...\n");
#endif
	      while (!xitf->in_mbox)
		sleep_ms (100);

#ifdef CONN_DEBUG
	      printf ("RUNPH_NET: starting rcv pid %d mbox 0x%x\n", getpid (),
		      xitf->in_mbox);
#endif
	      do
		{
		  n = net_packet_rcv (fd, xitf->in_mbox);
		}
	      while (n > 0);
	      close (fd);
#ifdef CONN_DEBUG
	      printf ("RUNPH_NET: In Connection lost. id 0x%x pid %d\n",
		      xitf->id, getpid ());
#endif
	      xitf->fd = 0;
	      if (do_snd)
		{
#ifdef CONN_DEBUG
		  printf ("killing snd chld proc %d\n", pid);
#endif
		  kill (pid, SIGKILL);
		}
	    }
	}
    }
}


/** Main Network Thread
 *
 * This is the main network thread. It launches a new thread for every external interface
 * depending on its characteristics.
 *
 * @todo Check this function.
 */
void
Net_thread (struct net_db *net_db, int msg_id)
{
  int i, n;
  int netpid;

  netpid = fork ();
  if (netpid < 0)
    {
      perror ("fork");
      exit (-1);
    }
  if (netpid > 0)
    return;

  data_msg_id = msg_id;
  net_db->net_chld = getpid ();

  /* Network child continues here */
  i = 0;

  for (i = 0; i < net_db->nof_ext_itf; i++)
    {
      net_db->ext_itf[i].pid = 0;
      net_db->ext_itf[i].fsock = 0;
    }

  while (1)
    {

      for (i = 0; i < net_db->nof_ext_itf; i++)
	{
	  if (!net_db->ext_itf[i].pid)
	    {
	      n = fork ();
	      if (n < 0)
		{
		  perror ("fork");
		  exit (-1);
		}
	      if (!n)
		{
		  do_net_chld (&net_db->ext_itf[i]);
		}
	      else
		{
		  printf ("net chld created with pid %d\n", n);
		  net_db->ext_itf[i].pid = n;
		}
	    }
	}
      sleep_ms (100);
    }
  printf ("net child is exiting\n");
  exit (0);
}


/** @} */


