#include <stdlib.h>
#include <string.h>

#include <assert.h>
#include "mem.h"

#define T mem_o

struct chunk
  {
    void *ptr;
    short slot_st;
    short slot_end;
  };

struct T
  {
    int nof_slots;
    int slot_sz;
    int buff_sz;
    int slots_offset;
    int chunks_offset;
    int data_offset;
  };

/** Create new memory pool region.
 *  This function creates a new pool given a memory region
 * defined by an address and a size. Within this buffer, the
 * function will allocate needed data to manage the pool.
 * The user must, also, provide the size of the minimun
 * memory slot that can be assigned.
 * 
 * @param addr Pointer to the buffer
 * @param buff_sz Size (in bytes) of the buffer
 * @param slot_sz Size (in bytes) of the minimum assigned slot.
 * 
 * @return Pointer to the object (same as addr), null if error.
 */
void * mem_new(void *addr, int buff_sz, int slot_sz)
  {
    T mem;
    int offset;
    int nof_slots;

    assert (addr && slot_sz>0 && buff_sz>0);

    /** First calculate a number of slots.
     * This number (N) depends on buffer size, slot size
     * and the needed size for the internal management:
     * 
     * N = (buff_sz - man_sz) / slot_sz;
     * man_sz = object_sz + 1*N + chunk_sz*N
     * N = (buff_sz - object_sz) / (slot_sz + 1 + chunk_sz)
     * 
     * where the lowest integer is taken from the division.
     * This means that at the end some data may be unused.
     */
    nof_slots = (int) (buff_sz - sizeof (*mem)) / (slot_sz + 1
        + sizeof(struct chunk));
    if (nof_slots <= 0)
      {
        return NULL;
      }

    offset = 0;
    /* set mem object at the beggining */
    mem = (T) addr;
    offset += sizeof (*mem);

    /* next used slots identifiers */
    mem->slots_offset = offset;
    offset += nof_slots;

    /* next chunks identifiers */
    mem->chunks_offset = offset;
    offset += nof_slots * sizeof(struct chunk);

    mem->data_offset = offset;
    offset += nof_slots * slot_sz;
        
    /* here if offset > buff_sz a fatal error ocurred */
    assert (offset < buff_sz);

    /* save parameters */
    mem->nof_slots = nof_slots;
    mem->buff_sz = buff_sz;
    mem->slot_sz = slot_sz;

    return addr;
  }

void mem_delete(void *addr)
  {
    T mem = (T) addr;
    
    assert (addr);

    memset(addr, 0, mem->buff_sz);
  }

void * mem_alloc(void *addr, int size)
  {
    int i, j;
    void *p;
    struct chunk *chunks, *chunk;
    char *slots;
    
    T mem = (T) addr;

    assert (mem && mem && size>=0);

    chunks = (struct chunk *) addr + mem->chunks_offset;
    slots = (char *) addr + mem->slots_offset;

    /* get a free chunk */
    for (i = 0; i < mem->nof_slots; i++)
      {
        if (!chunks[i].ptr)
          {
            break;
          }
      }


    chunk = &chunks[i];

    /* find available slots for required memory */
    j = 0; /* slot start index */
    i = 0; /* slot end index (+1) */

    do
      {
        if (slots[i])
          {
            j = i;
          }
        i++;
      } while (((i - j) * mem->slot_sz < size) && i < mem->nof_slots);

    if (i == mem->nof_slots)
      {
        return NULL;
      }


    /* save pointer */
    p = addr + mem->data_offset + j * mem->slot_sz;

    /* save chunk propierties */
    chunk->ptr = p;
    chunk->slot_st = j;
    chunk->slot_end = i;

    /* alloc space */
    for (; j < i; j++)
      {
        slots[j] = 1;
      }


    return p;
  }

void * mem_calloc(void *addr, int size)
  {
    void *p;
    T mem = (T) addr;

    assert (mem);

    p = mem_alloc(mem, size);
    if (p)
      {
        memset(p, 0, size);
      }
    return p;
  }

void mem_free(void *addr, void *ptr)
  {
    int i;
    struct chunk *chunks, *chunk;
    char *slots;
    T mem = (T) addr;

    assert (mem);

    /* do nothing if pointer is 0 */
    if (!ptr)
      {
        return;
      }

    chunks = (struct chunk *) addr + mem->chunks_offset;
    slots = (char *) addr + mem->slots_offset;

    /* find chunk */
    for (i = 0; i < mem->nof_slots; i++)
      {
        if (chunks[i].ptr == ptr)
          {
            break;
          }
      }

    chunk = &chunks[i];

    /* dealloc used space */
    for (i = chunk->slot_st; i < chunk->slot_end; i++)
      {
        slots[i] = 0;
      }

    /* and delete chunk data */
    memset(chunk, 0, sizeof(struct chunk));

  }

void * mem_realloc(void *addr, void *ptr)
  {
    return NULL;
  }
