/** @file parse_str.c
 * @brief PHAL Launcher for Linux Parsing Functions
 *
 * This file implements functions called to parse the following
 * files need to launch the PHAL Linux Launcher.
 *     	- Daemon Configuration File: Describes the daemons that will
 * 		be launched.
 * 	- Interface Configuration File: Describes the external physical
 *		interfaces (port, address, r/w modes, etc.)
 *	- CPU Configuration File: Describes the local processor characteristics.
 */


#include <sys/types.h>
#include <sys/mman.h>
#include <sys/fcntl.h>
#include <sys/stat.h>
#include <sys/param.h>
#include <sys/wait.h>
#include <unistd.h>
#include <errno.h>
#include <memory.h>
#include <dirent.h>
#include <fcntl.h>
#include <limits.h>
#include <unistd.h>
#include <setjmp.h>
#include <signal.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <sched.h>

/********************************/
/****   LOCAL INCLUDES       ****/
/********************************/

#define PARSE_LINE_LEN	163

#include "mem.h"
#include "phal_hw_api.h"
#include "hwapi_backd.h"
#include "parse_str.h"


/** @defgroup parse PHAL Launcher for Linux Parsing Functions
 *
 * This functions parse the PHAL Launcher Configuration files
 * need to launch PHAL main process.
 *
 * @{
 */



/** Interface Configuration Parameters
*/
char *itf_parameters[NOF_ITF_PARAMS+1];


/** Configure Interface Config file parameters
 */
void init_itf_param_strings(void)
{
	itf_parameters[ITF_ID]="id"; 		/* Id of the Interface */
	itf_parameters[ITF_PORT]="port";	/* TCP local/remote port */
	itf_parameters[ITF_ADDRESS]="address";	/* IP address */
	itf_parameters[ITF_MODE]="mode";	/* "in" or "out" */
	itf_parameters[ITF_BW]="bw";		/* float, in bps */
	itf_parameters[ITF_DELAY]="delay";	/* int, in milisec */
	itf_parameters[ITF_TYPE]="type";
	itf_parameters[NOF_ITF_PARAMS]="\0";
}

/** Interface Configuration Parsing Function.
 *
 * Actually parses the file and fills ext_itf structure with parsed data.
 *
 * @sa init_itf_param_string
 *
 * @param f FILE pointer
 * @param ext_itf Pointer to first element of the ext_itf database
 * @param max_itf Size in number of elements of the database
 *
 * @return >=0 Number of interfaces parsed
 * @return <0 Error parsing
 */
int parse_itf_cfg(FILE *f, struct ext_itf *ext_itf, int max_itf) {

	int file_len;
	char string[PARSE_LINE_LEN];
	int nof_itf;
	int error;
	int next_section,i,n;
	char *buffer,*pname,*pvalue;
	int line;
	int aux;

	init_itf_param_strings();

        file_len=read_file(f,&buffer,0);
        if (file_len<0)
        {
        	printf("ERROR: There were errors reading config file\n");
                return(-1);
	}

	i=0;
	nof_itf=0;
	while(i<file_len && nof_itf<max_itf)
	{
		n=read_line(buffer,i,string);
		i+=n;
		if (strcmp(string,"[xitf]")==0)
		{
			line=0;
			next_section=0;
			error=0;
			while((!next_section) && (i<file_len))
			{
				n=read_line(buffer,i,string);
				if (string[0]=='[')
					next_section=1;
				else
				{
					i+=n;
					line++;
					pname=string;
					pvalue=strchr(string,'=');
					pvalue[0]='\0';
					pvalue++;
					switch(param_name_start(pname,itf_parameters)) {
					case ITF_ADDRESS:
						if (strlen(pvalue)<16) {
							strcpy(ext_itf[nof_itf].address,pvalue);
						}
						else {
							printf("Error parsing xitf file. Address to long.\n");
							error=1;
							break;
						}
						break;
					case ITF_ID:
						sscanf(pvalue,"0x%x",&aux);
						ext_itf[nof_itf].id=aux;
						ext_itf[nof_itf].info.id=aux;
						break;
/*					case ITF_TYPE:
						if (!strcmp(pvalue,"udp"))
							ext_itf[nof_itf].type=NET_UDP;
						break;
*/					case ITF_MODE:
						if (!strcmp(pvalue,"out"))
							ext_itf[nof_itf].mode=FLOW_WRITE_ONLY;
						else if (!strcmp(pvalue,"in"))
							ext_itf[nof_itf].mode=FLOW_READ_ONLY;
						else if (!strcmp(pvalue,"inout"))
							ext_itf[nof_itf].mode=FLOW_READ_WRITE;
						else if (!strcmp(pvalue,"outin"))
							ext_itf[nof_itf].mode=FLOW_WRITE_READ;
						else {
							printf("Error parsing xitf file. Mode must be in/out.\n");
							error=1;
							break;
						}
						ext_itf[nof_itf].info.mode=ext_itf[nof_itf].mode;
						break;
					case ITF_PORT:
						sscanf(pvalue,"%d",&aux);
						ext_itf[nof_itf].port=aux;
						break;
					case ITF_DELAY:
						ext_itf[nof_itf].info.delay=atof(pvalue);
						break;
					case ITF_BW:
						ext_itf[nof_itf].info.BW=atof(pvalue);
						break;
					default:
						printf("Error parsing xitf file: Unknown parameter\n");
						error=1;
						break;
					}

					if (error) {
						printf("Error parsing xitf file. Itf %d, parameter %d\n",nof_itf+1,line);
						return -1;
					}
				}
			}
//			ext_itf[nof_itf].type=NET_TCP;
			nof_itf++;
			if (nof_itf==max_itf)
				printf("CAUTION: Maximum interfaces exceed\n");
		}
	}
	if (error) {
		return -1;
	}
	else
		return nof_itf;
}

/** Daemon Configuration Parameters
*/
char *d_parameters[NOF_DAEMON_PARAMS+1];

/** CPU Configuration Parameters
*/
char *cpu_parameters[NOF_CPU_PARAMS+1];


/** Daemon Configuration Parameters configuration.
 *
 * Configures the parameters accepted in the daemon configuration file
 *
 * @todo Separate CPU parsing from daemon parsing. More CPU parameters?
 * @todo Replace name by daemon code
 *
 */
void init_daemon_param_strings(void)
{
	d_parameters[DAEMON_NAME]="name";	/* name of the daemon*/
	d_parameters[DAEMON_PATH]="exe_path";	/* path to the executable */
	d_parameters[DAEMON_RUNABLE]="runable";
	d_parameters[DAEMON_OUTPUT]="output";
	d_parameters[NOF_DAEMON_PARAMS]="\0";

	cpu_parameters[CPU_MIPS]="mips";	/* float, MIPS of the processor */
	cpu_parameters[CPU_TSLOT]="tslot";
	cpu_parameters[CPU_VIRTUAL]="en_virtual_plt";
	cpu_parameters[NOF_CPU_PARAMS]="\0";
}


/** Daemon & CPU Configuration Parsing Function.
 *
 * Actually parses the file and fills daemon_db structure with parsed data.
 *
 * @sa init_daemon_param_string
 *
 * @param f FILE pointer
 * @param daemon_db Pointer to first element of the daemon_db database
 * @param cpu_info Pointer to cpu configuration structure
 * @param max_d Size in number of elements of the database
 *
 * @return >=0 Number of daemons parsed
 * @return <0 Error parsing
 */
int parse_cpu_cfg(FILE *f, struct launch_daemons *daemon_db, int max_d, struct hw_api_db *hwapi_db) {

	int file_len;
	char string[PARSE_LINE_LEN];
	int nof_d;
	int error;
	int next_section,i,n;
	char *buffer,*pname,*pvalue;
	int line;

	init_daemon_param_strings();

        file_len=read_file(f,&buffer,0);
        if (file_len<0)
        {
        	printf("ERROR: There were errors reading config file\n");
                return(-1);
	}

	i=0;
	nof_d=0;
	while(i<file_len && nof_d<max_d)
	{
		n=read_line(buffer,i,string);
		i+=n;
		if (strcmp(string,"[daemon]")==0)
		{
			line=0;
			next_section=0;
			error=0;
			while((!next_section) && (i<file_len))
			{
				n=read_line(buffer,i,string);
				if (string[0]=='[')
					next_section=1;
				else
				{
					i+=n;
					line++;
					pname=string;
					pvalue=strchr(string,'=');
					pvalue[0]='\0';
					pvalue++;
					switch(param_name_start(pname,d_parameters)) {
					case DAEMON_PATH:
						if (strlen(pvalue)<128) {
							strncpy(daemon_db[nof_d].path,pvalue,128);
						}
						else {
							printf("Error parsing daemons file. Exe_path too long.\n");
							error=1;
							break;
						}
						break;
					case DAEMON_OUTPUT:
						if (strlen(pvalue)<128) {
							strncpy(daemon_db[nof_d].output,pvalue,128);
						}
						else {
							printf("Error parsing daemons file. Exe_path too long.\n");
							error=1;
							break;
						}
						break;
					default:
						printf("Error parsing daemons file: Unknown parameter\n");
						error=1;
						break;
					}

					if (error) {
						printf("Error parsing daemons file. Daemon %d, parameter %d\n",nof_d+1,line);
						return -1;
					}
				}
			}
			nof_d++;
			if (nof_d==max_d)
				printf("CAUTION: Maximum daemons exceed\n");
		}
		if (strcmp(string,"[cpu_info]")==0)
		{
			line=0;
			next_section=0;
			error=0;
			while((!next_section) && (i<file_len))
			{
				n=read_line(buffer,i,string);
				if (string[0]=='[')
					next_section=1;
				else
				{
					i+=n;
					line++;
					pname=string;
					pvalue=strchr(string,'=');
					pvalue[0]='\0';
					pvalue++;
					switch(param_name_start(pname,cpu_parameters)) {
					case CPU_MIPS:
						hwapi_db->cpu_info.C=atof(pvalue);
						break;
					case CPU_TSLOT:
						hwapi_db->cpu_info.tslen_usec = atoi (pvalue);
						break;
					case CPU_VIRTUAL:
						hwapi_db->virtual_plt = strncmp(pvalue,"yes",3)?0:1;
						break;
					default:
						printf("Error parsing daemons file: Unknown parameter\n");
						error=1;
						break;
					}

					if (error) {
						printf("Error parsing daemons file. group %d, parameter %d\n",nof_d+1,line);
						return -1;
					}
				}
			}
		}

	}
	if (error) {
		return -1;
	}
	else
		return nof_d;
}
