#include <stdlib.h>
#include <string.h>

#include "phal_hw_api.h"

#include "phal_daemons.h"

#include "phid.h"
#include "pkt.h"
#include "daemon.h"
#include "str.h"
#include "var.h"

#include "cmdman_backend.h"

#include "frontend_cmds.h"
#include "swman_cmds.h"
#include "statsman_cmds.h"
#include "cmdman_cmds.h"
#include "exec_cmds.h"

int fd_r, fd_w;

void *daemon_x, *packet_x;

int wait_answer_pkt (int max_sec)
{
	time_t t[3];
	int n;

	get_time(&t[1]);
	do {
		n = daemon_rcvpkt((daemon_o) daemon_x);
		if (!n) {
			get_time(&t[2]);
			get_time_interval(t);
			hwapi_relinquish_daemon();
		}
	} while (!n && t[0].tv_sec < max_sec);

	return n;
}

void cmdman_execreports(void) 
{
	daemon_sendto((daemon_o) daemon_x, EXEC_STARTREPORT, 2, DAEMON_EXEC);
}


/** Initialize cmdmanager backend
 */
int cmdman_init(void)
{
	daemon_x = daemon_new(DAEMON_CMDMAN);
	if (daemon_x) {
		packet_x =  daemon_pkt((daemon_o) daemon_x);
		return 1;
	} else {
		return 0;
	}

}

/** De-allocates backend
 */
void cmdmand_close ()
{
	daemon_delete ((daemon_o*) &daemon_x);
}

/** Load an application in P-HAL
 * @param app_name Name of the app, as known by SWMAN
 * @param max_sec Timeout for the loading process
 */
int cmdman_loadapp(char *app_name, int max_sec)
{
	assert(app_name && (max_sec >= 0));

	pkt_clear((pkt_o) packet_x);
	strncpy(pkt_putval((pkt_o) packet_x, FIELD_APPNAME, strlen(app_name) + 1), app_name,
			pkt_freesz((pkt_o) packet_x));
	daemon_sendto((daemon_o) daemon_x, SWMAN_LOADAPP, 2, DAEMON_SWMAN);

	if (wait_answer_pkt(max_sec)) {
		if (pkt_getcmd((pkt_o) packet_x) == CMDMAN_SWLOADAPPOK) {
			return 1;
		}
	}
	return 0;
}

/** Change application status
 * @param app_name Name of the loaded application
 * @param new_status New P-HAL Status constant (see phal_daemons.h)
 * @param max_sec Timeout for status change
 */
int cmdman_appstatus(char *app_name, int new_status, int max_sec)
{
	assert(app_name);

	pkt_clear((pkt_o) packet_x);
	strncpy(pkt_putval((pkt_o) packet_x, FIELD_APPNAME, strlen(app_name) + 1), app_name,
			pkt_freesz((pkt_o) packet_x));
	pkt_putvalue((pkt_o) packet_x, FIELD_STATUS, new_status);

	pkt_putvalue((pkt_o) packet_x, FIELD_TSTAMP, get_tstamp() + 10);
	daemon_sendto((daemon_o) daemon_x, SWMAN_APPSTATUS, 2, DAEMON_SWMAN);

	if (wait_answer_pkt(max_sec)) {
		if (pkt_getcmd((pkt_o) packet_x) == CMDMAN_SWSTATUSOK) {
			return 1;
		}
	}
	return 0;
}

/** Get an statistic value
 * @param app_name Name of the application
 * @param obj_name Name of the object who registered the stat
 * @param stat_name Name of the variable
 * @param value Pointer to a buffer where data will be saved
 * @param type Pointer to where statistic type will be saved
 * @param size Pointer to where statistic size will be saved
 * @param max_sec Timeout for getting value
 */
int cmdman_statget(char *app_name, char *obj_name, char *stat_name,
	void *value, int *type, int *size, int max_sec)
{
	var_o var;

	assert(app_name && obj_name && stat_name);

	pkt_clear((pkt_o) packet_x);
	strncpy(pkt_putval((pkt_o) packet_x, FIELD_APPNAME, strlen(app_name) + 1), app_name,
			pkt_freesz((pkt_o) packet_x));
	strncpy(pkt_putval((pkt_o) packet_x, FIELD_OBJNAME, strlen(obj_name) + 1), obj_name,
			pkt_freesz((pkt_o) packet_x));
	strncpy(pkt_putval((pkt_o) packet_x, FIELD_STNAME, strlen(stat_name) + 1),
			stat_name, pkt_freesz((pkt_o) packet_x));
	daemon_sendto((daemon_o) daemon_x, STATSMAN_STGET, 2, DAEMON_STATSMAN);


	if (wait_answer_pkt(max_sec)) {
		if (pkt_getcmd((pkt_o) packet_x) == CMDMAN_STVALOK) {
			var = pkt_get((pkt_o) packet_x, FIELD_STAT, var_newfrompkt);
			if (value) {
				memcpy(value, var_getvalue(var), var_bsize(var));
			}
			if (type) {
				*type = var_type(var);
			}
			if (size) {
				*size = var_bsize(var);
			}
			return 1;
		}
	}
	return 0;
}

/** Set statistics variable value
 * @param app_name Name of the application
 * @param obj_name Name of the object who registered the stat
 * @param stat_name Name of the variable
 * @param value Pointer to a buffer where data will be read
 * @param size Size (in bytes) of the variable value
 * @param max_sec Timeout for setting value
 */
int cmdman_statset(char *app_name, char *obj_name, char *stat_name,
	void *value, int size, int max_sec)
{
	assert(app_name && obj_name && stat_name && value);

	pkt_clear((pkt_o) packet_x);

	strncpy(pkt_putval((pkt_o) packet_x, FIELD_APPNAME, strlen(app_name) + 1), app_name,
			pkt_freesz((pkt_o) packet_x));
	strncpy(pkt_putval((pkt_o) packet_x, FIELD_OBJNAME, strlen(obj_name) + 1), obj_name,
			pkt_freesz((pkt_o) packet_x));
	strncpy(pkt_putval((pkt_o) packet_x, FIELD_STNAME, strlen(stat_name) + 1),
			stat_name, pkt_freesz((pkt_o) packet_x));
	memcpy(pkt_putval((pkt_o) packet_x, FIELD_STVALUE, size), value, size);

	daemon_sendto((daemon_o) daemon_x, STATSMAN_STSET, 2, DAEMON_STATSMAN);

	if (wait_answer_pkt(max_sec)) {
		if (pkt_getcmd((pkt_o) packet_x) == CMDMAN_STSETOK) {
			return 1;
		}
	}
	return 0;
}

#define MAX_P	40

struct hwapi_proc_i procdb[MAX_P];
struct hwapi_cpu_i cpu;

int proc_print(int fd, struct hwapi_cpu_i cpu, struct hwapi_proc_i *procs, int nof_procs);
void hwinfo_print(int fd, struct hwapi_cpu_i *cpu_info);

void cmdman_hwcmd(char *cmd) {
	
	if (!strcmp(cmd, "ps")) {
		int n = hwapi_proc_list(procdb, MAX_P);
		hwapi_hwinfo_cpu(&cpu);
		proc_print(1, cpu, procdb, n);
	} else if (!strcmp(cmd, "info")) {
		hwapi_hwinfo_cpu(&cpu);
		hwinfo_print(1, &cpu);
	} else  {
		hwapi_hwcmd(cmd);
	} 
}



void print_to_fd(int fd, char *str) 
{
	write(fd,str,strlen(str)+1);
}

char pstr[80];




/** Print processes list
 * 
 * Prints processes to file descriptor as parameter
 * (set to 1 for stdout)
 * 
 */
int proc_print(int fd, struct hwapi_cpu_i cpu, struct hwapi_proc_i *procs, int nof_procs) 
{
	int i,n;
	char *t,*t2;
	
	sprintf(pstr, "\nPid\tName\t\tStatus\tCPU\t\t%% CPU\n");
	print_to_fd(fd, pstr);
	n=strlen(pstr)-3+5*8;
	for (i=0;i<n;i++)
		sprintf(&pstr[i], "=");
	sprintf(&pstr[i], "\n");
	print_to_fd(fd, pstr);
	for (i = 0; i < nof_procs; i++) {
		if (procs[i].obj_id) {
			if (strlen(procs[i].obj_name)<8)
				t="\t\t";
			else
				t="\t";
			sprintf(pstr, "%d us", procs[i].mean_cpu_usec);
			if (strlen(pstr)<8)
				t2="\t\t";
			else
				t2="\t";

			sprintf(pstr, "%d\t%s%s%d\t%d us%s%.2f%%\n", procs[i].obj_id, 
					procs[i].obj_name, t, 
					procs[i].status.cur_status,
					procs[i].mean_cpu_usec, t2, 
					(float) (100*procs[i].mean_cpu_usec/cpu.tslen_usec));
			print_to_fd(fd, pstr);
		}
	}
	for (i=0;i<n;i++)
		sprintf(&pstr[i], "=");
	sprintf(&pstr[i], "\n");
	
	return 1;
}


void hwinfo_print(int fd, struct hwapi_cpu_i *cpu_info) 
{
	sprintf(pstr, "\tPlatform MIPS = %.2f\n", cpu_info->C);
	print_to_fd(fd, pstr);
	sprintf(pstr, "\tTime slot length = %d usec\n", cpu_info->tslen_usec);
	print_to_fd(fd, pstr);
	sprintf(pstr, "\tDebug level = %d\n", cpu_info->debug_level);
	print_to_fd(fd, pstr);
	sprintf(pstr, "\tStatistics ");
	print_to_fd(fd, pstr);
	if (cpu_info->disable_stats) {
		sprintf(pstr, "disabled\n");
	} else {
		sprintf(pstr, "enabled\n");
	}
	print_to_fd(fd, pstr);
	sprintf(pstr, "\tLogs ");
	print_to_fd(fd, pstr);
	if (cpu_info->disable_logs) {
		sprintf(pstr, "disabled\n");
	} else {
		sprintf(pstr, "enabled\n");
	}
	print_to_fd(fd, pstr);
	sprintf(pstr, "\n\tFifo usage:\t%d Kb\t/ %d Kb\t(%.2f%%)\n", cpu_info->used_fifo/1024, cpu_info->total_fifo/1024, (float) (100*cpu_info->used_fifo/cpu_info->total_fifo));
	print_to_fd(fd, pstr);
	sprintf(pstr, "\n\tCPU usage:\t%d us\t/ %d ms\t(%.2f%%)\n", cpu_info->cpu_usage, cpu_info->tslen_usec/1000, (float) (100*cpu_info->cpu_usage/cpu_info->tslen_usec));
	print_to_fd(fd, pstr);
}

