#include <stdlib.h>
#include <string.h>

#include "phal_hw_api.h"

#include "phal_daemons.h"
#include "phid.h"

#include "set.h"
#include "pkt.h"
#include "daemon.h"

#include "frontend_cmds.h"

#define T daemon_o

#define DEFAULT_ITF_SZ	2048
#define WAIT_FE_SEC	5

#define PKT_BUFF_SZ	1024
int pkt_buff[PKT_BUFF_SZ];

struct cmd
{
	cmd_t cmd;
	int (*call)(T daemon);
};

struct T
{
	pkt_o pkt;
	int fd_r;
	int fd_w;
	dmid_t my_code;
	Set_o cmd_set;
};

int attach_itf_wait(itfid_t id, int mode, int timeout_sec);
int create_attach_itf(itfid_t id, int mode, int size);

/** Cmd destructor
 */
void cmd_delete(void **x)
{
	assert (x && *x);

	free(*x);
	x = NULL;
}

/** Find matching type x in member
 */
int cmd_findtype(const void * x, const void *member)
{
	assert (x && member);

	struct cmd *c = (struct cmd *) member;

	return c->cmd != *((cmd_t *) x);
}

/** Creates a new daemon object
 *
 * Create the daemon object, initializes hwapi and creates input and output
 * interfaces to FRONT-END daemon.
 *
 * \param daemon_code Code for the caller's daemon (see phal_daemons.h)
 *
 * \returns !=0 daemon object
 * \returns NULL if error
 */
T
daemon_new (dmid_t daemon_code)
{
	T daemon = NULL;
	int fd_r, fd_w;
	int input_itf_id;
	struct header *head = (struct header *) pkt_buff;

	assert (daemon_code);

	NEW (daemon);
	if (!daemon) {
		return NULL;
	}

	if (hwapi_init () < 0) {
		daemon_delete(&daemon);
		return NULL;
	}

	input_itf_id = DAEMON_INPUT_ITF_MASK | daemon_code;
	fd_r = utils_create_attach_itf (input_itf_id, FLOW_READ_ONLY, DEFAULT_ITF_SZ);
	if (fd_r < 0) {
		daemon_delete(&daemon);
		return NULL;
	}

	fd_w = utils_attach_itf_wait (FRONTEND_INPUT_ITF, FLOW_WRITE_ONLY, WAIT_FE_SEC);
	if (fd_w < 0) {
		xprintf ("Timeout while waiting to attach to front-end\n");
		daemon_delete(&daemon);
		return NULL;
	}

	/* send a packet to front-end indicating our input interface and daemon code */
	memset (pkt_buff, 0, HEAD_SZ + 2);
	head->cmd = FRONTEND_CFGITF;
	head->dst_daemon = DAEMON_FRONTEND;
	pkt_buff[HEAD_SZ] = daemon_code;
	pkt_buff[HEAD_SZ + 1] = input_itf_id;
	if (hwapi_itf_snd (fd_w, pkt_buff, 2 + HEAD_SZ) < 0) {
		xprintf ("Error sending through interface.\n");
		return NULL;
	}

	daemon->fd_r = fd_r;
	daemon->fd_w = fd_w;
	daemon->my_code=daemon_code;
	daemon->pkt = pkt_new (pkt_buff, PKT_BUFF_SZ<<2);
	daemon->cmd_set = Set_new (0,NULL);
	if (!daemon->pkt || !daemon->cmd_set)
	{
		daemon_delete(&daemon);
		return NULL;
	}
	return daemon;
}

/** Daemon destructor
 *
 */
void daemon_delete(T * daemon)
{
	assert (daemon);
	assert (*daemon);

	if ((*daemon)->cmd_set) {
		Set_destroy(&(*daemon)->cmd_set, cmd_delete);
	}

	if ((*daemon)->pkt) {
		pkt_delete(&(*daemon)->pkt);
	}

	free(*daemon);
	daemon = NULL;
}

/** Daemon packet
 *
 * Accesses in/out pkt buffer for daemon
 *
 * @param daemon daemon object
 * @returns packet object pointer
 */
pkt_o daemon_pkt(T daemon)
{
	assert (daemon);

	return daemon->pkt;
}

/** Add a command function to the selector
 *
 * This function adds a command to daemon object and associates it with a callback
 * function which will be called when processcmd() function is called and
 * the associated command type has been received.
 *
 * \param daemon Daemon object
 * \param type command type
 * \param call callback function to the processing function for this command
 *		type
 *
 * \returns 1 if ok
 * \returns 0 if error
 */
int daemon_addcmd(T daemon, cmd_t cmd, int call (T daemon))
{
	struct cmd *pcmd;

	assert (daemon && cmd && call);

	NEW(pcmd);
	if (!pcmd) {
		return 0;
	}
	pcmd->cmd=cmd;
	pcmd->call=call;

	Set_put(daemon->cmd_set, pcmd);

	return 1;
}

/** Receive packet
 *
 * This function is used to receive a packet through the input interface, created
 * during the creation of the daemon object.
 *
 * \param daemon Daemon object
 *
 * \returns >0 length of the packet received, if any
 * \returns 0 if no packet available
 * \returns -1 if error reading packet
 */
int daemon_rcvpkt(T daemon)
{
	int n;

	assert (daemon);

	n = hwapi_itf_rcv(daemon->fd_r, pkt_buff, PKT_BUFF_SZ);
	if (n < 0) {
		return -1;
	} else if (!n) {
		return 0;
	}

	if (!pkt_readvalues(daemon->pkt, n<<2)) {
		return 0;
	}
	return n;
}

/** Send Packet to a daemon and pe
 *
 * This function sends daemon packet (accessed through daemon_pkt() function)
 * through the output interface towards FRONT-END.
 * Note that input and output packets are the same, so, pkt_clear() function should
 * be called if we desire to erase its contents before sending it.
 *
 * It overwrittes type, destpe and destdaemon fields with pe & destdaemon params.
 *
 * \param daemon daemon object
 * \param type command type
 * \param pe destination Processor Id
 * \param destdaemon destination daemon code
 *
 * \returns same as sndpkt
 */
int daemon_sendto(T daemon, cmd_t cmd, peid_t pe, dmid_t destdaemon)
{
	int n;

	assert (daemon && cmd);

	pkt_setcmd(daemon->pkt, cmd);
	pkt_setdestpe(daemon->pkt, pe);
	pkt_setdestdaemon(daemon->pkt, destdaemon);

	do {
		n = hwapi_itf_snd(daemon->fd_w, pkt_buff,
						pkt_wlen(daemon->pkt));
		if (!n) {
			hwapi_relinquish_daemon();
		}
	} while (!n);

	return n;
}

/** Process command
 *
 * This function shall be called after daemon_rcvpkt() has returned something >0.
 * In that case, the callback selector will pick the function associated to the
 * received packet command type and call it.
 *
 * \param daemon daemon object
 *
 * \returns caller return code
 * \return -1 if command not found
 */
int daemon_processcmd(T daemon)
{
	struct cmd *pcmd;
	cmd_t c;

	assert (daemon);

	if (pkt_len(daemon->pkt)) {
		c = pkt_getcmd(daemon->pkt);

		/* find command type on command set */
		pcmd
				= (struct cmd*) Set_find(daemon->cmd_set, &c,
						cmd_findtype);
		if (!pcmd) {
			return -1;
		}

		assert (pcmd->call);
		return (pcmd->call)(daemon);
	} else {
		return -1;
	}
}

/**  Attaches to an interface.
 * In the case it has not been yet created (returns 0),
 * it waits for "timeout_sec" seconds.
 *
 * @param id id of the interface to attach
 * @param mode read/write
 * @param timeout_sec number of seconds to wait
 *
 * @return >0 if successfully, returns file descriptor
 * @return <0 if failed
 */
int utils_attach_itf_wait(itfid_t id, int mode, int timeout_sec)
{
	int n;
	time_t tdata[3];

	get_time(&tdata[1]);
	do {
		n = hwapi_itf_attach(id, mode);
		if (n == -1) {
			return -1;
		} else if (n == -2) {
			hwapi_relinquish_daemon();
			get_time(&tdata[2]);
			get_time_interval(tdata);
		}
	} while ((n == -2) && tdata[0].tv_sec < timeout_sec);

	if (n<1) {
		return -1;
	} else {
		return n;
	}
}

/** Creates an interface and attaches one side (depending on mode)
 *
 * @param id id of the interface to create
 * @param mode mode to attach the file descriptor
 * @param size size of the interface to create
 *
 * @return >0 if successfully, returns file descriptor
 * @return <0 if failed
 */
int utils_create_attach_itf(itfid_t id, int mode, int size)
{
	id = hwapi_itf_create(id, size);
	if (!id) {
		return -1;
	}
	return hwapi_itf_attach(id, mode);
}

int utils_create_attach_ext(int xitfid, int size, int *fd, int mode)
{
	int n;

	n = hwapi_itf_create (0, size);
	if (!n) {
		xprintf ("FRONT-END: Error creating itf\n");
		return 0;
	}
	if (hwapi_itf_link_phy(n, xitfid, mode) < 0) {
		xprintf ("FRONT-END: Error linking with external itf\n");
		return 0;
	}
	if (fd) {
		*fd = hwapi_itf_attach(n, mode);
		if (*fd < 0) {
			xprintf ("FRONT-END: Error attaching itf\n");
			return 0;
		}
	}
	return 1;
}

int utils_create_attach_ext_bi(int xitfid, int size, int *fd_w, int *fd_r)
{
	assert (xitfid && fd_w && fd_r);

	if (!utils_create_attach_ext (xitfid, size, fd_r, FLOW_READ_ONLY)) {
		return 0;
	}

	if (!utils_create_attach_ext (xitfid, size, fd_w, FLOW_WRITE_ONLY)) {
		return 0;
	}

	return 1;
}

