#include <stdlib.h>
#include <string.h>
#include <stdio.h>

#include "phal_hw_api.h"

#include "phid.h"
#include "stats.h"

#include "str.h"
#include "var.h"

/* public type */
#define T var_o

/* private type */
#define V var_x

typedef struct V *V;

struct V
{
	/* public members */
	str_o name;
	varid_t stat_id;
	/* private members */
	varsz_t size;
	vartype_t type;
	int elem_sz;
	void *value;
};

struct topkt
{
	/* public members */
	varid_t stat_id;
	/* private members */
	varsz_t size;
	vartype_t type;
};

/** Constructor
 *
 * @param size Size of the variable in elements
 * @param type Type of the variable
 *
 * @return !=0 pointer to the object
 * @return =0 Error
 *
 * @todo TEMA SIZES VARIABLES!! ANEM A WORDS O BYTS O QUE??
 * @todo Just Integers supported, type is ignored
 */
T
var_new (varsz_t size, vartype_t type)
{

	V var;
	int elem_sz = 0;

	NEW (var);
	if (!var)
	{
		return NULL;
	}

	var->size = size;
	var->type = type;

	switch (type)
	{
		case STAT_TYPE_CHAR:
		case STAT_TYPE_UCHAR:
		elem_sz = sizeof (char);
		break;
		case STAT_TYPE_FLOAT:
		elem_sz = sizeof (float);
		break;
		case STAT_TYPE_INT:
		elem_sz = sizeof (int);
		break;
		case STAT_TYPE_SHORT:
		elem_sz = sizeof (short);
		break;
		default:
		xprintf ("Unknown Var type %d\n", type);
		var_delete ((T*) &var);
		return NULL;
	}

	var->elem_sz = elem_sz;
	var->value = NULL; /* will be allocated when used */

	var->name = str_new ();
	if (!var->name) {
		var_delete ((T*) &var);
		return NULL;
	}

	return (T) var;
}

/** Destructor
 * 
 * Deallocates also the value and name pointers
 * 
 */
void var_delete(T * x)
{
	V * var = (V*) x;

	assert (var);
	assert (*var);

	if ((*var)->name) {
		str_delete(&(*var)->name);
	}
	if ((*var)->value) {
		free((*var)->value);
	}

	free(*var);
	var = NULL;
}

/** Get variable type
 */
vartype_t var_type (T x)
{
	V var = (V) x;
	assert (var);

	return var->type;
}

int var_bsize(T x)
{
	V var = (V) x;

	assert (x);

	return var->size*var->elem_sz;
}

/** Modify variable value 
 *
 * Modifies the stored variable value with the given value.
 * User must indicate new value size although if it is bigger
 * than the previously defined one, this excess will not be stored.
 *
 * @param var Variable object pointer
 * @param value Pointer to new value contentes
 * @param size Size of the new value to store
 *
 * @return >=0 Number of bytes written
 * @return -1 Error
 *
 */
varsz_t
var_setvalue (T x, void *value, varsz_t size)
{
	V var = (V) x;

	assert (var && value && size);

	if (!var->value) {
		var->value = malloc (var->elem_sz * var->size);
		if (!var->value) {
			return 0;
		}

	}

	/* check sizes */
	if (size> var->size) {
		size = var->size;
	}

	/* copy data */
	memcpy (var->value, value, size * var->elem_sz);

	return size;
}

/** Get variable value 
 *
 * Obtains the stored variable value address.
 * 
 * @warning Use this function with caution because there
 * is no control with sizes.
 */
void * var_getvalue(T x)
{
	V var = (V) x;

	assert (var);

	return var->value;
}

int var_writetohwapi(T x)
{
	V var = (V) x;

	assert (var);

	if (!var->value || !var->size) {
		return 0;
	}

	if (hwapi_var_write(var->stat_id, var->value, var->size * var->elem_sz)
			> 0) {
		return 1;
	} else {
		return 0;
	}
}

int var_readfromhwapi(T x)
{
	V var = (V) x;

	assert (var);

	if (!var->size) {
		return 0;
	}

	if (!var->value) {
		var->value = malloc(var->elem_sz * var->size);
		if (!var->value) {
			return 0;
		}

	}

	if (hwapi_var_read(var->stat_id, var->value, var->size * var->elem_sz)
			> 0) {
		return 1;
	} else {
		return 0;
	}
}

void var_printstr(T x, char *str)
{
	V var = (V) x;
	int i;
	char comma[] = ", ";
	char space[] = " ";
	char *sep;
	char *data_c;
	float *data_f;
	int *data_i;
	short *data_s;

	assert (var && str);

	if (!var->value) {
		return;
	}

	if (var->type == STAT_TYPE_CHAR && var->size > 1) {
		sprintf(str, "%s", (char *) var->value);
		return;
	}

	if (var->size > 1) {
		sep = comma;
	} else {
		sep = space;
	}

	data_c = (char *) var->value;
	data_f = (float *) var->value;
	data_i = (int *) var->value;
	data_s = (short *) var->value;

	for (i = 0; i < var->size; i++) {
		switch (var->type) {
		case STAT_TYPE_CHAR:
			sprintf(&str[i], "%c%s", data_c[i], sep);
			break;
		case STAT_TYPE_UCHAR:
			sprintf(&str[i], "%d%s", (unsigned char) data_c[i], sep);
		case STAT_TYPE_FLOAT:
			sprintf(&str[i], "%f.2%s", data_f[i], sep);
			break;
		case STAT_TYPE_INT:
			sprintf(&str[i], "%d%s", data_i[i], sep);
			break;
		case STAT_TYPE_SHORT:
			sprintf(&str[i], "%d%s", data_s[i], sep);
			break;
		}
	}
}

/** Duplicate variable
 *
 * Creates a new variable with the same contents of the given in params
 */
T
var_dup (T x)
{
	V var = (V) x;

	V newvar;

	assert (var);

	newvar = (V) var_new (var->size, var->type);

	newvar->stat_id = var->stat_id;

	newvar->name = str_dup (var->name);

	if (var->value)
	{
		newvar->value = malloc (newvar->size * newvar->elem_sz);

		if (!newvar->value)
		{
			var_delete ((T*) &newvar);

			return NULL;
		}

		memcpy (newvar->value, var->value, var->size * newvar->elem_sz);
	}

	return (void*) newvar;
}

/** Size of object structure when writting to flat array
 *
 */
int var_sizeof(T x)
{
	V var = (V) x;

	int size;

	assert (var);

	size = sizeof(struct topkt) + str_sizeof(var->name) + var->size
			* var->elem_sz;

	return size;
}

/** New Variable from Raw Data.
 *
 * Creates a new variable and fills its contents with flat aligned data
 * generated by the var_getdata() function.
 *
 */
void * var_newfrompkt(char **start, char *end)
{
	struct topkt *pkt;

	V var;

	assert (start && end && (end >= *start));

	pkt = (struct topkt *) *start;

	var = (V) var_new(pkt->size, pkt->type);
	if (!var) {
		return NULL;
	}

	var->stat_id = pkt->stat_id;
	*start += sizeof(struct topkt);

	var->name = str_newfrompkt(start, end);
	if (!var->name) {
		var_delete((T*) &var);
		return NULL;
	}

	/* 1 means there is value, 0 means there isn't */
	if (**start) {
		(*start)++;
		var_setvalue((T) var, *start, var->size * var->elem_sz);
		*start += var->size * var->elem_sz;
	} else {
		(*start)++;
	}

	return (void *) var;
}

/** Get Variable Raw Data.
 *
 * Copies all structure to a flat array of data. It can be reversed
 * with the newfromdata function
 */
int var_topkt(void *x, char **start, char *end)
{
	V var = (V) x;

	assert (var && start && end);
	assert (end >= *start);

	if ((end - *start) < var_sizeof((T) var)) {
		return 0;
	}

	memcpy(*start, &var->stat_id, sizeof(struct topkt));
	*start += sizeof(struct topkt);

	if (!str_topkt(var->name, start, end)) {
		return 0;
	}

	/* 1 means there is value, 0 means there isn't */
	if (var->value) {
		**start = 1;
		(*start)++;
		memcpy(*start, var->value, var->size * var->elem_sz);
		*start += var->size * var->elem_sz;
	} else {
		**start = 0;
		(*start)++;
	}

	return 1;
}

int var_findid(const void *x, const void *member)
{
	V var = (V) member;

	return *((varid_t *) x) != var->stat_id;
}

int var_findname(const void *x, const void *member)
{
	V var = (V) member;

	return str_scmp(var->name, (char *) x);
}

void var_xdelete(void **x)
{
	return var_delete((T*) x);
}
int var_xsizeof(const void *x)
{
	return var_sizeof((T) x);
}

void * var_xdup(const void *x)
{
	return (void *) var_dup((T) x);
}
