/** @file exec.c
 *
 * @brief EXEC SW Daemon Implementation
 *
 * - Daemon type: Sensor
 *
 * - Communicates with: SW MANAGER
 *
 * - Functions: This sensor has two main tasks:
 *	- Change object execution status under SWMANAGER command.
 *		- These status can be:
 *			-# INIT
 *			-# RUN
 *			-# PAUSE
 *			-# STOP
 *	- Monitor the correct execution of the objects depending on
 *		their status. If an object is in an incorrect state or 
 *		under incorrect conditions (e.g. time, etc.) a report
 *		is issued to SWMANAGER.
 */

/* standard libraries */
#include <stdlib.h>
#include <string.h>

#include "phal_hw_api.h"

#include "phal_daemons.h"
#include "phid.h"

#include "pkt.h"
#include "daemon.h"

#include "exec.h"

/* include command declarations */
#include "exec_cmds.h"

/* main function declaration */
void exec_run();

#ifdef RUN_AS_PROC
int main()
{
	exec_run();
}
#endif

/** EXEC Daemon Main Routine
 */
void exec_run()
{
	daemon_o daemon;

	int n;

	daemon = daemon_new(DAEMON_EXEC);
	if (!daemon) {
		xprintf ("swman: Error creating daemon\n");
		exit(-1);
	}

	/** Configure input commands
	 */

	/* Change an object status  */
	if (daemon_addcmd(daemon, EXEC_SETSTATUS, exec_incmd_setstatus) < 0) {
		exit(-1);
	}
	if (daemon_addcmd(daemon, EXEC_STARTREPORT, exec_incmd_startreport) < 0) {
		exit(-1);
	}

	init_state_str();

	xprintf ("EXEC: Init with PID=%d PPID=%d\n", getpid(), getppid());

	/**  main daemon loop 
	 * while packets in queue, process them
	 */
	while (1) {

		do {
			n = daemon_rcvpkt(daemon);
			if (n > 0) {
				daemon_processcmd(daemon);
			}
		} while (n);
		if (n < 0) {
			break;
		}

		/** After all commands are processed, we will execute the 
		 * real-time and execution status monitoring
		 */
		exec_background(daemon);

		hwapi_relinquish_daemon();
	}

	daemon_delete(&daemon);

	exit(-1);
}
