/**
 *
 *
 *	PHAL Exec Daemon Implementation
 *
 * FILE: exec_cmds.c
 *
 * Description: Here is where functions answering COMMANDS are
 *	implemented
 *
 */

#include <sys/types.h>
#include <sys/stat.h>
#include <sys/msg.h>
#include <sys/ipc.h>
#include <sys/shm.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <sys/wait.h>
#include <sys/time.h>
#include <time.h>
#include <errno.h>
#include <unistd.h>
#include <fcntl.h>



/* standard libraries */
#include <stdlib.h>
#include <string.h>

#include "phal_hw_api.h"
/* common objects */
#include "phid.h"
#include "phal_daemons.h"

#include "pkt.h"
#include "daemon.h"

#include "exec.h"

/* commands */
#include "exec_cmds.h"
#include "swman_cmds.h"


#define NEW_STATUS_TOUT	20

/** maximum objects to control
 */
#define MAX_OBJECTS		50

/** structure for gathering object execution information
 */
struct hwapi_proc_i hwapi_obj[MAX_OBJECTS];


//#define DEB

#define REPORT_WINDOW	100	// in timeslots
#define REPORT_PROCS	MAX_OBJECTS

struct report_data 
{
	unsigned int cpu_usage;
	unsigned int period;
	int tstamp;
	char executed;
};
struct report_obj
{
	int fd;
	int obj_id;
	int report_idx;
	int last_tstamp;
	struct report_data report[REPORT_WINDOW];
};

struct report_obj reports[REPORT_PROCS];

int enable_reports = 0;

char line[64];

void report_exec_stats(struct hwapi_proc_i *procs, int nof_procs)
{
	int tstamp;
	int i,j;
	
	tstamp = get_tstamp();
	
	for (i=0;i<nof_procs;i++) {
		if (procs[i].status.cur_status == PHAL_STATUS_RUN) {
			if (tstamp == reports[i].last_tstamp + 1
					&& procs[i].relinquished) {
				reports[i].report[reports[i].report_idx].cpu_usage = procs[i].cpu_time[0].tv_usec;
				reports[i].report[reports[i].report_idx].period = procs[i].tti;
				reports[i].report[reports[i].report_idx].executed = 1;
				reports[i].report[reports[i].report_idx].tstamp = tstamp;
				reports[i].last_tstamp = tstamp;
				reports[i].report_idx++;
			} else if (tstamp > reports[i].last_tstamp + 1) {
				reports[i].report[reports[i].report_idx].executed = 0;
				reports[i].last_tstamp = tstamp;
				reports[i].report_idx++;
			}
			if (reports[i].report_idx == REPORT_WINDOW) {
				if (reports[i].fd>1) {
					for (j=0;j<REPORT_WINDOW;j++) {
						sprintf(line, "%d, %d, %d, %d\n",
								reports[i].report[i].tstamp,
								reports[i].report[j].cpu_usage,
								reports[i].report[j].period,
								reports[i].report[j].executed);
						write(reports[i].fd, line, strlen(line));
					}
					reports[i].report_idx=0;
				}
			}
		}
	}
}

char file[64];

int exec_incmd_startreport(daemon_o daemon)
{
	int n, i, fd;
	
	memset(reports, 0, sizeof(struct report_obj)*REPORT_PROCS);
	
	n = hwapi_proc_list(hwapi_obj, MAX_OBJECTS);
	if (n < 0) {
		xprintf ("EXEC: Fatal error while getting object list\n");
		exit(-1);
	}

	for (i=0;i<n;i++) {
		sprintf(file, "example-work/execlogs/%s.log", hwapi_obj[i].obj_name);
		fd = open(file, O_CREAT | O_APPEND | O_WRONLY, S_IRUSR | S_IWUSR);
		if (fd<0) {
			xprintf("EXEC: Error creating file %s\n", file);
			perror("open");
			return;
		}
		reports[i].obj_id = hwapi_obj[i].obj_id;
		reports[i].fd = fd;
		reports[i].report_idx=0;
	}
	
	enable_reports = 1;
}


/** Processing Function
 *
 * REQUIRED PACKET FIELDS:
 *	- FIELD_OBJID: Id of the object to change the status.
 *	- FIELD_STATUS: New status for the object.
 *
 * ACTIONS:
 *	- Uses hw_api functions to set new status.
 *	- The previous function will return -1 if can't change stauts of the object because
 * either it does not exists or it is already in a change procedure.
 *
 * OUTPUT:
 *	- SWMAN_STATUSOK command if successfull
 *	- SWMAN_STATUSERR command if error
 */
int exec_incmd_setstatus(daemon_o daemon)
{
	pkt_o pkt;
	status_t status;
	objid_t obj_id;
	int c, n;
	int tstamp;
	struct hwapi_proc_i proc;
	struct hwapi_cpu_i cpu;

	assert (daemon);

	pkt = daemon_pkt(daemon);

	status = (status_t) pkt_getvalue(pkt, FIELD_STATUS);
	tstamp = (int) pkt_getvalue(pkt, FIELD_TSTAMP);
	obj_id = (objid_t) pkt_getvalue(pkt, FIELD_OBJID);

#ifdef DEB
	xprintf ("EXEC: Received status %d for object 0x%x\n",
			status, obj_id);
#endif

	if (!hwapi_proc_info (obj_id, &proc)) {
		xprintf ("EXEC: Error unknown object 0x%x\n", obj_id);
		return 0;
	}
	
	hwapi_hwinfo_cpu(&cpu);


	/* if received abnormal stop */
	if (status == PHAL_STATUS_ABNSTOP) {
		tstamp = get_tstamp() - 1;
		xprintf ("EXEC: Abnormal STOP for object 0x%x\n", obj_id);
	}


	/* wait it finishes it status change, if the case */
	if (proc.change_progress) {
		c = 0;
		do {
			c++;
			n = hwapi_proc_status_ack (obj_id);
			if (!n) {
				hwapi_relinquish_daemon();
			}
		} while (c < NEW_STATUS_TOUT && !n && proc.pid);
		if (!n) {
			/* ensure it didn't die */
			if (!proc.pid) {
				xprintf ("EXEC: Timeout while changing status for object 0x%x. Removing process...\n", obj_id);
			}
			if (!cpu.debug_level) {
				exec_statusrep(daemon, &proc, 0);
				hwapi_proc_remove (obj_id);
			}
			return 0;
		}
	}
	if (!hwapi_proc_status_new(obj_id, status, tstamp)) {
		xprintf ("EXEC: Error setting status for object 0x%x\n", obj_id);
	}


	/* *a successful change will be reported by exec_run() function when it detects it has been correctly
	 realized */
	return 1;
}

/** @defgroup sensors_exec EXEC Sensor Monitor Function
 *
 * EXEC Sensor, after processing all possible input commands, has the tasks to monitor
 * the object execution status and report changes or fault to manger (SW MANAGER).
 *
 * Status succesfull or error change report is executed by exec_statusrep function.
 *
 * @ingroup sensors
 *
 * @{
 */

/** structure for displaying names instead of status numbers
 * @todo Definirla com a constant
 */
struct st_db
{
	int max_ts;
	char name[16];
};
struct st_db st_db[PHAL_STATUS_MAX];

void init_state_str();


/** Exec Status Report Function
 *
 * This function build a SWMAN_STATUSOK or SWMAN_STATUSERR commands and sends it. It indicates the successfull or
 * error in an status change procedure for an object
 * @todo appid
 *
 * @param daemon Pointer to dameon object
 * @param obj Pointer to hwapi_proc_i structure of the object
 * @param ok 1 for STATUSOK 0 for STATUSERR
 *
 * @return 1 if send ok
 * @return -1 if send error
 */
int exec_statusrep(daemon_o daemon, struct hwapi_proc_i *obj, int ok)
{
	pkt_o pkt;
	assert (daemon && obj);
	pkt = daemon_pkt(daemon);
	pkt_clear(pkt);

	pkt_putvalue(pkt, FIELD_APPID, (uint) obj->app_id);
	pkt_putvalue(pkt, FIELD_OBJID, (uint) obj->obj_id);
	pkt_putvalue(pkt, FIELD_STATUS, (uint) obj->status.cur_status);

	daemon_sendto(daemon, ok ? SWMAN_STATUSOK : SWMAN_STATUSERR, 0,
			DAEMON_SWMAN);

#ifdef DEB
	xprintf ("EXEC: Sent status %d for object 0x%x\n", obj->status.cur_status,
			obj->obj_id);
#endif
	return 1;
}

int last_msg_ts=0;

/** Exec Background Function
 *
 * This function implements the monitoring of the execution status of the objects.
 *
 * @todo Document this funciton
 * @todo Rewrite this function
 */
void exec_background(daemon_o daemon)
{
	int i, max_ts, n;
	char *str_state;
	int tstamp;
	int c;
	assert (daemon);
	
	struct hwapi_cpu_i cpu;
	
	hwapi_hwinfo_cpu(&cpu);

	tstamp = get_tstamp();
	
	n = hwapi_proc_list(hwapi_obj, MAX_OBJECTS);
	if (n < 0) {
		xprintf ("EXEC: Fatal error while getting object list\n");
		exit(-1);
	}

	if (enable_reports) {
		report_exec_stats(hwapi_obj, n);
	}

	/** for all running objects
	 */
	for (i = 0; i < n; i++) {
		if (hwapi_obj[i].obj_id) {
			/** if object died, report it
			 */
			if (!cpu.debug_level) {
				if (!hwapi_obj[i].pid && hwapi_obj[i].status.cur_status) {
					if (hwapi_obj[i].status.cur_status < PHAL_STATUS_STOP) {
						xprintf ("EXEC: Object 0x%x died unexpectectly\n", hwapi_obj[i].obj_id);
						//xprintf ("report died 0x%x\n", hwapi_obj[i].obj_id);
		
						hwapi_obj[i].status.cur_status = PHAL_STATUS_ABNSTOP;
						exec_statusrep(daemon, &hwapi_obj[i], 0);
					}
					hwapi_proc_remove(hwapi_obj[i].obj_id);
					break;
				}
			}
			/** if object is in change procedure
			 */
			if (hwapi_obj[i].change_progress && hwapi_obj[i].status.cur_status != PHAL_STATUS_STEP) {
				max_ts = st_db[hwapi_obj[i].status.next_status].max_ts;
				str_state = st_db[hwapi_obj[i].status.next_status].name;

				/** If current status has been updated */
				if (hwapi_proc_status_ack(hwapi_obj[i].obj_id)) {
					//xprintf ("report change_progress 0x%x\n", hwapi_obj[i].obj_id);
					exec_statusrep(daemon, &hwapi_obj[i], 1);
				}
				/* if not yet changed */
				else {
					/** check if maximum timeslot allowed in the change procedure did elapsed
					 */
					if (tstamp > hwapi_obj[i].status.next_tstamp
						+ st_db[hwapi_obj[i].status.next_status].max_ts) {
						
						if (!cpu.debug_level) {
							xprintf ("EXEC: Caution object %s did not changed to %s after %d tslots (%d!=%d)\n",
								hwapi_obj[i].obj_name, str_state,
								tstamp - hwapi_obj[i].status.next_tstamp,
								hwapi_obj[i].status.cur_status,
								hwapi_obj[i].status.next_status);
							

							/** notify to manager object did not changed its status
							 */
							//xprintf ("report not changed 0x%x\n", hwapi_obj[i].obj_id);
							exec_statusrep (daemon, &hwapi_obj[i], 0);

							/** @todo What to do if an object can't change its status
							 */

							/* remove process */

							xprintf ("EXEC: Removing process 0x%x...\n", hwapi_obj[i].obj_id);
							hwapi_proc_remove (hwapi_obj[i].obj_id);
						}
					}
				}
			}
			/** if not in change operation, that is, normal operation
			 */
			else {
				max_ts = st_db[hwapi_obj[i].status.cur_status].max_ts;
				str_state = st_db[hwapi_obj[i].status.cur_status].name;

				/** @todo enable multiple slot execution time
				 */
				/** check that their execution cycle did not violated max_ts for the current status
				 */
				if (tstamp - hwapi_obj[i].rel_tstamp > max_ts
						&& tstamp > last_msg_ts + 10) {

#ifdef PRINT_RTCFAULT
					last_msg_ts = tstamp;
					xprintf
					("\nEXEC: CAUTION REAL-TIME VIOLATION at %s. Object %s. Cur TS=%d Last TS=%d. Cur ST=%d. Next ST=%d. Max TS=%d\n\n",
							str_state, hwapi_obj[i].obj_name, tstamp,
							hwapi_obj[i].rel_tstamp, hwapi_obj[i].status.cur_status,
							hwapi_obj[i].status.next_status, max_ts);
#endif
					/** @todo What to do when object fails its RTC Status */
#ifdef REPORT_RTCFAULT
					/*exec_statusrtcfault(daemon, &hwapi_obj[i]); */
#endif
				}
				/** @todo strict real-time monitor
				 * @todo cpu time monitor
				 */
			}
		}
	}

}

/** @} */
void init_state_str()
{
	int i;

	/* configure status names */
	memset(st_db, 0, sizeof(struct st_db));

	for (i = 0; i < PHAL_STATUS_MAX; i++) {
		switch (i) {
		case PHAL_STATUS_STARTED:
			st_db[i].max_ts = LOAD_LIMIT_TS;
			sprintf(st_db[i].name, "LOAD");
			break;
		case PHAL_STATUS_INIT:
			st_db[i].max_ts = INIT_LIMIT_TS;
			sprintf(st_db[i].name, "INIT");
			break;
		case PHAL_STATUS_RUN:
			st_db[i].max_ts = RUN_LIMIT_TS;
			sprintf(st_db[i].name, "RUN");
			break;
		case PHAL_STATUS_PAUSE:
			st_db[i].max_ts = PAUSE_LIMIT_TS;
			sprintf(st_db[i].name, "PAUSE");
			break;
		case PHAL_STATUS_ABNSTOP:
		case PHAL_STATUS_STOP:
			st_db[i].max_ts = STOP_LIMIT_TS;
			sprintf(st_db[i].name, "STOP");
			break;
		case PHAL_STATUS_STEP:
			st_db[i].max_ts = STEP_LIMIT_TS;
			sprintf(st_db[i].name, "STEP");
			break;
		}
	}

}
