/**@file stats_cmds.c
 * Description: Here is where functions answering COMMANDS are 
 * implemented
 */

/* standard libraries */
#include <stdlib.h>
#include <string.h>

/* phal hwlibrary constants */
#include "phal_hw_api.h"

#include "phal_daemons.h"
#include "phid.h"

/* common objects */
#include "set.h"
#include "str.h"
#include "var.h"
#include "rtcobj.h"
#include "phobj.h"
#include "pkt.h"
#include "log.h"
#include "daemon.h"

#include "stats.h"

/* commands */
#include "stats_cmds.h"
#include "statsman_cmds.h"

/** Processing Function
 *
 * REQUIRED PACKET FIELDS:
 *	- FIELD_STID: (for STAT INIT) Id of the stat for which the init request was done
 *	- FIELD_PMID: (for PARAM INIT) Id of the global variable used by SERVICES to request
 * the initialization procedure (see function implementation stats_incmd_initack at stats_cmds.c)
 *
 * ACTIONS:
 *	- Uses hw_api to set the global variable option to the appropiate constant. This way, the 
 * services API knows the stat/parameters have been successfully initiated and are ready to be used.
 *
 * OUTPUT: (no output)
 */
int stats_incmd_initack(daemon_o daemon)
{

	pkt_o pkt;
	assert(daemon);

	pkt = daemon_pkt(daemon);

	switch (pkt_getcmd(pkt)) {
	case STATS_STINITOK:
		if (hwapi_var_setopt((varid_t) pkt_getvalue(pkt, FIELD_STID),
				VAR_IS_STOK) < 0) {
			xprintf("STATS: Error setting stat option\n");
			return -1;
		}

		break;
	case STATS_STINITERR:
		if (hwapi_var_setopt((varid_t) pkt_getvalue(pkt, FIELD_STID),
				VAR_IS_STERR) < 0) {
			xprintf("STATS: Error setting stat option\n");
			return -1;
		}
		break;
	case STATS_PMINITOK:
		if (hwapi_var_setopt((varid_t) pkt_getvalue(pkt, FIELD_PMID),
				VAR_IS_PMINITOK) < 0) {
			xprintf("STATS: Error setting stat option\n");
			return -1;
		}
		break;
	case STATS_PMINITERR:
		/* set variable as error initiating */
		if (hwapi_var_setopt((varid_t) pkt_getvalue(pkt, FIELD_PMID),
				VAR_IS_PMINITERR) < 0) {
			xprintf("STATS: Error setting stat option\n");
			return -1;
		}
		break;
	case STATS_LOGINITOK:
		if (hwapi_var_setopt((varid_t) pkt_getvalue(pkt, FIELD_LOGID),
				VAR_IS_LOGINITOK) < 0) {
			xprintf("STATS: Error setting stat option\n");
			return -1;
		}
		break;
	case STATS_LOGINITERR:
		/* set variable as error initiating */
		if (hwapi_var_setopt((varid_t) pkt_getvalue(pkt, FIELD_LOGID),
				VAR_IS_LOGINITERR) < 0) {
			xprintf("STATS: Error setting stat option\n");
			return -1;
		}
		break;
	default:
		xprintf("error\n");
	}

	return 1;
}

/** Processing Function
 *
 * REQUIRED PACKET FIELDS:
 *	- FIELD_STID: Id of the stat to view/modify
 *	- FIELD_STTYPE: (for SET) Type of the stat
 *	- FIELD_STVALUE: (for SET) Data Value
 *
 * ACTIONS:
 *	- Uses hw_api functions to set or get the variable value
 *
 * OUTPUT: 
 *	- STATSMAN_STSETOK command if value has been changed successfully
 *	- STATSMAN_STSETERR command if impossible to change value
 *
 *	- STATSMAN_STVALOK command if value was obtained successfully
 *	- STATSMAN_STVALERR command if impossible to get value
 */
int stats_incmd_stact(daemon_o daemon)
{
	var_o var;
	pkt_o pkt;
	objid_t obj_id;

	assert(daemon);

	pkt = daemon_pkt(daemon);

	var = pkt_get(pkt, FIELD_STAT, var_newfrompkt);
	if (!var) {
		xprintf("STATS: Error reading stat from packet\n");
		return 0;
	}

	obj_id = (objid_t) pkt_getvalue(pkt, FIELD_OBJID);
	pkt_clear(pkt);

	pkt_putvalue(pkt, FIELD_OBJID, obj_id);
	pkt_putvalue(pkt, FIELD_STID, (uint) var->id);

	if (pkt_getcmd(pkt) == STATS_STGET) {
		if (!var_readfromhwapi(var)) {
			xprintf("STATS: Error getting stats variable\n");
			daemon_sendto(daemon, STATSMAN_STVALERR, 0,
					DAEMON_STATSMAN);
		} else {
			memcpy(pkt_putval(pkt, FIELD_STVALUE, var_bsize(var)),
					var_getvalue(var), var_bsize(var));

			daemon_sendto(daemon, STATSMAN_STVALOK, 0,
					DAEMON_STATSMAN);
		}
	} else if (pkt_getcmd(pkt) == STATS_STSET) {
		if (!var_writetohwapi(var)) {
			xprintf("STATS: Error setting stats variable\n");
			daemon_sendto(daemon, STATSMAN_STSETERR, 0,
					DAEMON_STATSMAN);
		} else {
			daemon_sendto(daemon, STATSMAN_STSETOK, 0,
					DAEMON_STATSMAN);
		}
	} else {
		xprintf("STATS: Unknown command\n");
	}

	var_delete(&var);

	return 1;
}

/** Processing Function
 *
 * REQUIRED PACKET FIELDS:
 *	- FIELD_PMID: Id of the global variable used by SERVICES to store the value of the parameter.
 *	- FIELD_PMTYPE: Type of the parameter
 *	- FIELD_PMVALUE: Value
 *
 * ACTIONS:
 *	- Uses hw_api to set the global variable value to the received one .
 *
 * OUTPUT: (no output)
 */
int stats_incmd_pmval(daemon_o daemon)
{

	pkt_o pkt;
	var_o var;
	varid_t pmid;

	assert(daemon);

	pkt = daemon_pkt(daemon);

	
	if (pkt_getcmd(pkt) == STATS_PMVALOK) {
		var = pkt_get(pkt, FIELD_PARAM, var_newfrompkt);
		if (!var) {
			xprintf("STATS: Error reading variable\n");
			return -1;
		}

		if (!var_writetohwapi(var)) {
			xprintf("STATS: Error setting stat\n");
		}
		if (hwapi_var_setopt(var->id, VAR_IS_PMVALOK) < 0) {
			xprintf("STATS: Error setting stat opt\n");
		}
		var_delete(&var);
	} else if (pkt_getcmd(pkt) == STATS_PMVALERR) {
		if (hwapi_var_setopt((varid_t) pkt_getvalue(pkt, FIELD_PMID), VAR_IS_PMVALERR) < 0) {
			xprintf("STATS: Error setting stat opt\n");
		}
	} else {
		xprintf("STATS: Error in command\n");
	}	

	return 1;
}

/** @defgroup sensors_stats STATS Sensor Background Functions
 *
 * The STATS Sensor monitors object's requests for registering an statistics
 * variable or requesting a parameter value.
 *
 * @ingroup sensors
 *
 * @{
 */

/** Maximum variables of new creation to process
 */
#define VAR_LIST_SZ		30

/** HWAPI structure for keeping information about variables 
 */
struct hwapi_var_i var_list[VAR_LIST_SZ];

/** STATS Background Processign Function
 *
 * @todo Comment this function
 */
void stats_background(daemon_o daemon)
{
	int n, i;

	assert(daemon);

	n = hwapi_var_list(VAR_IS_STINIT, var_list, VAR_LIST_SZ);
	if (n < 0) {
		xprintf("fatal error\n");
		exit(-1);
	} else {
		for (i = 0; i < n; i++) {
			stats_stinit(daemon, &var_list[i]);
		}
	}

	n = hwapi_var_list(VAR_IS_STCLOSE, var_list, VAR_LIST_SZ);
	if (n < 0) {
		xprintf("fatal error\n");
		exit(-1);
	} else {
		for (i = 0; i < n; i++) {
			stats_stclose(daemon, &var_list[i]);
		}
	}

	n = hwapi_var_list(VAR_IS_PMINIT, var_list, VAR_LIST_SZ);
	if (n < 0) {
		xprintf("fatal error\n");
		exit(-1);
	} else {
		for (i = 0; i < n; i++) {
			stats_pminit(daemon, &var_list[i]);
		}
	}

	n = hwapi_var_list(VAR_IS_PMCLOSE, var_list, VAR_LIST_SZ);
	if (n < 0) {
		xprintf("fatal error\n");
		exit(-1);
	} else {
		for (i = 0; i < n; i++) {
			stats_pmclose(daemon, &var_list[i]);
		}
	}

	n = hwapi_var_list(VAR_IS_PMGET, var_list, VAR_LIST_SZ);
	if (n < 0) {
		xprintf("fatal error\n");
		exit(-1);
	} else {
		for (i = 0; i < n; i++) {
			stats_pmget(daemon, &var_list[i]);
		}
	}
	
	
	n = hwapi_var_list(VAR_IS_LOGINIT, var_list, VAR_LIST_SZ);
	if (n < 0) {
		xprintf("fatal error\n");
		exit(-1);
	} else {
		for (i = 0; i < n; i++) {
			stats_loginit(daemon, &var_list[i]);
		}
	}

	n = hwapi_var_list(VAR_IS_LOGCLOSE, var_list, VAR_LIST_SZ);
	if (n < 0) {
		xprintf("fatal error\n");
		exit(-1);
	} else {
		for (i = 0; i < n; i++) {
			stats_logclose(daemon, &var_list[i]);
		}
	}
	
	n = hwapi_var_list(VAR_IS_LOGWRITE, var_list, VAR_LIST_SZ);
	if (n < 0) {
		xprintf("fatal error\n");
		exit(-1);
	} else {
		for (i = 0; i < n; i++) {
			stats_logwrite(daemon, &var_list[i]);
		}
	}
}

int sizeoftype(int type)
{
	switch (type) {
	case STAT_TYPE_CHAR:
	case STAT_TYPE_UCHAR:
		return sizeof(char);
	case STAT_TYPE_FLOAT:
		return sizeof(float);
	case STAT_TYPE_INT:
		return sizeof(int);
	case STAT_TYPE_SHORT:
		return sizeof(short);
	default:
		xprintf("SWAPI: Error invalid stat type %d\n", type);
		return 0;
	}
}

/** stats_stinit
 *
 * Issues a command to STATSMAN announcing that a new variable
 * has been created.
 * In order the higher layers (CMD_MAN, STATSMAN) to be able to change
 * this value, SERVICES daemon should wait the confiramtion of this 
 * initialization packet by checking the value of the options field
 * of the variable. In any case, object can still making use of this 
 * variable although anyone will be able to read or write it.
 * 
 */
int stats_stinit(daemon_o daemon, struct hwapi_var_i *var)
{
	char *objname, *appname, *statname;
	pkt_o pkt;
	var_o ovar;
	char type;

	assert(daemon && var);

	pkt = daemon_pkt(daemon);

	pkt_clear(pkt);

	/* variable name contains the application, object & stat name 
	 sepparated by ':' e.g.: appname:objname:statname */
	appname = var->name;
	objname = strstr(appname, ":");
	*objname = '\0';
	objname++;
	statname = strstr(objname, ":");
	*statname = '\0';
	statname++;

	hwapi_var_read(var->stat_id, (void *) &type, 1);

	/* create temporal variable using hwapi structure information */
	/**@todo watch here, size/type must be multiple, actually they are, because
	 * services created them. 
	 */
	ovar = var_new(var->size/sizeoftype(type), type);
	if (!ovar) {
		xprintf("STATS: Error creating stat %d.\n", var->stat_id);
		hwapi_var_setopt(var->stat_id, VAR_IS_STERR);
		return 0;
	}

	ovar->id = var->stat_id;
	str_set(ovar->name, statname);

	/** @todo change strcpy
	 */

	strcpy(pkt_putval(pkt, FIELD_APPNAME, strlen(appname) + 1), appname);
	strcpy(pkt_putval(pkt, FIELD_OBJNAME, strlen(objname) + 1), objname);

	pkt_putvalue(pkt, FIELD_OBJID, (uint) var->obj_id);
	pkt_put(pkt, FIELD_STAT, ovar, var_topkt);

	var_delete(&ovar);

	daemon_sendto(daemon, STATSMAN_STINIT, 0, DAEMON_STATSMAN);

	hwapi_var_setopt(var->stat_id, 0);

	return 1;
}

/** stats_stclose
 */
int stats_stclose(daemon_o daemon, struct hwapi_var_i *var)
{
	pkt_o pkt;

	assert(daemon);

	pkt = daemon_pkt(daemon);

	pkt_clear(pkt);

	pkt_putvalue(pkt, FIELD_OBJID, (uint) var->obj_id);
	pkt_putvalue(pkt, FIELD_STID, (uint) var->stat_id);

	daemon_sendto(daemon, STATSMAN_STCLOSE, 0, DAEMON_STATSMAN);
	
	//xprintf ("Closing stat %d for object %d\n", var->stat_id, var->obj_id);

	hwapi_var_delete(var->stat_id);

	return 1;

}

/** stats_pminit
 *
 * Issues a command to STATSMANAGER requesting to initialize
 * or read the parameter configuration file
 */
int stats_pminit(daemon_o daemon, struct hwapi_var_i *var)
{
	char *objname, *appname;
	pkt_o pkt;

	assert(daemon && var);

	pkt = daemon_pkt(daemon);

	pkt_clear(pkt);

	/* variable name contains the application & object
	 sepparated by : e.g.: appname:objname:statname */
	appname = var->name;
	objname = strstr(appname, ":");
	*objname = '\0';
	objname++;

	strcpy(pkt_putval(pkt, FIELD_APPNAME, strlen(appname) + 1), appname);
	strcpy(pkt_putval(pkt, FIELD_OBJNAME, strlen(objname) + 1), objname);

	pkt_putvalue(pkt, FIELD_OBJID, (uint) var->obj_id);
	pkt_putvalue(pkt, FIELD_PMID, (uint) var->stat_id);

	daemon_sendto(daemon, STATSMAN_PMINIT, 0, DAEMON_STATSMAN);

	hwapi_var_setopt(var->stat_id, 0);

	return 1;
}


/** stats_pmclose
 */
int stats_pmclose(daemon_o daemon, struct hwapi_var_i *var)
{
	pkt_o pkt;

	assert(daemon);

	pkt = daemon_pkt(daemon);

	pkt_clear(pkt);

	pkt_putvalue(pkt, FIELD_OBJID, (uint) var->obj_id);

	daemon_sendto(daemon, STATSMAN_PMCLOSE, 0, DAEMON_STATSMAN);

	hwapi_var_delete(var->stat_id);

	return 1;

}

/** stats_pmget
 *
 * Issues a command to STATSMAN requesting to obtain a parameter
 * value 
 */
int stats_pmget(daemon_o daemon, struct hwapi_var_i *var)
{
	pkt_o pkt;
	var_o ovar;
	char type;

	assert(daemon && var);

	pkt = daemon_pkt(daemon);

	pkt_clear(pkt);

	hwapi_var_read(var->stat_id, (void *) &type, 1);

	/** @todo check strcpy!! */

	strcpy(pkt_putval(pkt, FIELD_PMNAME, strlen(var->name) + 1), var->name);

	pkt_putvalue(pkt, FIELD_OBJID, (uint) var->obj_id);
	pkt_putvalue(pkt, FIELD_PMID, (uint) var->stat_id);
	pkt_putvalue(pkt, FIELD_PMTYPE, (uint) type);

	daemon_sendto(daemon, STATSMAN_PMGET, 0, DAEMON_STATSMAN);

	hwapi_var_setopt(var->stat_id, 0);

	return 1;

}


/** stats_loginit
 *
 * Issues a command to STATSMANAGER requesting to initialize
 * a log file
 */
int stats_loginit(daemon_o daemon, struct hwapi_var_i *var)
{
	char *objname, *appname;
	pkt_o pkt;

	assert(daemon && var);

	pkt = daemon_pkt(daemon);

	pkt_clear(pkt);

	/* variable name contains the application, object & log name
	 sepparated by : e.g.: appname:objname:logname */
	appname = var->name;
	objname = strstr(appname, ":");
	*objname = '\0';
	objname++;

	strcpy(pkt_putval(pkt, FIELD_APPNAME, strlen(appname) + 1), appname);
	strcpy(pkt_putval(pkt, FIELD_OBJNAME, strlen(objname) + 1), objname);

	pkt_putvalue(pkt, FIELD_OBJID, (uint) var->obj_id);
	pkt_putvalue(pkt, FIELD_LOGID, (uint) var->stat_id);

	daemon_sendto(daemon, STATSMAN_LOGINIT, 0, DAEMON_STATSMAN);

	hwapi_var_setopt(var->stat_id, 0);
	
//	xprintf ("STATS: Requesting creating of log %d for object %s\n", var->stat_id, objname);

	return 1;
}


/** stats_logclose
 */
int stats_logclose(daemon_o daemon, struct hwapi_var_i *var)
{
	pkt_o pkt;

	assert(daemon);

	pkt = daemon_pkt(daemon);

	pkt_clear(pkt);
	pkt_putvalue(pkt, FIELD_OBJID, (uint) var->obj_id);
	pkt_putvalue(pkt, FIELD_LOGID, (uint) var->stat_id);
	daemon_sendto(daemon, STATSMAN_LOGCLOSE, 0, DAEMON_STATSMAN);

	hwapi_var_delete(var->stat_id);

//	xprintf ("STATS: Requesting close of log %d\n", var->stat_id);

	return 1;
}

char logbuffer[LOG_LINE_SZ];

/** stats_logwrite
 */
int stats_logwrite(daemon_o daemon, struct hwapi_var_i *var)
{
	pkt_o pkt;

	assert(daemon);

	pkt = daemon_pkt(daemon);

	pkt_clear(pkt);
	pkt_putvalue(pkt, FIELD_OBJID, (uint) var->obj_id);
	pkt_putvalue(pkt, FIELD_LOGID, (uint) var->stat_id);
	
	/* contents are read from stat */
	hwapi_var_read(var->stat_id, logbuffer, LOG_LINE_SZ);
	if (strlen(logbuffer) > 0) {
		memcpy (pkt_putval (pkt, FIELD_LOGTXT, strlen(logbuffer)+1), logbuffer, strlen(logbuffer)+1);
		daemon_sendto(daemon, STATSMAN_LOGWRITE, 0, DAEMON_STATSMAN);		
		//xprintf ("STATS: Logging to %d for object %d len %d\n", var->stat_id, var->obj_id, strlen(logbuffer)+1);
	}
	hwapi_var_setopt(var->stat_id, 0);
	
	return 1;
}


/** @} */
