/** 
 * PHAL Stats Manager SW Daemon Implementation
 * 
 * FILE: statsman.c
 * 
 * Daemon type: Manager
 * 
 * Communicates with: STATS sensor, CMD MANAGER
 * 
 * Inputs: main input interface from front-end
 * 
 * Outputs: main output interface to front-end
 * 
 * Functions: This daemon has two main functionalities:
 * 		* provide initialization parameters to objects
 * 		  when they are requested.
 * 		* access or modify stats variables values when 
 * 		  it is requested by CMD MANAGER.
 * 	
 *  	The initialization values are different depending on the 
 *   	application which the object has been instantiated. Thus,
 * 	the user must define them for every application.
 * 	
 * Config Files:
 * 	* appname/objname.stats: It has the initialization values of object objname
 * 		in application "appname"   
 * 
 * 
 */

/* standard libraries */
#include <stdlib.h>
#include <string.h>

#include "phal_hw_api.h"

#include "phid.h"
#include "phal_daemons.h"
#include "pkt.h"
#include "daemon.h"

/* include command declarations */
#include "statsman_cmds.h"

/* main function declaration */
void statsman_run();

#ifdef RUN_AS_PROC
int main()
{
	statsman_run();
}
#endif

/** STATS MANAGER DAEMON MAIN FUNCTION 
 *
 */
void statsman_run()
{
	daemon_o daemon;
	int n;

	daemon = daemon_new(DAEMON_STATSMAN);
	if (!daemon) {
		xprintf ("STATSMAN: Error creating daemon\n");
		exit(-1);
	}

	/** Configure input commands
	 */

	if (daemon_addcmd(daemon, STATSMAN_STINIT, statsman_incmd_init) < 0) {
		exit(-1);
	}
	if (daemon_addcmd(daemon, STATSMAN_PMINIT, statsman_incmd_init) < 0) {
		exit(-1);
	}
	if (daemon_addcmd(daemon, STATSMAN_LOGINIT, statsman_incmd_init) < 0) {
		exit(-1);
	}

	if (daemon_addcmd(daemon, STATSMAN_STCLOSE, statsman_incmd_close) < 0) {
		exit(-1);
	}
	if (daemon_addcmd(daemon, STATSMAN_PMCLOSE, statsman_incmd_close) < 0) {
		exit(-1);
	}
	if (daemon_addcmd(daemon, STATSMAN_LOGCLOSE, statsman_incmd_close) < 0) {
		exit(-1);
	}

	if (daemon_addcmd(daemon, STATSMAN_STGET, statsman_incmd_stact) < 0) {
		exit(-1);
	}
	if (daemon_addcmd(daemon, STATSMAN_STSET, statsman_incmd_stact) < 0) {
		exit(-1);
	}

	if (daemon_addcmd(daemon, STATSMAN_STVALOK, statsman_incmd_stactack)
			< 0) {
		exit(-1);
	}
	if (daemon_addcmd(daemon, STATSMAN_STSETOK, statsman_incmd_stactack)
			< 0) {
		exit(-1);
	}
	if (daemon_addcmd(daemon, STATSMAN_STVALERR, statsman_incmd_stactack)
			< 0) {
		exit(-1);
	}
	if (daemon_addcmd(daemon, STATSMAN_STSETERR, statsman_incmd_stactack)
			< 0) {
		exit(-1);
	}

	if (daemon_addcmd(daemon, STATSMAN_PMGET, statsman_incmd_pmget) < 0) {
		exit(-1);
	}

	if (daemon_addcmd(daemon, STATSMAN_LOGWRITE, statsman_incmd_logwrite) < 0) {
		exit(-1);
	}

	if (statsman_init() < 0) {
		exit(-1);
	}

	xprintf ("STATSMAN: Init with PID=%d PPID=%d\n", getpid(), getppid());

	/** main daemon loop 
	 * while packets in queue, process them
	 */
	while (1) {
		do {
			n = daemon_rcvpkt(daemon);
			if (n > 0) {
				daemon_processcmd(daemon);
			}
		} while (n);
		if (n < 0) {
			break;
		}
		hwapi_relinquish_daemon();
	}

	daemon_delete(&daemon);

	exit(-1);
}
