/** @file swload_cmds.c
 * @brief PHAL Software Load SW Daemon Implementation
 *
 * This file contains the implementation of the command functions
 *
 * @todo tema errors al carregar executables o aplicacions
 *
 */

/* standard libraries */
#include <stdlib.h>
#include <string.h>

/* phal hwlibrary constants */
#include "phal_hw_api.h"

#include "phid.h"
#include "phal_daemons.h"

#include "swman.h"

/* common objects */
#include "set.h"
#include "str.h"
#include "rtcobj.h"
#include "phitf.h"
#include "phobj.h"
#include "app.h"
#include "pkt.h"
#include "downexe.h"
#include "daemon.h"

/* commands */
#include "swload_cmds.h"
#include "swman_cmds.h"

#include "bridge.h"
#include "bridge_cmds.h"

/* downloading queue */
Set_o down_q= NULL;

#define DEFAULT_ITF_SZ	0x1000

void load_error (daemon_o daemon, downexe_o exe);

int swload_init(void)
{
	down_q = Set_new(0, NULL);
	if (!down_q) {
		xprintf("SWLOAD: Fatal error while initiating\n");
		return -1;
	}
	return 1;
}

void request_exec(daemon_o daemon, int plt_family, str_o exe_name) {
	pkt_o pkt;
	
	assert (daemon && exe_name);
	
	pkt = daemon_pkt(daemon);
	xprintf("SWLOAD: Requesting executable %s for platform %d\n",
			str_str(exe_name), plt_family);

	pkt_clear(pkt);
	pkt_put(pkt, FIELD_EXENAME, exe_name, str_topkt);
	pkt_putvalue(pkt, FIELD_PLATID, (uint) plt_family);

	daemon_sendto(daemon, SWMAN_EXECREQ, 0, DAEMON_SWMAN);

}

/** LOAD OBJECT COMMAND
 *
 * This function is called when a command to load a list of objects
 * with a list of interfaces connecting them is received. The procedure
 * followed to provide such connectivity is the following:
 *
 * - Internal interfaces (the remote object is in the same processor): 
 *		A name like "w_obj_name:w_itf_name" and "r_obj_name:r_itf_name"
 * 		is associated using the HWAPI Interface Names Service with the 
 *		write and read side of the interface, created by this function.
 * - External Interfaces (the local interface is associated to a physical 
 *	interface id and a remote interface id decided by HWMAP.
 *		This function will create the object side of the interface and assign
 *		the same name for both sides. The bridge, will attach
 *		its interface when receives the order from HWMAP. 
 * 
 * @todo Configure Bridge Interfaces
 * @todo Setup interface options (delay)
 * @todo Executable Request error management
 * @todo Absolute Addressing Computers (Relocation)
 */
int swload_incmd_loadobj(daemon_o daemon)
{

	pkt_o pkt;
	phobj_o obj, r_obj;
	downexe_o exe;
	Set_o objset;
	app_o app;
	int nof_obj;
	int i, n;
	int plt_family;
	struct hwapi_cpu_i cpu;

	assert(daemon);

	pkt = daemon_pkt(daemon);

	hwapi_hwinfo_cpu(&cpu);
	plt_family = cpu.plat_family;

	app = (app_o) pkt_get(pkt, FIELD_APP, app_newfrompkt);
	if (!app) {
		xprintf("SWLOAD: Received an invalid object list. Ignoring command\n");
		return 0;
	}
	objset = app->objects;
	if (!objset || !Set_length(objset)) {
		xprintf("SWLOAD: Received a void object list. Ignoring command.\n");
		if (objset) {
			Set_delete(&objset);
		}
		/** @todo send error if received invalid list
		 */
		return 0;
	}

	for (i = 0; i < Set_length(objset); i++) {
		obj = Set_get(objset, i);
		if (!obj) {
			return 0;
		}

		/** Now check if we need to request an executable 
		 */

		exe = Set_find(down_q, str_str(obj->exename), downexe_findname);
		if (!exe) {
			exe = downexe_new();
			if (!exe) {
				return 0;
			}

			str_cpy(exe->name, obj->exename);
			Set_put(down_q, exe);
			if (!i) {
				request_exec(daemon, plt_family, obj->exename);
			}
		}
		Set_put(exe->objects, phobj_dup(obj));
	}

	app_delete(&app);
	return 1;
}

/** SWLOAD: Executable request error 
 * @todo what to do with errors
 */
int swload_incmd_execerr(daemon_o daemon)
{
	pkt_o pkt;
	phobj_o obj;
	downexe_o exe;

	assert(daemon);

	pkt = daemon_pkt(daemon);

	exe = Set_find(down_q, pkt_getval(pkt, FIELD_EXENAME),
		downexe_findname);
	if (!exe) {
		xprintf("SWLOAD: Can't find executable at downloading queue\n\tCaution! Some executables may be still in queue.\n");
		return 0;
	}

	xprintf ("SWLOAD: Error when requesting executable %s\n", str_str (exe->name));
	
	load_error (daemon, exe);

}

/** SWLOAD: Load start command
 *
 * Now the final executable is generated and ready to be downloaded. This command
 * indicates the length (byte and word aligned) and the size of the packet that will
 * be used to slice the executable when downloading.
 * 
 * Here, as a lot of memory needs to be allocated when receiveing files, dynamic
 * memory is used.
 *
 * @todo Variable packet size
 *
 * @todo ACK del LOADSTART (per si no pot allocar)
 */
int swload_incmd_loadstart(daemon_o daemon)
{
	pkt_o pkt;
	phobj_o obj;
	downexe_o exe;
	int error = 0;

	assert(daemon);

	pkt = daemon_pkt(daemon);

	exe = Set_find(down_q, pkt_getval(pkt, FIELD_EXENAME),
					downexe_findname);
	if (!exe) {
		xprintf("SWLOAD: Can't find executable %s at downloading queue\n", pkt_getval (pkt, FIELD_EXENAME));
		load_error (daemon, NULL);
	}

	memcpy(&exe->pinfo, pkt_getval(pkt, FIELD_PINFO), sizeof(struct pinfo));
	if (!exe->data && exe->pinfo.blen) {
		exe->data = calloc(1, exe->pinfo.blen);
	}
	if (!exe->data) {
		xprintf("SWLOAD: Error allocating memory\n");
		obj = Set_get(exe->objects, 0);
		pkt_put(pkt, FIELD_APPNAME, obj->appname, str_topkt);
	}

	xprintf("SWLOAD: File Allocated, ready to download.\n");

	return 1;
}

int save_packet(void *x, void *data, int size)
{
	downexe_o downexe = (downexe_o) x;
	int i;
	unsigned int *p;

	assert(x && data && size);

	if (downexe->rpm + size > downexe->pinfo.blen) {
		xprintf("error with size\n");
		return -1;
	}

	memcpy(&downexe->data[downexe->rpm], data, size);

	p = (unsigned int *) data;

	/* refresh checksum */
	for (i = 0; i < ((size-1)/4) + 1; i++) {
		downexe->checksum += p[i];
	}

	downexe->rpm += ((size-1)/4) + 1;

	return 1;
}

void load_error (daemon_o daemon, downexe_o exe) 
{
	pkt_o pkt;
	phobj_o obj;
	assert(daemon);
	
	pkt = daemon_pkt (daemon);
	pkt_clear(pkt);
	if (exe) {
		obj = Set_get(exe->objects, 0);
		pkt_put(pkt, FIELD_APPNAME, obj->appname, str_topkt);
		Set_remove(down_q, exe);
		downexe_delete(&exe);
	}
	daemon_sendto(daemon, SWMAN_EXECERR, 0, DAEMON_SWMAN);
}

/** SWLOAD: Loading progress command
 *
 * With this function the actual body of the executable is 
 * saved into the buffer allocated during the loadstart function.
 *
 * @todo Errors when downloading
 */
int swload_incmd_loading(daemon_o daemon)
{
	pkt_o pkt;
	downexe_o exe;
	phobj_o obj;
	
	assert(daemon);

	pkt = daemon_pkt(daemon);

	exe = Set_find(down_q, pkt_getval(pkt, FIELD_EXENAME),
			downexe_findname);
	if (!exe) {
		xprintf("SWLOAD: Error downloading. Unknown executable name %s.\n",
				pkt_getval(pkt, FIELD_EXENAME));
		load_error(daemon, NULL);
		return 0;
	}
	if (!exe->data) {
		xprintf("SWLOAD: Error not received LOAD_START packet\n");
		load_error(daemon, exe);
		return 0;
	}
	if (!save_packet(exe, pkt_getval(pkt, FIELD_BINDATA),
			pkt_getvalsz(pkt, FIELD_BINDATA))) {
		xprintf("SWLOAD: Error saving binary data\n");
		load_error(daemon, exe);
		return 0;
	}		
	return 1;
}

/** Launch an instance (object) for every process 
 * @todo APPID???
 */
int swload_launchall(downexe_o downexe)
{
	Set_o objset;
	phobj_o obj;

	int i;
	struct hwapi_proc_launch plaunch;

	assert(downexe);

	objset = downexe->objects;

	for (i = 0; i < Set_length(objset); i++) {
		obj = Set_get(objset, i);

		plaunch.app_id = obj->app_id;
		plaunch.obj_id = obj->obj_id;
		plaunch.obj_name = str_str(obj->objname);
		plaunch.app_name = str_str(obj->appname);
		plaunch.exe_name = str_str(downexe->name);
		plaunch.blen = downexe->pinfo.blen;
		plaunch.wlen = downexe->pinfo.wlen;
		if (hwapi_proc_create((char *) downexe->data, &plaunch) < 0) {
			xprintf("Error Launching executable.\n");
			return 0;
		}
	}
	return 1;
}

/** SWLOAD: Loading finish command
 *
 * When the transmission ends, this packet is received with the 
 * checksum to check errors. 
 *
 * If all is ok, we continue with launching the process.
 */
int swload_incmd_loadend(daemon_o daemon)
{

	pkt_o pkt;
	phobj_o obj;
	downexe_o exe;
	int i;
	int plt_family;
	struct hwapi_cpu_i cpu;

	assert(daemon);

	pkt = daemon_pkt(daemon);

	hwapi_hwinfo_cpu(&cpu);
	plt_family = cpu.plat_family;
	
	exe = Set_find(down_q, pkt_getval(pkt, FIELD_EXENAME),
			downexe_findname);
	if (!exe) {
		xprintf("SWLOAD: Can't find executable at downloading queue\n");
		load_error(daemon, NULL);
		return 0;
	}

	if (exe->checksum ^ ((uint) pkt_getvalue(pkt, FIELD_CHKSUM))) {
		xprintf("SWLOAD: Error with checksum (0x%x)\n",
				pkt_getvalue(pkt, FIELD_CHKSUM));
		//load_error(daemon, exe);
		//return 0;
	}
	if (!swload_launchall(exe)) {
		xprintf("SWMAN: Error launching processes\n");
		load_error(daemon, exe);
		return 0;
	}
	for (i = 0; i < Set_length(exe->objects); i++) {
		if (!configure_itfs(daemon, Set_get(exe->objects, i))) {
			xprintf("SWLOAD: Error configuring itfs.\n");
			load_error(daemon, exe);
			return 0;
		}
	}
	
	Set_remove(down_q, exe);
	downexe_delete(&exe);
	
	/* request next one */
	exe = Set_get(down_q, 0);
	if (exe) {
		request_exec (daemon, plt_family, exe->name);
	}

	return 1;
}

int add_bridge_itf (daemon_o daemon, int itf_id, objid_t obj_id, phitf_o itf)
{
	pkt_o pkt;
	data_header_t header;
	struct hwapi_cpu_i cpu;
	
	pkt = daemon_pkt(daemon);
	
	pkt_clear(pkt);

	if (phitf_iswriteonly(itf)) {
		header.ItfId = itf->remote_id;
		header.ObjId = itf->remote_obj_id;		
	} else {
		header.ItfId = itf->id;
		header.ObjId = obj_id;
	}
	
	memcpy (pkt_putval (pkt, FIELD_DATAHEADER, sizeof (data_header_t)), &header, sizeof (data_header_t));
	pkt_putvalue (pkt, FIELD_OBJITFID, itf_id);
	pkt_putvalue (pkt, FIELD_OBJITFMODE, itf->mode);
	pkt_putvalue (pkt, FIELD_OBJXITFID, itf->xitf_id);
	
	hwapi_hwinfo_cpu (&cpu);
	
	return daemon_sendto(daemon, BRIDGE_ADDITF, cpu.pe_id, DAEMON_BRIDGE);
}

int configure_itfs(daemon_o daemon, phobj_o obj)
{
	int j;
	phitf_o itf;
	int n;
	char *w_obj, *r_obj, *w_itf, *r_itf;

	/* for every interface in the object */
	for (j = 0; j < Set_length(obj->itfs); j++) {
		itf = Set_get(obj->itfs, j);

		/* create the interface just for one side */
		if (phitf_iswriteonly(itf) || !phitf_isinternal(itf)) {
			n = hwapi_itf_create(0, DEFAULT_ITF_SZ);
			if (!n) {
				xprintf("SWLOAD: Error creating interface\n");
				return 0;
			}
			
	
			if (phitf_iswriteonly(itf)) {
				w_obj = str_str(obj->objname);
				w_itf = str_str(itf->name);
				r_obj = str_str(itf->remoteobjname);
				r_itf = str_str(itf->remotename); 
			} else {
				r_obj = str_str(obj->objname);
				r_itf = str_str(itf->name);
				w_obj = str_str(itf->remoteobjname);
				w_itf = str_str(itf->remotename);
			}
		 
			/* assign same interface id for read/write side */
			if (hwapi_itf_match(n, w_obj, w_itf, /* write itf name */
						r_obj, r_itf) /* reader itf name */
				< 0) {
				xprintf("SWLOAD: Error assigning interface\n");
				return 0;
			}
			
			if (!phitf_isinternal(itf)) {
				if (!add_bridge_itf(daemon, n, obj->obj_id, itf)) {
					xprintf ("SWLOAD: Error sending itf to bridge\n");
					return 0;
				}
			}
		}
		/** @todo Setup interface options (delay, etc...)
		 */
		/*set_itf_opts(phitf[itf_idx+j].delay); */
	}

	return 1;
}

/** SWLOAD: Executable Information Answer
 *
 * This command is received after a request for memory information, 
 * here we need to obtain processor memory resources and request
 * the executable addressed to such memory.
 *
 * As memory has to be allocated, it may happen that not enough is 
 * available. In such case, an error should be reported indicating 
 * the impossibiliness of loading the executable.
 */

