/** @file swman.c
 * @brief PHAL Software Manager SW Daemon Implementation
 * 
 * Functions: This manager has two main tasks, each of them
 * 	is achieved using a different sensor:
 * 	- Management of software applications --> SWLOAD.
 * 		This includes reading the application configuration and 
 * 		downloading the correct executable depending on platform
 * 		characteristics (obtained by SWLOAD).
 * 	- Management of application execution --> EXEC.
 * 		This includes modifying its execution state (init/run/pause/stop)
 * 		and ensuring all objects are in correct state under real-time
 * 		considerations (monitored and reported by EXEC).
 * 	
 * Config Files:
 * 	* appname.swman: Contains the list of objects, their interfaces
 * 	  	and how are they connected.
 * 	* executables.swman: Contains the list of available executables, 
 * 		their path, and in which form are they available:
 * 			- source code
 * 			- compiled (not linked with p-hal libs)
 * 			- linked relocatable 
 * 			- linked non-relocatable
 */

/* standard libraries */
#include <stdlib.h>
#include <string.h>

#include "phal_hw_api.h"

#include "phid.h"
#include "phal_daemons.h"

/* common objects */
#include "pkt.h"
#include "daemon.h"

/* include command declarations */
#include "swman_cmds.h"

/* main function declaration */
void swman_run();

#ifdef RUN_AS_PROC
int main()
{
	swman_run();
}
#endif

/** SW MANAGER DAEMON MAIN FUNCTION 
 *
 * Configures man_daemon_cfg structure and launches UTILS main function
 *
 * The following commands are accepted (constants are defined at swman_cmds.h) 
 * and produce the following results:
 *								=== command produced ==
 *   Commands from CMD_MAN:					
 * 	- SWMAN_LOADAPP: Command to load an applicatoin		--> SWLOAD_LOADOBJ/CMDMAN_SWLOADERR
 * 	- SWMAN_APPSTATUS: Modify the status of an app		--> EXEC_STATUS/CMDMAN_SWSTATUSERR
 *
 *   Commands from SENSOR(S):
 *	(from SWLOAD)
 *	- SWMAN_EXECREQ: Processor requests an executable	--> SWLOAD_EXECLOAD/SWLOAD_EXECERR/CMDMAN_SWLOADOK/ERR
 *	- SWMAN_EXECINFO: Processor requests exec. info		--> SWLOAD_EXECINFO/SWLOAD_EXECERR
 *	- SWMAN_EXECERR: Error while loading executable		--> CMDMAN_SWLOADERR
 *
 *	(from EXEC)
 *	- SWMAN_STATUSOK: Object notifies status change		--> CMDMAN_SWSTATUSOK
 *	- SWMAN_STATUSERR: Object not changed at time		--> CMDMAN_SWSTATUSERR
 *
 *   todo commands:
 *	- SWMAN_RTCFAULT: Exec reports failure in real-time	--> CMDMAN_RTCFAULT
 */
int swman_init(void);

void swman_run()
{
	daemon_o daemon;
	int n;

	daemon = daemon_new(DAEMON_SWMAN);
	if (!daemon) {
		xprintf ("swman: Error creating daemon\n");
		exit(-1);
	}

	/** Configure input commands
	 */

	if (daemon_addcmd(daemon, SWMAN_LOADAPP, swman_incmd_loadapp) < 0) {
		exit(-1);
	}
	if (daemon_addcmd(daemon, SWMAN_APPSTATUS, swman_incmd_appstatus) < 0) {
		exit(-1);
	}
	if (daemon_addcmd(daemon, SWMAN_EXECREQ, swman_incmd_execreq) < 0) {
		exit(-1);
	}
	if (daemon_addcmd(daemon, SWMAN_EXECERR, swman_incmd_execerr) < 0) {
		exit(-1);
	}
	if (daemon_addcmd(daemon, SWMAN_STATUSOK, swman_incmd_statusrep) < 0) {
		exit(-1);
	}
	if (daemon_addcmd(daemon, SWMAN_STATUSERR, swman_incmd_statusrep) < 0) {
		exit(-1);
	}

	if (swman_init() < 0) {
		exit(-1);
	}

	xprintf ("SWMAN: Init with PID=%d PPID=%d\n", getpid(), getppid());

	/** main daemon loop 
	 * while packets in queue, process them
	 */
	while (1) {

		do {
			n = daemon_rcvpkt(daemon);
			if (n>0) {
				daemon_processcmd(daemon);
			}
		} while (n>0);
		if (n < 0) {
			break;
		}
		hwapi_relinquish_daemon();
	}

	daemon_delete(&daemon);
	
	exit(-1);
}
