/**
 *
 *
 *	PHAL SW Manager SW Daemon Implementation
 *
 * FILE: swman_cmds.c
 *
 * Description: Here is where functions answering COMMANDS are 
 *	implemented
 *
 *
 * @TODO tema errors
 *
 */

/* standard libraries */
#include <stdlib.h>
#include <string.h>

#include "phal_hw_api.h"

#include "phid.h"
#include "phal_daemons.h"

/* common objects */
#include "set.h"
#include "str.h"
#include "var.h"
#include "app.h"
#include "rtcobj.h"
#include "phobj.h"
#include "pkt.h"
#include "daemon.h"

#include "swman.h"

//#define DEB


/* commands */
#include "swload_cmds.h"
#include "exec_cmds.h"
#include "cmdman_cmds.h"
#include "swman_cmds.h"
#include "hwman_cmds.h"

#include "swman_parser.h"

#define UPLOAD_PACKETS 1000

#define DEFAULT_STATUS_INC_TS 5

/* This is the main database for objects */
Set_o apps_db= NULL;

peid_t local_id = 0;

/* function to initialize database */
int swman_init(void)
{
	apps_db = Set_new(0, NULL);
	if (!apps_db) {
		xprintf("SWMAN: Error creating main database\n");
		return 0;
	}
	return 1;
}

/** Set app status
 *
 * Tries to set a new application status. It only can be changed
 * when all objects did change their status. 
 *
 * @return 1 if status did changed, 0 if not
 */
int set_app_status(app_o app, status_t new_status)
{
	if (status_ok(app) || new_status == PHAL_STATUS_STOP
			|| new_status == PHAL_STATUS_ABNSTOP) {
		app->cur_status = new_status;
		return 1;
	} else {
		return 0;
	}
}

/** Check if app changed ok
 *
 * Checks if all the application objects did change their status. 
 * That means that any object is in an status different of 
 * current application status.
 */
int status_ok(app_o app)
{
	int ok = 0;
	
	if (Set_find(app->objects, &(app)->cur_status, phobj_finddiffstatus)) {
		ok = 0;
	} else {
		ok = 1;
	}
	return ok;
}

void loadapp_error(daemon_o daemon, app_o app)
{
	pkt_o pkt;
	
	assert (daemon);
	
	pkt = daemon_pkt(daemon);
	daemon_sendto(daemon, CMDMAN_SWLOADAPPERR, 0, DAEMON_CMDMAN);
	if (app) {
		Set_remove(apps_db, app);
		app_delete(&app);		
	}
}

/** SWMAN: Load an application function
 *							
 * This function reads the application config file, parses it
 *	and sends it to SWLOAD the plain list of objects and their
 *	interconnection so the loading and linking is performed.
 *
 *	This packet may be captured by any HWMAP daemon located in the 
 *	same or lower processor and perform a re-assignation of sw into
 *	processing resources (see doc)							
 *	
 *	The configuration file of the application must have the following name:
 *		- sw_daemons/sw_cfg/appname.swman
 *	(see constants in the beggining of this file).
 *
 * 	In these files, objects, aplications and interfaces are descripted
 *	using names. After reading the configuration file, an id is generated 
 * 	for every object in the application. This id must be unique in all P-HAL 
 *	environment.
 *
 *	This id will be internally associated with the object and, after being
 *	registered by SWLOAD when process is launched, it will be available to the 
 *	whole processor, enabling the EXEC sensor to monitor its execution status.
 *	
 *	So, EXEC reports will be associated to Ids instead of names, also enabling
 *	low overhead in such reports as everything in P-HAL that must be used during
 *	run-time must be as lightweight as possible, this explains why strings are not
 *	used.
 *	@todo comment this
 *	@todo errors in downloading
 */
int swman_incmd_loadapp(daemon_o daemon)
{
	pkt_o pkt;
	app_o app= NULL;
	phobj_o obj;
	str_o app_name;
	int i;
	int n;
	struct hwapi_cpu_i cpu_i;

	assert(daemon);
	
	if (!local_id) {
		hwapi_hwinfo_cpu(&cpu_i);
		if (cpu_i.pe_id) {
			local_id = cpu_i.pe_id;
		} else {
			xprintf ("SWMAN: Can't load application, pe not registered.\n");
			loadapp_error(daemon, NULL);
			return 0;
		}
	}

	pkt = daemon_pkt(daemon);

	app_name = (str_o) pkt_get(pkt, FIELD_APPNAME, str_newfrompkt);
	app = Set_find(apps_db, str_str(app_name), app_findname);
	if (app) {
		xprintf("SWMAN: Error application already loaded\n");
		loadapp_error(daemon, NULL);
		return 0;
	} else {
		app = app_new();
		if (!app) {
			xprintf("SWMAN: Error creating application\n");
			loadapp_error(daemon, app);
			return 0;
		} else {
			str_cpy(app->name, app_name);
			Set_put(apps_db, app);
		}
	}
	
	if (!swman_parser_app(str_str(app_name), app->objects)) {
		xprintf("SWMAN: Error parsing application configuration file.\n");
		loadapp_error (daemon, app);
		return 0;
	}
	if (!Set_length(app->objects)) {
		xprintf("SWMAN: Any object was parsed, can't load application.\n");
		loadapp_error(daemon, app);
		return 0;
	} 
	if (!set_app_status(app, PHAL_STATUS_STARTED)) {
		xprintf("SWMAN: Error setting application initial status.\n");
		loadapp_error(daemon, app);
		return 0;
	}

	pkt_clear(pkt);

	for (i = 0; i < Set_length(app->objects); i++) {
		obj = Set_get(app->objects, i);
		str_cpy(obj->appname, app->name);
		obj->app_id = app->app_id;
	}

	if (pkt_put(pkt, FIELD_APP, app, app_topkt)) {
		daemon_sendto(daemon, HWMAN_MAPOBJ, local_id, DAEMON_HWMAN);
	} else {
		xprintf("SWMAN: Error building output packet\n");
	}
	
	return 1;
}

int swman_sendobjstatus(daemon_o daemon, app_o app, status_t status, int tstamp)
{
	int i;
	Set_o objset;
	phobj_o obj;
	pkt_o pkt;
	int xtstamp;

	pkt = daemon_pkt(daemon);

	objset = app->objects;
	if (!objset) {
		return 0;
	}


	if (!tstamp) {
		xtstamp = get_tstamp() + DEFAULT_STATUS_INC_TS;
	} else {
		xtstamp = tstamp;
	}

	/* for every object */
	for (i = 0; i < Set_length(objset); i++) {
		obj = Set_get(objset, i);

		if (!obj->pe_id) {
			xprintf("SWMAN: Error some objects have not been registered yet.\n");
			return 0;
		}

		/* send only if object is not stopped */
		if (obj->status != PHAL_STATUS_STOP && obj->status
				!= PHAL_STATUS_ABNSTOP) {

			pkt_clear(pkt);
			pkt_putvalue(pkt, FIELD_STATUS, (uint) status);

			//obj->status = 0;
			pkt_putvalue(pkt, FIELD_OBJID,
					(uint) obj->obj_id);
			pkt_putvalue(pkt, FIELD_TSTAMP, (uint) tstamp);
			daemon_sendto(daemon, EXEC_SETSTATUS,
					obj->pe_id, DAEMON_EXEC);

		}
	}

	return 1;
}

/** SWMAN: Change application status command.
 *
 * This function is called when a command to change an application
 *	status is received from CMD_MAN. If the application has been
 *	loaded from this swman and the objects have been (all of them)
 * 	correctly loaded, the status change will be ordered to the 
 *	(under us) array of sensors (EXEC) controlling objects execution.
 */
int swman_incmd_appstatus(daemon_o daemon)
{
	pkt_o pkt;
	app_o app= NULL;
	Set_o objset;
	phobj_o obj;
	str_o app_name;
	int i, n;
	status_t status;
	int tstamp;

	assert(daemon);

	pkt = daemon_pkt(daemon);

	app_name = (str_o) pkt_getval(pkt, FIELD_APPNAME);
	status = (status_t) pkt_getvalue(pkt, FIELD_STATUS);
	tstamp = (int) pkt_getvalue(pkt, FIELD_TSTAMP);

	app = Set_find(apps_db, app_name, app_findname);
	if (!app) {
		xprintf("SWMAN: Error can't find application\n");
		loadapp_error(daemon, NULL);
		return 0;
	}

	if (status == PHAL_STATUS_STEP) {
		n = set_app_status(app, PHAL_STATUS_PAUSE);
	} else {
		n = set_app_status(app, status);
	}
	if (!n) {
		xprintf("SWMAN: Can't change application status\n");
		loadapp_error(daemon, NULL);
		return 0;
	}

	if (!swman_sendobjstatus(daemon, app, status, tstamp)) {
		xprintf("SWMAN: Error setting application status\n");
		loadapp_error(daemon, NULL);
		return 0;
	}

	return 1;
}

/** SWMAN: Object reports status change 
 *
 * Here EXEC will tell us a successfull change in its status had been done
 *	or if an error occurred.
 * It can also be received any time if the EXEC sensor detects that some
 *	object is not in correct state. In such case, it will be reported to
 * 	CMD_MAN daemon.
 * In the case that all the objects reach the correct status, it will also 
 *	be reported to CMD_MAN indicating a successful change of the application status
 *
 * @todo What to do when an application status change failed in some object
 * @todo RealTime Monitoring of execution time. What to do when it fails
 *
 */
int swman_incmd_statusrep(daemon_o daemon)
{
	pkt_o pkt;
	app_o app;
	status_t status;
	phobj_o obj;
	short out_type = 0;

	assert(daemon);

	pkt = daemon_pkt(daemon);
	
	app = Set_find(apps_db, pkt_getval(pkt, FIELD_APPID), app_findid);
	if (!app) {
		xprintf("SWMAN: Error can't find application with id %d (object 0x%x, status %d, ok %d)\n",
			pkt_getvalue(pkt, FIELD_APPID), pkt_getvalue(pkt, FIELD_OBJID),
			pkt_getvalue (pkt, FIELD_STATUS),
			pkt_getcmd(pkt) == SWMAN_STATUSOK);
		return 0;
	}
	obj = Set_find(app->objects, pkt_getval(pkt, FIELD_OBJID),
			phobj_findid);
	if (!obj) {
		xprintf("SWMAN: Error object 0x%x not found.\n", pkt_getvalue(
				pkt, FIELD_OBJID));
		return -0;
	}

	status = (status_t) pkt_getvalue(pkt, FIELD_STATUS);
	
	if (pkt_getcmd(pkt) == SWMAN_STATUSOK 
			&& ((status == app->cur_status || app->cur_status == PHAL_STATUS_ABNSTOP))) {

		obj->status = status;

		/* if this is the first packet, register pe where it is running */
		if (obj->status == PHAL_STATUS_STARTED) {
			obj->pe_id = pkt_getsrcpe(pkt);
		}

		
	} else {
		if (status != app->cur_status && status != PHAL_STATUS_ABNSTOP) {
			xprintf("SWMAN: Incorrect status for object 0x%x (obj %d, app %d)\n",
					obj->obj_id, pkt_getvalue(pkt, FIELD_STATUS),
					app->cur_status);			
		} else if (status == PHAL_STATUS_ABNSTOP) {
			xprintf ("SWMAN: Object 0x%x died unexpectectly.\n", obj->obj_id);
			obj->status = PHAL_STATUS_ABNSTOP;
		} else {
			xprintf ("SWMAN: Object 0x%x did not changed to status %d\n", obj->obj_id, app->cur_status);
		}
		
		/* stop all the application */
		if (app->cur_status != PHAL_STATUS_ABNSTOP) {
			if (!set_app_status(app, PHAL_STATUS_ABNSTOP)) {
				xprintf("SWMAN: Error setting stop status.\n");
			}
			
			swman_sendobjstatus(daemon, app, PHAL_STATUS_ABNSTOP, 0);			
		}
		obj->status = PHAL_STATUS_ABNSTOP;

	}
	
	/* if all objects changed their status we need to report it to CMDMAN */
	if (status_ok(app)) {
		if (obj->status == PHAL_STATUS_STARTED) {
			out_type = CMDMAN_SWLOADAPPOK;
		} else if (app->cur_status == PHAL_STATUS_ABNSTOP) {
			out_type = CMDMAN_SWSTATUSERR;
		} else {
			out_type = CMDMAN_SWSTATUSOK;
		}

		if (app->cur_status == PHAL_STATUS_STOP || app->cur_status == PHAL_STATUS_ABNSTOP) {
			Set_remove(apps_db, app);
			app_delete(&app);
			xprintf("SWMAN: Removed application from db\n");
		}
	}

	if (out_type) {
		pkt_clear(pkt);
		pkt_put(pkt, FIELD_APPNAME, app->name, str_topkt);
		daemon_sendto(daemon, out_type, 0, DAEMON_CMDMAN);
	}

	return 1;
}

/** SWLOAD reports executable error.
 * @todo what to do with errors
 */
int swman_incmd_execerr(daemon_o daemon)
{
	pkt_o pkt;
	app_o app;

	assert(daemon);

	pkt = daemon_pkt(daemon);

	xprintf("SWMAN: Error loading application %s\n", (char *) pkt_getval(
			pkt, FIELD_APPNAME));

	app = Set_find(apps_db, pkt_getval(pkt, FIELD_APPNAME), app_findname);
	if (!app) {
		xprintf("app not found\n");
		return 1;
	}

	Set_remove(apps_db, app);
	app_delete(&app);

	daemon_sendto(daemon, CMDMAN_SWLOADAPPERR, 0, DAEMON_CMDMAN);

	return 1;
}

struct exec
{
	str_o name;
	pltid_t platform;
	struct pinfo pinfo;
	unsigned int checksum;
	int fd;
};

int pinfotopkt(void *x, char **start, char *end)
{
	struct exec *exeobj = (struct exec*) x;

	assert(x && start && end);
	assert(end >= *start && *start);

	if ((end - *start) < sizeof(struct pinfo)) {
		return 0;
	}

	memcpy(*start, &exeobj->pinfo, sizeof(struct pinfo));
	*start += sizeof(struct pinfo);

	return 1;
}

int popbindata(void *x, char **start, char *end)
{
	int t, n, i;
	struct exec *exeobj = (struct exec*) x;
	unsigned int *p;

	assert(x && start && end);
	assert(end >= *start && *start);

	if ((end - *start) < DEFAULT_PKT_SZ) {
		return 0;
	}

	n =hwapi_res_read(exeobj->fd, *start, DEFAULT_PKT_SZ);
	if (n < 0) {
		xprintf("SWMAN:\t\t Error reading file\n");
		return -1;
	}

	t = ((n - 1) >> 2) + 1;
	p = (unsigned int*) *start;
	for (i = 0; i < t; i++) {
		exeobj->checksum += p[i];
	}

	*start += n;

	return 1;
}

char * get_path(struct exec *exec);

void loadexec_error (daemon_o daemon, struct exec *exec, peid_t src_pe)
{
	pkt_o pkt;
	assert(daemon);
	
	pkt = daemon_pkt(daemon);
	if (exec) {
		if (exec->name) {
			pkt_clear(pkt);
			pkt_put(pkt, FIELD_EXENAME, exec->name, str_topkt);
			str_delete(&exec->name);
		}		
	}
	daemon_sendto(daemon, SWLOAD_EXECERR, src_pe, DAEMON_SWLOAD);
}

/** SWMAN: Processor downloading function
 *
 * After the SWLOAD sensor obtains information about platform characteristics
 * and loading procedures, and in the case absolute addressing is required, memory
 * has been allocated, a command to download the executable for certain platform is
 * issued.
 * 
 * This function processes such command. The downloading method is very simple:
 *	1) Check in which form is presented the executable and do the appropiate, that means:
 *		a) source code: Compile for the platform.
 *		b) compiled: Link with P-HAL specific platform libraries.
 *		c) linked: Do nothing.
 *	2) If absolute addressing is required and relocator for the platform is available
 *		do the relocation procedure to the appropiate addresses.
 *	3) Send the executable file in packets of pre-defined size (TODO: platform
 *		can choose appropiate size, or negotiate it). This downloading is divided in 
 *		three stages:
 *			a) LOADSTART packet: Send a packet indicating the loading starting 
 *				and the total number of bytes that will be transmitted.
 *			b) LOADING packets: Send the executable itself, splitted in packets
 *			c) LOADEND packet: Indicate the end of the transmision and the checksum
 */
int swman_incmd_execreq(daemon_o daemon)
{
	pkt_o pkt;
	struct exec exec;
	peid_t src_pe;
	objid_t obj_id;
	int n;
	void *ptr;
	int c=0;
	int fd = 0;
	int proc_blen;
	char *exe_name, *path;
	pltid_t platform;

	assert(daemon);

	pkt = daemon_pkt(daemon);

	src_pe = pkt_getsrcpe(pkt);
	exe_name = pkt_getval(pkt, FIELD_EXENAME);
	platform = (pltid_t) pkt_getvalue(pkt, FIELD_PLATID);

	exec.platform = platform;
	exec.name = str_snew(exe_name);
	
	if (!src_pe || !exe_name || !platform || !exec.name) {
		xprintf ("SWMAN: Error with executable %s\n", exe_name);
		loadexec_error(daemon, &exec, src_pe);
		return 0;
	}

	path = get_path(&exec);

	fd = hwapi_res_open(path, FLOW_READ_ONLY, &proc_blen);
	if (fd < 0) {
		xprintf ("SWMAN: Error opening %s\n", path);
		loadexec_error(daemon, &exec, src_pe);
		return 0;
	}

	exec.fd = fd;
	exec.pinfo.blen = proc_blen;
	exec.pinfo.wlen = ((proc_blen - 1) >> 2) + 1;
	

	/** @todo Depending on mode, some actions should be taken, like for example, compile it
	 * to gather memory requirements
	 */
	pkt_clear(pkt);
	pkt_put(pkt, FIELD_EXENAME, exec.name, str_topkt);
	pkt_put(pkt, FIELD_PINFO, &exec, pinfotopkt);
	daemon_sendto(daemon, SWLOAD_LOADSTART, src_pe, DAEMON_SWLOAD);

	do {
		pkt_clear(pkt);
		pkt_put(pkt, FIELD_EXENAME, exec.name, str_topkt);
		n = pkt_put(pkt, FIELD_BINDATA, &exec, popbindata);
		if (n > 0) {
			daemon_sendto(daemon, SWLOAD_LOADING, src_pe,
					DAEMON_SWLOAD);
			c++;
			if (c == UPLOAD_PACKETS) {
				c = 0;
				hwapi_relinquish_daemon();
			}
		} else if (n < 0) {
			xprintf("SWLOAD: Error reading file\n");
			loadexec_error(daemon, &exec, src_pe);
			return 0;
		}
	} while (n);

	pkt_clear(pkt);
	pkt_put(pkt, FIELD_EXENAME, exec.name, str_topkt);
	pkt_putvalue(pkt, FIELD_CHKSUM, exec.checksum);
	daemon_sendto(daemon, SWLOAD_LOADEND, src_pe, DAEMON_SWLOAD);
	
	str_delete(&exec.name);

	if (fd > 0) {
		hwapi_res_close(fd);
	}

	return 1;
}

struct exec_dirs
{
	pltid_t plt_id;
	char *dir;
};

#define NOF_PLATS 2

/* insert here platform <-> directory
 * where executables can be found 
 */
struct exec_dirs exec_dirs[NOF_PLATS] = { { PLATFORM_LNX26, "linux" }, {
		PLATFORM_C6000, "c6000" } };

char * get_path(struct exec *exec)
{
	int i;
	char *path;

	assert(exec);

	for (i = 0; i < NOF_PLATS; i++) {
		if (exec_dirs[i].plt_id == exec->platform) {
			break;
		}
	}
	if (i == NOF_PLATS) {
		printf("Platform Id unknown. (%d)\n", exec->platform);
		return NULL;
	}

	path = calloc(1, 3 + strlen((char*) EXECS_BASE_DIR)
			+ strlen(exec_dirs[i].dir) + str_sizeof(exec->name));
	if (!path) {
		return NULL;
	}

	sprintf(path, "%s/%s/%s", EXECS_BASE_DIR, exec_dirs[i].dir,
			str_str(exec->name));

	return path;
}

/** SWMAN: Processor executable information
 *
 * Here a processor (SWLOAD sensor) receives the order
 * 	to load an executable and, depending on the case 
 * 	(e.g. the platform requires absolute addressing executables)
 *	it may require to obtain executable memory requirements
 *	in order to realize a pre-allocation. This is obtained using
 * 	this command function.
 */
#ifdef RELOCATING
int
swman_incmd_execinfo (daemon_o daemon)
{
	pkt_o pkt;
	exeobj_o exec;
	int obj_id;
	meminfo_t minfo = NULL;
	pkt = daemon_pkt (daemon);
	if (!pkt)
	return -1;

	/* get executable for this platform from database */
	exec =
	Set_find2 (exec_db, pkt_getval (pkt, FIELD_EXENAME),
			pkt_getval (pkt, FIELD_PLATID), exeobj_findname);
	if (!exec)
	{
		xprintf ("SWMAN: Error can't find executable\n");
	}

	/** TODO: Depending on mode, some actions should be taken, like for example, compile it
	 * to gather memory requirements
	 */
	else
	{
		switch (exeobj_mode (exec))
		{
			case SWMAN_PROCMODE_SOURCE:
			/* compile & link and get memory requirements? */
			break;
			case SWMAN_PROCMODE_COMPILED:
			/* link ... */
			break;
			case SWMAN_PROCMODE_LINKED:
			/* Obtain executable memory requirements */
			minfo = exeobj_meminfo (exec);
			if (!minfo)
			{
				xprintf ("SWMAN: Error getting information for file\n");
			}
			break;
		}
	}

	/* save object id */
	obj_id = id_get (pkt_getval (pkt, FIELD_OBJID));
	pkt_clear (pkt);

	/* put name */
	name_cpy (pkt_putval (pkt, FIELD_EXENAME, sizeof (name_t)),
			exeobj_name (exec));

	/* and associated object id */
	id_put (pkt_putval (pkt, FIELD_OBJID, sizeof (id_t)), obj_id);

	/* Send answer to sensor */
	if (!minfo)
	{
		meminfo_cpy (pkt_putval (pkt, FIELD_MEMINFO, sizeof (meminfo_t)),
				minfo);
		type_set (pkt_oype (pkt), SWLOAD_EXECINFO);
		daemon_sendto (daemon, pkt_srcpe (pkt), DAEMON_SWLOAD);
	}
	else
	{

		/* set type to error */
		type_set (pkt_oype (pkt), SWLOAD_EXECERR);

		/* and send it */
		daemon_sendto (daemon, pkt_srcpe (pkt), DAEMON_SWLOAD);
	}

	return 1;
}
#endif
