/** @file swman_parser.c
 * @brief SWMAN Parsing functions implementation, for applications and executables.
 */

/* standard libraries */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "phal_hw_api.h"
#include "phal_hw_api_man.h"

#include "phid.h"
#include "phal_daemons.h"

/* common objects */
#include "set.h"
#include "str.h"
#include "var.h"
#include "app.h"
#include "rtcobj.h"
#include "phobj.h"
#include "phitf.h"
#include "pkt.h"
#include "daemon.h"

/* include global phal daemon constans */

#include "swman.h"

#include "swman_parser.h"

#include "cfg_parser.h"

char path[128];

/** Function that opens the file, reads to memory
 * and calls parsing function
 */
int swman_parser_app(char *app_name, Set_o objects)
{
	int i, j, n, fd, offset;
	char *buffer;
	cfg_o cfg;
	sect_o subsect, sect;
	key_o key, obj_name, exe_name, itf_name, remote_itf_name,
			remote_obj_name, proc_req, bw_req;
	phobj_o obj;
	phitf_o itf;

	assert (app_name && objects);

	sprintf(path, "%s/%s.%s", APPS_CFG_DIR, app_name, 
	APPS_CFG_EXT);

	offset = hwapi_res_parseall(path, &buffer);
	if (offset < 0) {
		xprintf ("SWMAN: Error parsing file %s\n", path);
		return 0;
	}

	cfg = cfg_new(buffer, offset);
	if (!cfg) {
		xprintf ("SWMAN: Error parsing file %s\n", path);
		free(buffer);
		return 0;
	}

	for (i = 0; i < Set_length(cfg_sections(cfg)); i++) {
		sect = Set_get(cfg_sections(cfg), i);

		if (strcmp(sect_title(sect), "object")) {
			xprintf ("SWMAN: Error unknown section %s\n", sect_title (sect));
			break;
		}

		obj_name = Set_find(sect_keys(sect), "obj_name", key_findname);
		exe_name = Set_find(sect_keys(sect), "exe_name", key_findname);
		proc_req = Set_find(sect_keys(sect), "proc", key_findname);
		
		if (obj_name && exe_name) {
			obj = phobj_new();
			if (!obj) {
				xprintf ("SWMAN: Error creating object\n");
				break;
			}

			str_set(obj->objname, key_value(obj_name, 0,
					PARAM_STRING));
			str_set(obj->exename, key_value(exe_name, 0,
					PARAM_STRING));
			
			if (proc_req) {
				obj->rtc->mips_req = *((float*) key_value(proc_req, 0, PARAM_FLOAT));
			} 
			
			Set_put(objects, obj);
		}

		for (j = 0; j < Set_length(sect_subsects(sect)); j++) {

			subsect = Set_get(sect_subsects(sect), j);

			itf_name = Set_find(sect_keys(subsect), "name",
					key_findname);
			remote_itf_name = Set_find(sect_keys(subsect),
					"remote_itf", key_findname);
			remote_obj_name = Set_find(sect_keys(subsect),
					"remote_obj", key_findname);
				
			bw_req = Set_find(sect_keys(subsect), "bw", key_findname);
			
			if (itf_name && remote_itf_name && remote_obj_name) {

				if (!strcmp(sect_title(subsect), "inputs")) {
					itf = phitf_new(FLOW_READ_ONLY);

				} else if (!strcmp(sect_title(subsect),
						"outputs")) {
					itf = phitf_new(FLOW_WRITE_ONLY);
				} else {
					xprintf ("SWMAN: Error unknown sub-section %s\n", sect_title(subsect));
					break;
				}
				
				if (!itf) {
					xprintf ("SWMAN: Error creating itf\n");
					break;
				}

				str_set(itf->name, key_value(itf_name, 0,
						PARAM_STRING));
				str_set(itf->remotename, key_value(
						remote_itf_name, 0,
						PARAM_STRING));
				str_set(itf->remoteobjname, key_value(
						remote_obj_name, 0,
						PARAM_STRING));

				if (bw_req) {
					itf->bw_req = *((float*) key_value(bw_req, 0, PARAM_FLOAT));					
				}
				
				Set_put(obj->itfs, itf);
			}
		}

		if (j < Set_length(sect_subsects(sect))) {
			xprintf ("SWMAN: Caution some interfaces were not parsed.\n");
		}

	}

	free(buffer);

	if (i < Set_length(cfg_sections(cfg))) {
		xprintf ("SWMAN: Caution some objects were not parsed.\n");
	}

	cfg_delete(&cfg);

	return 1;
}

